# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: check.pl

use strict;

if ($] < 5.006) {
  use vars qw(%bot %cfg %check %fserv %fsquery %irc %msg %services %sock $localtime $netmerge $ts $TMP);
} else {
  our (%bot, %cfg, %check, %fserv, %fsquery, %irc, %msg, %services, %sock, $localtime, $netmerge, $ts, $TMP);
}

my $last = 0;

sub check_timer {
  $ts = time();
  if ($last == $ts) { return 0; }

  $last = $ts;
  $localtime = strftime("%H:%M:%S", localtime);

  if ($cfg{'irc-connected'}) {
    if (!$netmerge && scalar keys %check) {
      foreach (keys %check) {
        my $nick = $check{$_}->{'nick'};
        if (!exists $irc{'client'}->{lc($nick)} || $check{$_}->{'kill'}) {
          if ($check{$_}->{'kill'}) { delete $check{$_}; }
          next;
        }

        if (&check_gline($nick)) { next; }
        if (exists $services{'ScanServ'}) { &scanserv_check($nick); }
        delete $check{$_};
      }
    }

    if ($netmerge && ($ts - $netmerge > 2)) { $netmerge = 0; }
    if ($ts - $cfg{'irc-ping'} > 60) {
      $cfg{'irc-ping'} = $ts;

      foreach (keys %{$irc{'server'}}) {
        if ($irc{'server'}->{$_}->{'pong'}) {
          $irc{'server'}->{$_}->{'ping'} = $ts;
          $irc{'server'}->{$_}->{'pong'} = 0;
          &putsock($sock{'irc'}, ":%s PING %s %s", $cfg{'irc-fakeserv'}, $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hub'}, $irc{'server'}->{$_}->{'name'});
        }
      }
    }

    foreach (keys %{$irc{'client'}}) {
      if ($irc{'client'}->{$_}->{'ignored'} && ($ts - $irc{'client'}->{$_}->{'ignored'} > $cfg{'irc-floodignore'})) {
        $irc{'client'}->{$_}->{'lastcmd'}  = 0;
        $irc{'client'}->{$_}->{'floodcnt'} = 0;
        $irc{'client'}->{$_}->{'ignored'}  = 0;
        &putinfo("y", "Expired ignore: %s (%s\@%s)", $irc{'client'}->{$_}->{'nick'}, $irc{'client'}->{$_}->{'user'}, $irc{'client'}->{$_}->{'host'});
      }
    }
  }

  foreach (keys %fsquery) {
    if ($ts - $fsquery{$_}->{'time'} > 30) { delete $fsquery{$_}; }
  }

  foreach (keys %{$irc{'glined'}}) {
    if (($irc{'glined'}->{$_}->{'expire'} > 0) && ($irc{'glined'}->{$_}->{'set'} + $irc{'glined'}->{$_}->{'expire'} < $ts)) {
      &putinfo("y", "Expired G-line: %s (set by %s %s) [%s]", $irc{'glined'}->{$_}->{'mask'}, $irc{'glined'}->{$_}->{'author'}, &format_date($irc{'glined'}->{$_}->{'set'}), $irc{'glined'}->{$_}->{'reason'});
      if ($cfg{'irc-connected'}) {
        if ($cfg{'server-type'} eq "bahamut") {
          &putsock($sock{'irc'}, ":%s UNSGLINE 0 :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{$_}->{'mask'});
        } elsif ($cfg{'server-type'} eq "ultimate") {
          &putsock($sock{'irc'}, ":%s REMGLINE :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{$_}->{'mask'});
        }
        delete $irc{'glined'}->{$_};
      }
    }
  }

  foreach (keys %{$sock{'list'}}) {
    if ($sock{'list'}->{$_}->{'type'} eq "BOT-B") {
      if ($ts - $sock{'list'}->{$_}->{'time'} > $cfg{'bot-timeout'}) { &del_socket($sock{'list'}->{$_}->{'sock'}); }
    } elsif (($sock{'list'}->{$_}->{'type'} eq "DCC-B") || ($sock{'list'}->{$_}->{'type'} eq "DCC-C") || ($sock{'list'}->{$_}->{'type'} eq "DCC-D")) {
      if ($ts - $sock{'list'}->{$_}->{'time'} > $cfg{'dcc-timeout'}) { &del_socket($sock{'list'}->{$_}->{'sock'}); }
    } elsif (($sock{'list'}->{$_}->{'type'} eq "FSERV-A") && ($ts - $sock{'list'}->{$_}->{'time'} > $cfg{'fserv-timeout'})) {
      &del_socket($sock{'list'}->{$_}->{'sock'});
    } elsif (($sock{'list'}->{$_}->{'type'} eq "FSERV-G") && ($ts - $sock{'list'}->{$_}->{'time'} > $cfg{'fserv-timeout'})) {
      $sock{'list'}->{$_}->{'type'} = "FSERV-Z";
      &del_socket($sock{'list'}->{$_}->{'sock'});
    } elsif ((($sock{'list'}->{$_}->{'type'} eq "FSERV-B") || ($sock{'list'}->{$_}->{'type'} eq "FSERV-C")) && ($ts - $sock{'list'}->{$_}->{'time'} > $cfg{'fserv-timeout'})) {
      $sock{'list'}->{$_}->{'type'} = "FSERV-C";
      &del_socket($sock{'list'}->{$_}->{'sock'});
    } elsif ($sock{'list'}->{$_}->{'type'} eq "FSERV-X") {
      $sock{'list'}->{$_}->{'type'} = "FSERV-C";
      &fserv_send($sock{'list'}->{$_}->{'sock'}, $sock{'list'}->{$_}->{'seek'});
    } elsif (($sock{'list'}->{$_}->{'type'} eq "FSERV") && ($ts - $sock{'list'}->{$_}->{'lastcmd'} + $cfg{'fserv-warn'} > $cfg{'fserv-inactive'})) {
      if (!$sock{'list'}->{$_}->{'warning'}) {
        $sock{'list'}->{$_}->{'warning'} = $ts;
        &putsock($sock{'list'}->{$_}->{'sock'}, $msg{'fserv-warn'});
      } elsif ($ts - $sock{'list'}->{$_}->{'lastcmd'} > $cfg{'fserv-inactive'}) {
        &putsock($sock{'list'}->{$_}->{'sock'}, $msg{'fserv-timeout'});
        &del_socket($sock{'list'}->{$_}->{'sock'});
      }
    }
  }

  if ($cfg{'mysql-host'} && ($ts - $cfg{'last-mysql'} > $cfg{'mysql-sync'})) { &save_mysql(); }
  if ($ts - $cfg{'last-sync'} > $cfg{'db-sync'}) { &save_db(); }
  if ($cfg{'hub-server'} && !$sock{'irc'} && ($ts - $cfg{'irc-attempt'} > $cfg{'irc-reconnect'})) { &connect2irc(); }
  if (exists $services{'PageServ'}) {
    if ($cfg{'icq-uin'} && !$sock{'icq'} && ($ts - $cfg{'icq-attempt'} > $cfg{'icq-reconnect'})) { &connect2icq(0); }
    if ($cfg{'aim-uid'} && !$sock{'aim'} && ($ts - $cfg{'aim-attempt'} > $cfg{'aim-reconnect'})) { &connect2aim(0); }
    if ($cfg{'ymsg-uid'} && !$sock{'ymsg'} && ($ts - $cfg{'ymsg-attempt'} > $cfg{'ymsg-reconnect'})) { &connect2ymsg(0); }
    if ($cfg{'msn-uid'} && !$sock{'msn'} && ($ts - $cfg{'msn-attempt'} > $cfg{'msn-reconnect'})) { &connect2msn(0); }
  }

  if (%bot) {
    foreach (keys %bot) {
      if (!$bot{$_}->{'sock'} && ($ts - $bot{$_}->{'attempt'} > $cfg{'bot-reconnect'})) { &connect2bot($bot{$_}->{'nick'}); }
    }
  }

  if ($cfg{'irc-connected'} && ($ts - $cfg{'irc-pinghub'} > $cfg{'irc-timeout'})) {
    &putinfo("y", "Ping timeout (%dsec) for IRC server %s:%d", $cfg{'irc-timeout'}, (split(/:/, $cfg{'hub-server'}, 0))[0], (split(/:/, $cfg{'hub-server'}, 0))[1]);
    &disconnect("Ping timeout");
  }

  if ($cfg{'irc-connected'}) { &check_tmp(); }
  return 0;
}

sub check_flood {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    $irc{'client'}->{lc($nick)}->{'lastcmd'}  = $ts;
    $irc{'client'}->{lc($nick)}->{'floodcnt'} = 0;
    return 0;
  } elsif ($irc{'client'}->{lc($nick)}->{'ignored'}) {
    return 1;
  } elsif ($ts - $irc{'client'}->{lc($nick)}->{'lastcmd'} < (split(/:/, $cfg{'irc-flood'}, 0))[1]) {
    if (++$irc{'client'}->{lc($nick)}->{'floodcnt'} < (split(/:/, $cfg{'irc-flood'}, 0))[0]) {
      $irc{'client'}->{lc($nick)}->{'lastcmd'} = $ts;
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'flood-irc'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, (split(/:/, $cfg{'irc-flood'}, 0))[0], (split(/:/, $cfg{'irc-flood'}, 0))[1]);
      $irc{'client'}->{lc($nick)}->{'ignored'} = $ts;
      &putinfo("y", "Detected flood from %s!%s\@%s (%dsec ignore)", $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $cfg{'irc-floodignore'});
      return 1;
    }
  } else {
    $irc{'client'}->{lc($nick)}->{'lastcmd'}  = $ts;
    $irc{'client'}->{lc($nick)}->{'floodcnt'} = 0;
    return 0;
  }
}

sub fserv_flood {
  my $socket = shift;

  if (($ts - $sock{'list'}->{$socket}->{'lastcmd'}) < (split(/:/, $cfg{'fserv-flood'}, 0))[1]) {
    if (++$sock{'list'}->{$socket}->{'floodcnt'} < (split(/:/, $cfg{'fserv-flood'}, 0))[0]) {
      $sock{'list'}->{$socket}->{'lastcmd'} = $ts;
      $sock{'list'}->{$socket}->{'warning'} = 0;
      return 0;
    } else {
      &putsock($socket, $msg{'flood-fserv'}, (split(/:/, $cfg{'fserv-flood'}, 0))[0], (split(/:/, $cfg{'fserv-flood'}, 0))[1]);
      &putinfo("y", "%s: detected flood from %s\@%s (closing socket)", $services{'FileServ'}->{'nick'}, $fserv{$sock{'list'}->{$socket}->{'ip'}}->{'nick'}, $sock{'list'}->{$socket}->{'ip'});
      &del_socket($socket);
      return 1;
    }
  } else {
    $sock{'list'}->{$socket}->{'lastcmd'}  = $ts;
    $sock{'list'}->{$socket}->{'floodcnt'} = 0;
    $sock{'list'}->{$socket}->{'warning'}  = 0;
    return 0;
  }
}

sub check_gline {
  my $nick = shift;
  my $test = 0;

  if (!exists $irc{'client'}->{lc($nick)}) { return 0; }

  foreach (keys %{$irc{'glined'}}) {
    my $usermask = &plain2regexp($irc{'glined'}->{$_}->{'usermask'});
    my $hostmask = &plain2regexp($irc{'glined'}->{$_}->{'hostmask'});

    if (($irc{'client'}->{lc($nick)}->{'user'} =~ /$usermask/i) && ($irc{'client'}->{lc($nick)}->{'host'} =~ /$hostmask/i)) {
      if (lc($irc{'client'}->{lc($nick)}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }

      if (my $pid = fork) {
        print $TMP sprintf("%s %s %s %s %d %d %s\n", $irc{'glined'}->{$_}->{'usermask'}, $irc{'glined'}->{$_}->{'hostmask'}, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'glined'}->{$_}->{'author'}, $irc{'glined'}->{$_}->{'set'}, $irc{'glined'}->{$_}->{'expire'}, $irc{'glined'}->{$_}->{'reason'});
        exit(0);
      }

      if ($cfg{'server-type'} eq "hybrid") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $irc{'glined'}->{$_}->{'reason'});
      } elsif ($cfg{'server-type'} eq "bahamut") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'glined'}->{$_}->{'reason'});
      } elsif ($cfg{'server-type'} eq "ultimate") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'glined'}->{$_}->{'reason'});
      }

      $test = 1; last;
    }
  }

  return $test;
}

sub check_tmp {
  my $delete = 1;
  my $buffer;
  my $size = -s $cfg{'tmp-file'};

  if (!$size) { return 0; }

  seek($TMP, 0, 0);
  read($TMP, $buffer, $size);

  my @lines = split(/\n/, $buffer);
  if (!scalar @lines) { return 0; }

  foreach (@lines) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//;

    if ($data !~ /\S/) { next; }

    my ($usermask, $hostmask, $nick, $author, $set, $expire, $reason) = split(/ /, $data, 7);

    if (!exists $irc{'client'}->{lc($nick)}) { next; }
    if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { next; }

    &putinfo("O", "Client G-Lined: [%s] %s\@%s (%s)", $irc{'client'}->{lc($nick)}->{'nick'}, $usermask, $hostmask, $reason);

    $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
    $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
    $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
    $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
    $irc{'seen'}->{lc($nick)}->{'what'} = "QUIT";
    $irc{'seen'}->{lc($nick)}->{'xtra'} = $reason;

    ++$irc{'stats'}->{'kills'}[0];
    $check{lc($nick)}->{'kill'} = $ts;
    &gline($usermask, $hostmask, $author, $reason, $set, $expire, 0);
    &del_client($nick);
  }

  if ($delete) {
    close($TMP);
    unlink($cfg{'tmp-file'});
    if (!open($TMP, "+>>" . $cfg{'tmp-file'})) { &putlog("y", "Unable to write tmp file: %s", $cfg{'tmp-file'}); }
  }

  return 0;
}

1;