# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: client.pl

use strict;

if ($] < 5.006) {
  use vars qw(%cfg %flags %irc %sock $ts);
} else {
  our (%cfg, %flags, %irc, %sock, $ts);
}

sub del_client {
  my $nick   = shift;
  my @host   = split(/\./, $irc{'client'}->{lc($nick)}->{'host'});
  my $suffix = $host[$#host];
  my $domain = $host[$#host - 1] . "." . $suffix;

  foreach (keys %{$irc{'channel'}}) {
    if (exists $irc{'channel'}->{$_}->{'names'}->{lc($nick)}) { delete $irc{'channel'}->{$_}->{'names'}->{lc($nick)}; }
    if (scalar keys %{$irc{'channel'}->{$_}->{'names'}} == 0) { delete $irc{'channel'}->{$_}; }
  }

  --$irc{'server'}->{lc($irc{'client'}->{lc($nick)}->{'serv'})}->{'users'}[0];

  if ($irc{'client'}->{lc($nick)}->{'mode'} & $flags{'client'}->{'o'}) {
    --$irc{'stats'}->{'opers'}[0];
    --$irc{'server'}->{lc($irc{'client'}->{lc($nick)}->{'serv'})}->{'opers'}[0];
  }

  delete $irc{'client'}->{lc($nick)};

  if ($suffix !~ /\d+/) {
    --$irc{'stats'}->{'tld'}->{lc($suffix)}[0];
    --$irc{'stats'}->{'domain'}->{lc($domain)}[0];
  } else {
    --$irc{'stats'}->{'domain'}->{'[non-resolved]'}[0];
  }

  foreach (keys %{$sock{'list'}}) {
    if (($sock{'list'}->{$_}->{'type'} eq "FSERV") && (lc($sock{'list'}->{$_}->{'nick'}) eq lc($nick))) {
      &putsock($sock{'list'}->{$_}->{'sock'}, "Sorry, you MUST stay on the IRC server");
      &del_socket($sock{'list'}->{$_}->{'sock'});
      last;
    }
  }

  return 0;
}

sub client_stats {
  my $nick = shift;
  my $serv = $irc{'client'}->{lc($nick)}->{'serv'};

  my @host   = split(/\./, $irc{'client'}->{lc($nick)}->{'host'});
  my $suffix = $host[$#host];
  my $domain = $host[$#host - 1] . "." . $suffix;

  if ($suffix !~ /\d+/) {
    if (++$irc{'stats'}->{'domain'}->{lc($domain)}[0] > $irc{'stats'}->{'domain'}->{lc($domain)}[1]) {
      $irc{'stats'}->{'domain'}->{lc($domain)}[1] = $irc{'stats'}->{'domain'}->{lc($domain)}[0];
      $irc{'stats'}->{'domain'}->{lc($domain)}[2] = $ts;
    }

    if (++$irc{'stats'}->{'tld'}->{lc($suffix)}[0] > $irc{'stats'}->{'tld'}->{lc($suffix)}[1]) {
      $irc{'stats'}->{'tld'}->{lc($suffix)}[1] = $irc{'stats'}->{'tld'}->{lc($suffix)}[0];
      $irc{'stats'}->{'tld'}->{lc($suffix)}[2] = $ts;
    }
  } else {
    if (++$irc{'stats'}->{'domain'}->{'[non-resolved]'}[0] > $irc{'stats'}->{'domain'}->{'[non-resolved]'}[1]) {
      $irc{'stats'}->{'domain'}->{'[non-resolved]'}[1] = $irc{'stats'}->{'domain'}->{'[non-resolved]'}[0];
      $irc{'stats'}->{'domain'}->{'[non-resolved]'}[2] = $ts;
    }
  }

  if (++$irc{'server'}->{lc($serv)}->{'users'}[0] > $irc{'server'}->{lc($serv)}->{'users'}[1]) {
    $irc{'server'}->{lc($serv)}->{'users'}[1] = $irc{'server'}->{lc($serv)}->{'users'}[0];
    $irc{'server'}->{lc($serv)}->{'users'}[2] = $ts;
  }

  if (scalar keys %{$irc{'client'}} > $irc{'stats'}->{'users'}[0]) {
    @{$irc{'stats'}->{'users'}} = (scalar keys %{$irc{'client'}}, $ts);
    if ($cfg{'irc-connected'} && !$netmerge) { &putinfo("x", "New max client count: %d", $irc{'stats'}->{'users'}[0]); }
  }

  return 0;
}

sub gline {
  my ($usermask, $hostmask, $author, $reason, $set, $expire, $wallops) = (shift, shift, shift, shift, shift, shift, shift);
  my $mask = $usermask . "\@" . $hostmask;

  if (!exists $irc{'glined'}->{lc($mask)}) {
    $irc{'glined'}->{lc($mask)}->{'mask'}     = $mask;
    $irc{'glined'}->{lc($mask)}->{'usermask'} = $usermask;
    $irc{'glined'}->{lc($mask)}->{'hostmask'} = $hostmask;
    $irc{'glined'}->{lc($mask)}->{'set'}      = $set;
    $irc{'glined'}->{lc($mask)}->{'expire'}   = $expire;
    $irc{'glined'}->{lc($mask)}->{'author'}   = $author;
    $irc{'glined'}->{lc($mask)}->{'reason'}   = $reason;
    $irc{'glined'}->{lc($mask)}->{'new'}      = 1;
  }

  if ($cfg{'server-type'} eq "hybrid") {
    &putsock($sock{'irc'}, ":%s GLINE %s %s %s %s %s %s :%s", $cfg{'irc-fakeserv'}, "fxGl1", "fxGl1", "gliner1.fxmakers.net", "gliner1.fxmakers.net", $usermask, $hostmask, $reason);
    &putsock($sock{'irc'}, ":%s GLINE %s %s %s %s %s %s :%s", $cfg{'irc-fakeserv'}, "fxGl2", "fxGl2", "gliner2.fxmakers.net", "gliner2.fxmakers.net", $usermask, $hostmask, $reason);
    &putsock($sock{'irc'}, ":%s GLINE %s %s %s %s %s %s :%s", $cfg{'irc-fakeserv'}, "fxGl3", "fxGl3", "gliner3.fxmakers.net", "gliner3.fxmakers.net", $usermask, $hostmask, $reason);
  } elsif ($cfg{'server-type'} eq "bahamut") {
    &putsock($sock{'irc'}, "SGLINE %d :%s:%s", length($mask), $mask, $reason);
  } elsif ($cfg{'server-type'} eq "ultimate") {
    &putsock($sock{'irc'}, ":%s GLINE %s %d %d %s :%s", $cfg{'irc-fakeserv'}, $mask, $expire ? $set + $expire : 0, $set, $author, $reason);
  }

  if ($wallops) {
    if ($expire > 0) {
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: GLINE from %s for %s [%s] (%s)", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $author, $mask, $reason, &format_time($set + $expire));
    } else {
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: GLINE from %s for %s [%s]", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $author, $mask, $reason);
    }
  }

  return 0;
}

1;