# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: dataserv.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %bot %cfg %flags %irc %msg %services %sock $bit $ts $VERSION);
} else {
  our (%access, %bot, %cfg, %flags, %irc, %msg, %services, %sock, $bit, $ts, $VERSION);
}

my %map;

sub process_dataserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'DataServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'DataServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'DataServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'DataServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'DataServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'DataServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'DataServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "DCC") {
      if ($query =~ /DCC CHAT chat (\d+) (\d+)/i) { &dataserv_dcc($nick, $1, $2); }
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'DataServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (uc($cmd) eq "HELP") {
    &dataserv_help($nick, $arg);
  } elsif ((uc($cmd) eq "ID") || (uc($cmd) eq "IDENTIFY")) {
    &dataserv_identify($nick, $arg);
  } elsif (uc($cmd) eq "MAP") {
    &dataserv_map($nick, $arg);
  } elsif (uc($cmd) eq "RAW") {
    &dataserv_raw($nick, $arg);
  } elsif (uc($cmd) eq "CLONES") {
    &dataserv_clones($nick);
  } elsif (uc($cmd) eq "RESTART") {
    &dataserv_restart($nick);
  } elsif (uc($cmd) eq "REHASH") {
    &dataserv_rehash($nick);
  } elsif (uc($cmd) eq "DIE") {
    &dataserv_die($nick, $arg);
  } elsif (uc($cmd) eq "OMODE") {
    &dataserv_omode($nick, $arg);
  } elsif (uc($cmd) eq "KILL") {
    &dataserv_kill($nick, $arg);
  } elsif (uc($cmd) eq "GLINE") {
    &dataserv_gline($nick, $arg);
  } elsif (uc($cmd) eq "UNGLINE") {
    &dataserv_ungline($nick, $arg);
  } elsif (uc($cmd) eq "STATS") {
    &dataserv_stats($nick, $arg);
  } elsif (uc($cmd) eq "STATUS") {
    &dataserv_status($nick);
  } elsif (uc($cmd) eq "SAVE") {
    &dataserv_save($nick);
  } elsif (uc($cmd) eq "NOTICE") {
    &dataserv_notice($nick, $arg);
  } elsif (uc($cmd) eq "TRACE") {
    &dataserv_trace($nick, $arg);
  } elsif ((uc($cmd) eq "SEEN") || (uc($cmd) eq "WHOWAS")) {
    &dataserv_seen($nick, $arg);
  } elsif (uc($cmd) eq "UNSEEN") {
    &dataserv_unseen($nick, $arg);
  } elsif (uc($cmd) eq "SEENSTAT") {
    &dataserv_seenstat($nick);
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
  }

  return 0;
}

sub dataserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'DataServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'DataServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/dataserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub dataserv_map {
  my ($nick, $server) = (shift, shift);
  my @prefix;

  for (my $inc = 0; $inc < 64; ++$inc) { $prefix[$inc] = ""; }

  if (!$server) {
    $server = $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hub'};
  } else {
    $server = (split(/ /, $server, 0))[0];

    if (!exists $irc{'server'}->{lc($server)}) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-server'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $server);
      return 0;
    }
  }

  &putsock($sock{'irc'}, ":%s %s %s :-- IRC Network Map --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
  &dataserv_dumpmap($nick, $server, 0, @prefix);
  &putsock($sock{'irc'}, ":%s %s %s : ", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
  &putsock($sock{'irc'}, ":%s %s %s :Total: %d user%s on %d server%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick,  $map{lc($nick)}->{'users'},  $map{lc($nick)}->{'users'} > 1 ? "s" : "", $map{lc($nick)}->{'servers'}, $map{lc($nick)}->{'servers'} > 1 ? "s" : "");
  &putsock($sock{'irc'}, ":%s %s %s :-- End of Map --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

  delete $map{lc($nick)};
  &serv_record($nick, $services{'DataServ'}->{'nick'}, "MAP $server");
  return 0;
}

sub dataserv_dumpmap {
  my ($nick, $server, $prompt, @prefix) = (shift, shift, shift, @_);
  my $cnt = scalar keys %{$irc{'server'}->{lc($server)}->{'leaf'}};
  my $output;

  ++$map{lc($nick)}->{'servers'};
  $map{lc($nick)}->{'users'} += $irc{'server'}->{lc($server)}->{'users'}[0];
  $map{lc($nick)}->{'opers'} += $irc{'server'}->{lc($server)}->{'opers'}[0];

  $prefix[$prompt] = "\0";

  foreach (@prefix) {
    if ($_ eq "\0") { last; }
    $output .= $_;
  }

  &putsock($sock{'irc'}, ":%s %s %s :%s%s (%d user%s, %d oper%s) [lag: %s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $output, $irc{'server'}->{lc($server)}->{'name'}, $irc{'server'}->{lc($server)}->{'users'}[0], $irc{'server'}->{lc($server)}->{'users'}[0] > 1 ? "s" : "", $irc{'server'}->{lc($server)}->{'opers'}[0], $irc{'server'}->{lc($server)}->{'opers'}[0] > 1 ? "s" : "", &format_time($irc{'server'}->{lc($server)}->{'lag'} + $ts));

  if ($prompt > 0) {
    $prefix[$prompt - 1] = " ";
    if ($prefix[$prompt - 2] eq "`") { $prefix[$prompt - 2] = " "; }
  }

  $prefix[$prompt] = "|";
  $prefix[$prompt + 1] = "-";

  foreach (keys %{$irc{'server'}->{lc($server)}->{'leaf'}}) {
    if (--$cnt == 0) { $prefix[$prompt] = "`"; }
    &dataserv_dumpmap($nick, $irc{'server'}->{$_}->{'name'}, $prompt + 2, @prefix);
  }

  if ($prompt > 0) { $prefix[$prompt - 1] = "-"; }

  return 0;
}

sub dataserv_dcc {
  my ($nick, $addr, $port) = (shift, shift, shift);
  my $idx = 0;
  my $prefix;

  if (exists $access{lc($nick)} || $irc{'client'}->{lc($nick)}->{'access'}) {
    if (($port < 1024) || ($port > 65536)) {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-invalid'});
      &putlog("y", "Bogus DCC Chat port (%d) specified by %s!%s\@%s", $port, $nick, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
      return 0;
    }

    foreach (keys %{$sock{'list'}}) {
      if (($sock{'list'}->{$_}->{'type'} eq "DCC") && (++$idx > $cfg{'dcc-max'})) {
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-max'});
        return 0;
      }
    }

    my $ip  = inet_ntoa(pack("N", $addr));
    my $dcc = IO::Socket::INET->new(
      Proto    => "tcp",
      PeerAddr => $ip,
      PeerPort => $port,
      Timeout  => 3
    );

    if (!$dcc) {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-failed'});
      return 0;
    }

    $dcc->autoflush(1);

    $sock{'list'}->{$dcc}->{'sock'}   = $dcc;
    $sock{'list'}->{$dcc}->{'host'}   = $dcc->peerhost();
    $sock{'list'}->{$dcc}->{'port'}   = $dcc->peerport();
    $sock{'list'}->{$dcc}->{'type'}   = "DCC";
    $sock{'list'}->{$dcc}->{'fileno'} = $dcc->fileno();
    $sock{'list'}->{$dcc}->{'nick'}   = $irc{'client'}->{lc($nick)}->{'nick'};
    $sock{'list'}->{$dcc}->{'user'}   = $irc{'client'}->{lc($nick)}->{'user'};
    $sock{'list'}->{$dcc}->{'time'}   = $ts;
    vec($bit, $sock{'list'}->{$dcc}->{'fileno'}, 1) = 1;

    if ($irc{'client'}->{lc($nick)}->{'access'}) {
      my $login = $irc{'client'}->{lc($nick)}->{'access'};
      $sock{'list'}->{$dcc}->{'access'}  = $login;
      $sock{'list'}->{$dcc}->{'lastcmd'} = $ts;
      &send_motd($dcc, $cfg{'dcc-motd'});

      if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
        $prefix = "Admin";
      } elsif ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
        $prefix = "Oper";
      } else {
        $prefix = "User";
      }

      &putinfo(0, "%s %s (%s\@%s) has connected", $prefix, $sock{'list'}->{$dcc}->{'nick'}, $sock{'list'}->{$dcc}->{'user'}, $sock{'list'}->{$dcc}->{'host'});
      &putbnet("(%s) %s %s (%s\@%s) has connected", $services{'DataServ'}->{'nick'}, $prefix, $sock{'list'}->{$dcc}->{'nick'}, $sock{'list'}->{$dcc}->{'user'}, $sock{'list'}->{$dcc}->{'host'});
      return 0;
    } else {
      $sock{'list'}->{$dcc}->{'type'}    = "DCC-D";
      $sock{'list'}->{$dcc}->{'warning'} = 0;
      &putinfo("y", "Received DCC connection from %s [%s\@%s:%d]", $sock{'list'}->{$dcc}->{'nick'}, $sock{'list'}->{$dcc}->{'user'}, $sock{'list'}->{$dcc}->{'host'}, $sock{'list'}->{$dcc}->{'port'});
      &putsock($dcc, "You must .identify before you may use commands");
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-unauth'});
    &putinfo("y", "Unauthorized DCC Chat attempt from %s (%s\@%s)", $nick, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    return 0;
  }
}

sub dataserv_identify {
  my ($nick, $arg) = (shift, shift);
  my $count = $arg =~ tr/ / /; ++$count;
  my $login;
  my $passwd;

  if ($count > 1) {
    $login  = (split(/ /, $arg, 0))[0];
    $passwd = (split(/ /, $arg, 0))[1];
  } else {
    $login  = $nick;
    $passwd = (split(/ /, $arg, 0))[0];
  }

  if (!$passwd) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "IDENTIFY [login] <password>");
    return 0;
  }

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'id-already'});
    return 0;
  }

  if (exists $access{lc($login)}) {
    $passwd = (split(/ /, $passwd, 0))[0];

    if ($passwd eq $access{lc($login)}->{'passwd'}) {
      $irc{'client'}->{lc($nick)}->{'access'} = $access{lc($login)}->{'login'};
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'id-ok'});
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "IDENTIFY");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'bad-pass'});
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "failed IDENTIFY");
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_restart {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: RESTART", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'});
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "RESTART");
      &restart();
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_rehash {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      &rehash();
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: REHASH", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'});
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'rehash-done'});
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "REHASH");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_die {
  my ($nick, $reason) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "DIE $reason");
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: DIE %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason ? "[" . $reason . "]" : "");
      &terminate(1, $reason);
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_omode {
  my ($nick, $omode) = (shift, shift);
  my $count = $omode =~ tr/ / /; ++$count;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if ($count < 2) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "OMODE <channel> <modes>");
        return 0;
      }

      my $chan = (split(/ /, $omode, 0))[0];
      my $mode = (split(/ /, $omode, 0))[1];
      my $args = (split(/ /, $omode, 3))[2];
      my @arg; foreach (split(/ /, $args)) { push @arg, $_; }

      if (!exists $irc{'channel'}->{lc($chan)}) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-chan'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $chan);
        return 0;
      } else {
        my $idx  = 0;
        my $what = 1;
        my ($exec, $target);

        foreach (split(//, $mode)) {
          if ($_ eq "+") {
            $what = 1;
            $exec .= "+";
          } elsif ($_ eq "-") {
            $what = 0;
            $exec .= "-";
          } elsif (($_ eq "o") || ($_ eq "h") || ($_ eq "v")) {
            if (exists $irc{'channel'}->{lc($chan)}->{'names'}->{lc($arg[$idx])}) {
              $exec   .= $_;
              $target .= $arg[$idx] . " ";
            }
            ++$idx;
          } elsif (($_ eq "b") || ($_ eq "d") || ($_ eq "e")) {
            $exec   .= $_;
            $target .= $arg[$idx] . " ";
          } elsif (($_ =~ /[fIkl]/) && !(($_ eq "l") && !$what)) {
            $exec   .= $_;
            $target .= $arg[$idx] . " ";
            ++$idx;
          } elsif ($_ =~ /[acilmnpst]/) {
            $exec .= $_;
          }
        }

        $target =~ s/\s+$//;
        &putsock($sock{'irc'}, ":%s MODE %s %s %s", $services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
        &channel_mode($services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
        &putsock($sock{'irc'}, ":%s WALLOPS :%s: OMODE %s %s %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "OMODE $irc{'channel'}->{lc($chan)}->{'channel'} $exec $target");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_stats {
  my ($nick, $switch) = (shift, shift);
  my $idx = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$switch) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "STATS <switch>");
        return 0;
      }

      $switch = (split(/ /, $switch, 0))[0];
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "STATS $switch");

      if (uc($switch) eq "C") {
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing Channels --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Service ] [Channel                  ]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %services) {
          if (!exists $services{$_}->{'channel'}) { next; }

          $idx += scalar keys %{$services{$_}->{'channel'}};
          foreach my $chan (keys %{$services{$_}->{'channel'}}) {
            my $mode;

            if ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "o") {
              $mode = "\@";
            } elsif ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "h") {
              $mode = "\%";
            } elsif ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "v") {
              $mode = "+";
            }

            &putsock($sock{'irc'}, ":%s %s %s :[%-8s] [%-25s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $services{$_}->{'nick'}, $mode . $services{$_}->{'channel'}->{$chan}->{'channel'});
          }
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d channel%s] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;

      } elsif (uc($switch) eq "E") {
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing E-lines --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Hostmask                 ] [Server mask              ]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %{$irc{'noscan'}}) { &putsock($sock{'irc'}, ":%s %s %s :[%-25s] [%-25s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'noscan'}->{$_}->{'hostmask'}, $irc{'noscan'}->{$_}->{'servmask'}); }

        $idx = scalar keys %{$irc{'noscan'}};
        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d mask%s found] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;

      } elsif (uc($switch) eq "I") {
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing I-lines --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Hostmask                 ]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %{$irc{'ignore'}}) { &putsock($sock{'irc'}, ":%s %s %s :  %s\@%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'ignore'}->{$_}->{'hostmask'}, $irc{'ignore'}->{$_}->{'usermask'}); }

        $idx = scalar keys %{$irc{'ignore'}};
        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d mask%s found] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;

      } elsif (uc($switch) eq "O") {
        my ($unick, $uhost, $udcc, $uflag);

        &putsock($sock{'irc'}, ":%s %s %s :-- Authorized Users --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Nickname       ] [Hostname                      ] [Flags] [DCC]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %access) {
          $unick = $access{$_}->{'login'};

          foreach my $client (keys %{$irc{'client'}}) {
            if ($irc{'client'}->{$client}->{'access'} && ($irc{'client'}->{$client}->{'access'} eq $access{$_}->{'login'})) {
              $unick = $irc{'client'}->{$client}->{'nick'};
              $uhost = $irc{'client'}->{$client}->{'host'};

              foreach my $socket (keys %{$sock{'list'}}) {
                if (($sock{'list'}->{$socket}->{'type'} eq "DCC") && ($sock{'list'}->{$socket}->{'access'} eq $access{$_}->{'login'})) { $udcc = " X "; }
              }
            }
          }

          foreach my $flag (keys %{$flags{'access'}}) {
            if ($access{$_}->{'flags'} & $flags{'access'}->{$flag}) { $uflag .= $flag; }
          }

          &putsock($sock{'irc'}, ":%s %s %s :[%-15s] [%-30s] [%-5s] [%-3s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $unick, $uhost, $uflag, $udcc);
          undef($unick); undef($uhost); undef($udcc); undef($uflag);
        }

        $idx = scalar keys %access;
        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d user%s] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;

      } elsif (uc($switch) eq "U") {
        my $started = gmtime($cfg{'started'});
        &putsock($sock{'irc'}, ":%s %s %s :Uptime: %s (started: %s)", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, &format_time($cfg{'started'}), $started);
        return 0;

      } elsif (uc($switch) eq "G") {
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing G-lines --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Mask                     ] [Reason                   ] [Expires       ] [Author         ]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %{$irc{'glined'}}) {
          my $mask   = $irc{'glined'}->{$_}->{'mask'};
          my $expire = $irc{'glined'}->{$_}->{'expire'};
          my $reason = $irc{'glined'}->{$_}->{'reason'};
          if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }

          &putsock($sock{'irc'}, ":%s %s %s :[%-25s] [%-25s] [%-14s] [%-15s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask, $reason, $expire ? &format_time($irc{'glined'}->{$_}->{'set'} + $expire) : "never", $irc{'glined'}->{$_}->{'author'});
        }

        $idx = scalar keys %{$irc{'glined'}};
        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d G-line%s found] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;

      } elsif (uc($switch) eq "B") {
        &putsock($sock{'irc'}, ":%s %s %s :-- Bot Lines --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Bot Nickname      ] [Hostname                      ] [Port ] [Link]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %bot) {
          &putsock($sock{'irc'}, ":%s %s %s :[%-2s %-15s] [%-30s] [%-5s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $bot{$_}->{'prefix'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $bot{$_}->{'port'});
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        return 0;

      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-switch'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $switch);
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_status {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      my $system   = "computer, probably :-)";
      my $autokill = "no";
      my $started  = gmtime($cfg{'started'});
      my $kernel;

      &serv_record($nick, $services{'DataServ'}->{'nick'}, "STATUS");
      &putsock($sock{'irc'}, ":%s %s %s :General Status", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

      if (open(FILE, "</proc/version")) {
        while (<FILE>) {
          if (!$kernel) { ($kernel) = $_ =~ /version (\S+)/; last; }
        }
        close(FILE);
      }

      if ($^O && $kernel) { $system = $^O . " " . $kernel; }

      &putsock($sock{'irc'}, ":%s %s %s :          Running on: %s (Perl %vd)", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $system, $^V);
      &putsock($sock{'irc'}, ":%s %s %s :         Current Hub: %s (connected for: %s)", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'hub-server'}, &format_time($cfg{'irc-connected'}));
      &putsock($sock{'irc'}, ":%s %s %s :Hub Connect Interval: %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'irc-reconnect'});
      &putsock($sock{'irc'}, ":%s %s %s :   Bot Link Interval: %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'bot-reconnect'});
      &putsock($sock{'irc'}, ":%s %s %s :          Started at: %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $started);
      &putsock($sock{'irc'}, ":%s %s %s : ", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      &putsock($sock{'irc'}, ":%s %s %s :Security Settings", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      &putsock($sock{'irc'}, ":%s %s %s :       Flood trigger: %d messages in %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, (split(/:/, $cfg{'irc-flood'}, 0))[0], (split(/:/, $cfg{'irc-flood'}, 0))[1]);
      &putsock($sock{'irc'}, ":%s %s %s :         Ignore Time: %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'irc-floodignore'});

      if (exists $services{'ScanServ'}) {
        my $output;
        my %port;

        foreach (split(/ /, $cfg{'scan-wingate'})) { $port{$_}->{'port'} = $_; }
        foreach (split(/ /, $cfg{'scan-http'}))    { $port{$_}->{'port'} = $_; }
        foreach (split(/ /, $cfg{'scan-socks4'}))  { $port{$_}->{'port'} = $_; }
        foreach (split(/ /, $cfg{'scan-socks5'}))  { $port{$_}->{'port'} = $_; }

        my @list = sort { $port{$a}->{'port'} <=> $port{$b}->{'port'} } keys %port;
        foreach (@list) { $output .= $port{$_}->{'port'} . " "; }

        &putsock($sock{'irc'}, ":%s %s %s :    Ports Monitoring: %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $output);
        if ($cfg{'kill-clones'} && (lc($cfg{'kill-clones'}) eq "yes")) { $autokill = "yes"; }
        &putsock($sock{'irc'}, ":%s %s %s :          Max Clones: %d (autokill: %s)", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'max-clones'}, $autokill);
      }

      &putsock($sock{'irc'}, ":%s %s %s : ", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      &putsock($sock{'irc'}, ":%s %s %s :Databases", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      &putsock($sock{'irc'}, ":%s %s %s :       Save Interval: %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'db-sync'});

      if ($cfg{'mysql-host'}) {
        &putsock($sock{'irc'}, ":%s %s %s :      MySQL Interval: %d seconds", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'mysql-sync'});
        &putsock($sock{'irc'}, ":%s %s %s :        MySQL Server: %s:%d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, (split(/:/, $cfg{'mysql-host'}, 0))[0], (split(/:/, $cfg{'mysql-host'}, 0))[1]);
      }

      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_raw {
  my ($nick, $raw) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      if (!$raw) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "RAW <command>");
        return 0;
      }

      &serv_record($nick, $services{'DataServ'}->{'nick'}, "RAW $raw");
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'perform'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $raw);
      &putsock($sock{'irc'}, $raw);
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_save {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "SAVE");
      &save_db();
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'db-saved'});
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_notice {
  my ($nick, $arg) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      my $target = (split(/ /, $arg, 0))[0]; $target = lc($target);
      my $notice = (split(/ /, $arg, 2))[1];
      my $idx    = 0;

      if (!$target || !$notice) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "NOTICE <target> <message>");
        return 0;
      }

      if ($target eq "-all") {
        foreach (keys %{$irc{'client'}}) {
          if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }

          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{$_}->{'nick'}, $notice);
          ++$idx;
        }
      } elsif ($target eq "-oper") {
        foreach (keys %{$irc{'client'}}) {
          if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
          if ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'}) {
            &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{$_}->{'nick'}, $notice);
            ++$idx;
          }
        }
      } elsif ($target =~ /^-chan:(\S+)$/) {
        if (!$1) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "NOTICE <target> <message>");
          return 0;
        } elsif (!exists $irc{'channel'}->{lc($1)}) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-chan'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $1);
          return 0;
        }

        foreach (keys %{$irc{'channel'}->{lc($1)}->{'names'}}) {
          my $user = $irc{'channel'}->{lc($1)}->{'names'}->{$_}->{'nick'};
          if (lc($irc{'client'}->{lc($user)}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $user, $notice);
          ++$idx;
        }

      } elsif (($target =~ /^-serv:(\S+)$/) || ($target =~ /^-server:(\S+)$/)) {
        if (!$1) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "NOTICE <target> <message>");
          return 0;
        } elsif (!exists $irc{'server'}->{lc($1)}) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-serv'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $1);
          return 0;
        }

        foreach (keys %{$irc{'client'}}) {
          if ((lc($irc{'client'}->{lc($_)}->{'serv'}) eq lc($1)) && (lc($cfg{'irc-fakeserv'}) ne lc($1))) {
            &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{lc($_)}->{'nick'}, $notice);
            ++$idx;
          }
        }

      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "NOTICE <target> <message>");
        return 0;
      }

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'msg-sent'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx);
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "NOTICE $target $notice");
      return 0;

    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_gline {
  my ($nick, $arg) = (shift, shift);
  my $count = $arg =~ tr/ / /; ++$count;
  my $idx   = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "[time] <user\@host> [reason]");
        return 0;
      }

      my ($mask, $reason);
      my $expire = &format_timestr((split(/ /, $arg, 0))[0]);

      if (($count > 1) && ($expire =~ /\d+/) && ($expire > 0)) {
        $mask   = (split(/ /, $arg, 0))[1];
        $reason = (split(/ /, $arg, 3))[2];
      } else {
        $mask   = (split(/ /, $arg, 0))[0];
        $reason = (split(/ /, $arg, 2))[1];
        $expire = 0;
      }

      if ($mask !~ "\@") {
        $mask = "*\@" . $mask;
      } else {
        if ((split(/\@/, $mask, 0))[0] eq "") { $mask = "*" . $mask; }
      }

      foreach (split(//, $mask)) {
        if (($_ eq "\@") || ($_ eq "*") || ($_ eq "?") || ($_ eq ".")) { next; }
        ++$idx;
      }

      if ($idx < $cfg{'nonspecial-chars'}) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-bad'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
        return 0;
      }

      if (exists $irc{'glined'}->{lc($mask)}) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-in'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
        return 0;
      }

      my $usermask = (split(/\@/, $mask, 0))[0]; my $umaskrx = &plain2regexp($usermask);
      my $hostmask = (split(/\@/, $mask, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);

      &gline($usermask, $hostmask, $nick, $reason, $ts, $expire, 1);

      foreach (keys %{$irc{'client'}}) {
        if (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i)) {
          if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
          if ($irc{'client'}->{$_}->{'access'} && ($access{lc($irc{'client'}->{$_}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) { next; }

          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'nick'} = $irc{'client'}->{$_}->{'nick'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'time'} = $ts;
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'user'} = $irc{'client'}->{$_}->{'user'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'host'} = $irc{'client'}->{$_}->{'host'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'what'} = "QUIT";
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'xtra'} = $reason;

          if ($cfg{'server-type'} eq "hybrid") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "bahamut") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "ultimate") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
          }

          ++$irc{'stats'}->{'kills'}[0];
          &del_client($irc{'client'}->{$_}->{'nick'});
        }
      }

      if ($expire > 0) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-tmp'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, &format_time($ts + $expire), $mask);
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "GLINE [$expire] $mask [$reason]");
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-add'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "GLINE $mask [$reason]");
      }

      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_ungline {
  my ($nick, $mask) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$mask) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "UNGLINE <user\@host>");
        return 0;
      }

      $mask = (split(/ /, $mask, 0))[0];

      if (exists $irc{'glined'}->{lc($mask)}) {
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "UNGLINE $mask");
        delete $irc{'glined'}->{lc($mask)};

        if ($cfg{'server-type'} eq "bahamut") {
          &putsock($sock{'irc'}, ":%s UNSGLINE 0 :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{lc($mask)}->{'mask'});
        } elsif ($cfg{'server-type'} eq "ultimate") {
          &putsock($sock{'irc'}, ":%s REMGLINE :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{lc($mask)}->{'mask'});
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-del'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
        &putsock($sock{'irc'}, ":%s WALLOPS :%s: UNGLINE %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $mask);
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-mask'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
      }

      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_kill {
  my ($nick, $arg) = (shift, shift);
  my $idx  = 0;
  my $kill = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "KILL <nick|user\@host> [reason]");
        return 0;
      }

      my $target = (split(/ /, $arg, 0))[0];
      my $reason = (split(/ /, $arg, 2))[1];
      if (!$reason) { $reason = "Killed"; }

      if (exists $irc{'client'}->{lc($target)}) {
        if (lc($irc{'client'}->{lc($target)}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) {
          putsock($sock{'irc'}, ":%s %s %s :Cannot kill Network Service Daemon [%s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{lc($target)}->{'nick'});
          return 0;
        }

        if ($irc{'client'}->{lc($target)}->{'access'} && ($access{lc($irc{'client'}->{lc($target)}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'kill-friend'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'});
          &putlog("y", "%s attempted to KILL protected user %s!%s\@%s %s", $nick, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'}, $reason ? "[" . $reason . "]" : "");
          return 0;
        }


        $irc{'seen'}->{lc($target)}->{'nick'} = $irc{'client'}->{lc($target)}->{'nick'};
        $irc{'seen'}->{lc($target)}->{'time'} = $ts;
        $irc{'seen'}->{lc($target)}->{'user'} = $irc{'client'}->{lc($target)}->{'user'};
        $irc{'seen'}->{lc($target)}->{'host'} = $irc{'client'}->{lc($target)}->{'host'};
        $irc{'seen'}->{lc($target)}->{'what'} = "QUIT";
        $irc{'seen'}->{lc($target)}->{'xtra'} = $reason;

        if ($cfg{'server-type'} eq "hybrid") {
          &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
        } elsif ($cfg{'server-type'} eq "bahamut") {
          &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
        } elsif ($cfg{'server-type'} eq "ultimate") {
          &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'kill-nick'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'}, $reason);
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "KILL $irc{'client'}->{lc($target)}->{'nick'} [$reason]");

        ++$irc{'stats'}->{'kills'}[0];
        &del_client($irc{'client'}->{lc($target)}->{'nick'});
        return 0;
      } else {
        if ($target !~ "\@") {
          $target = "*\@" . $target;
        } else {
          if ((split(/\@/, $target, 0))[0] eq "") { $target = "*" . $target; }
        }

        foreach (split(//, $target)) {
          if (($_ eq "\@") || ($_ eq "*") || ($_ eq "?") || ($_ eq ".")) { next; }
          ++$idx;
        }

        if ($idx < $cfg{'nonspecial-chars'}) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'gline-bad'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $target);
          return 0;
        }

        my $usermask = lc((split(/\@/, $target, 0))[0]); my $umaskrx = &plain2regexp($usermask);
        my $hostmask = lc((split(/\@/, $target, 0))[1]); my $hmaskrx = &plain2regexp($hostmask);

        foreach (keys %{$irc{'client'}}) {
          if (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i)) {
            if ($kill >= $cfg{'max-kill'}) {
              &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'kill-max'});
              last;
            }

            if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
            if ($irc{'client'}->{$_}->{'access'} && ($access{lc($irc{'client'}->{$_}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) { next; }

            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'nick'} = $irc{'client'}->{$_}->{'nick'};
            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'time'} = $ts;
            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'user'} = $irc{'client'}->{$_}->{'user'};
            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'host'} = $irc{'client'}->{$_}->{'host'};
            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'what'} = "QUIT";
            $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'xtra'} = $reason;

            if ($cfg{'server-type'} eq "hybrid") {
              &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
            } elsif ($cfg{'server-type'} eq "bahamut") {
              &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $irc{'client'}->{$_}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
            } elsif ($cfg{'server-type'} eq "ultimate") {
              &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $nick, $services{'DataServ'}->{'nick'});
            }

            ++$irc{'stats'}->{'kills'}[0]; ++$kill;
            &del_client($irc{'client'}->{$_}->{'nick'});
          }
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'kill-host'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $kill, $target);
        &serv_record($nick, $services{'DataServ'}->{'nick'}, "KILL $target [$reason]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_clones {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      my $idx = 0;
      my $cln = 0;
      my %clones;

      &putsock($sock{'irc'}, ":%s %s %s :-- Listing Clones --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

      foreach (keys %{$irc{'client'}}) {
        if ($irc{'client'}->{$_}->{'serv'} ne $cfg{'irc-fakeserv'}) { push @{$clones{lc($irc{'client'}->{$_}->{'host'})}}, $irc{'client'}->{$_}->{'nick'}; }
      }

      foreach (keys %clones) {
        if (scalar @{$clones{$_}} > 1) {
          ++$cln;
          &putsock($sock{'irc'}, ":%s %s %s :*\@%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $_);
          foreach my $clone (@{$clones{$_}}) {
            ++$idx;
            &putsock($sock{'irc'}, ":%s %s %s :  %02d) %-15s [%s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $irc{'client'}->{lc($clone)}->{'nick'}, $irc{'client'}->{lc($clone)}->{'serv'});
          }

          $idx = 0;
        }
      }

      &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d clone%s found] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cln, $cln > 1 ? "s" : "");
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "CLONES");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_trace {
  my ($nick, $arg) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      my ($oper, $nonoper, $long) = (0, 0, 0);
      my ($i, $j, $s) = (0, 0, 0); my $server;
      my $search = my $trace = (split(/ /, $arg, 0))[0];
      my $option = (split(/ /, $arg, 2))[1];

      if (!$search) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "TRACE <pattern> [options]");
        return 0;
      }

      if ($option) {
        foreach (split(/ /, $option)) {
          ++$i;

          if (lc($_) eq "-oper") {
            $oper = 1;
          } elsif (lc($_) eq "-nonoper") {
            $nonoper = 1;
          } elsif (lc($_) eq "-long") {
            $long = 1;
          } elsif (lc($_) eq "-server") {
            $s = $i + 1;
          } elsif ($s == $i) {
            $server = $_;
          }
        }

        if ($s && !$server) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "TRACE <pattern> [options]");
          return 0;
        } elsif ($server && !exists $irc{'server'}->{lc($server)}) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-server'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $server);
          return 0;
        }
      }

      if ($trace !~ "\@") {
        $trace = "*\@" . $trace;
      } else {
        if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
      }

      my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
      my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
      my $searchrx = &plain2regexp($search);

      &putsock($sock{'irc'}, ":%s %s %s :-- Listing Results --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

      foreach (keys %{$irc{'client'}}) {
        if (($irc{'client'}->{$_}->{'nick'} =~ /$searchrx/i) || (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i))) {
          if ($nonoper && ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'})) { next; }
          if ($oper && !($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'})) { next; }
          if ($server && (lc($irc{'client'}->{$_}->{'serv'}) ne lc($server))) { next; }
          ++$j;

          if ($long) {
            my $modes;
            my $chanlist;

            foreach my $flag (keys %{$flags{'client'}}) {
              if ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{$flag}) { $modes .= $flag; }
            }

            foreach my $chan (keys %{$irc{'channel'}}) {
              if (exists $irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}) {
                if ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'o'}) {
                  $chanlist .= "\@" . $irc{'channel'}->{$chan}->{'channel'} . " ";
                } elsif ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'h'}) {
                  $chanlist .= "\%" . $irc{'channel'}->{$chan}->{'channel'} . " ";
                } elsif ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'v'}) {
                  $chanlist .= "+" . $irc{'channel'}->{$chan}->{'channel'} . " ";
                } else {
                  $chanlist .= $irc{'channel'}->{$chan}->{'channel'} . " ";
                }
              }
            }

            $chanlist =~ tr/ //s; $chanlist =~ s/\s+$//;

            if ($j > 1) { &putsock($sock{'irc'}, ":%s %s %s : ", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick); }
            &putsock($sock{'irc'}, ":%s %s %s :User:     %s (%s\@%s)", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{$_}->{'nick'}, $irc{'client'}->{$_}->{'user'}, $irc{'client'}->{$_}->{'host'});
            &putsock($sock{'irc'}, ":%s %s %s :Realname: %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{$_}->{'real'});
            &putsock($sock{'irc'}, ":%s %s %s :Server:   %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{$_}->{'serv'});
            &putsock($sock{'irc'}, ":%s %s %s :Usermode: +%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $modes);
            &putsock($sock{'irc'}, ":%s %s %s :Signon:   %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, &format_time($irc{'client'}->{$_}->{'time'}));
            if ($chanlist) { &putsock($sock{'irc'}, ":%s %s %s :Channel:  %s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $chanlist); }
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :%s!%s\@%s (%s) [%s]", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{$_}->{'nick'}, $irc{'client'}->{$_}->{'user'}, $irc{'client'}->{$_}->{'host'}, $irc{'client'}->{$_}->{'real'}, $irc{'client'}->{$_}->{'serv'});
          }
        }
      }

      &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d match%s found] --", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $j, $j > 1 ? "es" : "");
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "TRACE " . $search . ($option ? " [" . $option . "]" : ""));
      return 0;

    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_seen {
  my ($nick, $search) = (shift, shift);
  my $idx  = 0;
  my @seen = ();
  my @list = ();
  my $format;

  if (!$search) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SEEN <nick|hostmask>");
    return 0;
  }

  my $trace = $search = (split(/ /, $search, 0))[0];

  if ($trace !~ "\@") {
    $trace = "*\@" . $trace;
  } else {
    if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
  }

  my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
  my $searchrx = &plain2regexp($search);

  if (($search =~ /\*/) || ($search =~ /\?/)) {
    foreach (keys %{$irc{'seen'}}) {
      if (($irc{'seen'}->{$_}->{'nick'} =~ /$searchrx/i) || (($irc{'seen'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'seen'}->{$_}->{'host'} =~ /$hmaskrx/i)) || (($irc{'seen'}->{$_}->{'what'} eq "NICK") && ($irc{'seen'}->{$_}->{'xtra'} =~ /$searchrx/i))) { ++$idx; push @seen, $_; }
    }

    if (!$idx) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-norec'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $search);
      return 0;
    } elsif ($idx > $cfg{'seen-maxwild'}) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-refine'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'seen-maxwild'}, $search);
      return 0;
    } else {
      if ($idx > $cfg{'seen-maxprint'}) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-recent'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $search);
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-print'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $search);
      }

      $idx  = 0;
      @list = sort { $irc{'seen'}->{$b}->{'time'} <=> $irc{'seen'}->{$a}->{'time'} } @seen;

      foreach (@list) {
        if ($irc{'seen'}->{$_}->{'what'} eq "NICK") {
          $format = $msg{'seen-nick'};
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "JOIN") {
          $format = $msg{'seen-join'};
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "PART") {
          $format = $msg{'seen-part'};
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "KICK") {
          $format = $msg{'seen-kick'};
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "QUIT") {
          $format = $msg{'seen-quit'};
        }

        &putsock($sock{'irc'}, ":%s %s %s :  %d) $format", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, ++$idx, $irc{'seen'}->{$_}->{'nick'}, $irc{'seen'}->{$_}->{'user'}, $irc{'seen'}->{$_}->{'host'}, $irc{'seen'}->{$_}->{'xtra'}, &format_time($irc{'seen'}->{$_}->{'time'}));
        if ($idx >= $cfg{'seen-maxprint'}) { last; }
      }
    }
  } else {
    if (exists $irc{'client'}->{lc($search)}) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-online'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'client'}->{lc($search)}->{'nick'});
    } elsif (exists $irc{'seen'}->{lc($search)}) {
      if ($irc{'seen'}->{lc($search)}->{'what'} eq "NICK") {
        $format = $msg{'seen-nick'};
      } elsif ($irc{'seen'}->{lc($search)}->{'what'} eq "JOIN") {
        $format = $msg{'seen-join'};
      } elsif ($irc{'seen'}->{lc($search)}->{'what'} eq "PART") {
        $format = $msg{'seen-part'};
      } elsif ($irc{'seen'}->{lc($search)}->{'what'} eq "KICK") {
        $format = $msg{'seen-kick'};
      } elsif ($irc{'seen'}->{lc($search)}->{'what'} eq "QUIT") {
        $format = $msg{'seen-quit'};
      }

      &putsock($sock{'irc'}, ":%s %s %s :$format", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'seen'}->{lc($search)}->{'nick'}, $irc{'seen'}->{lc($search)}->{'user'}, $irc{'seen'}->{lc($search)}->{'host'}, $irc{'seen'}->{lc($search)}->{'xtra'}, &format_time($irc{'seen'}->{lc($search)}->{'time'}));
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-norec'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $search);
    }
  }

  &serv_record($nick, $services{'DataServ'}->{'nick'}, "SEEN $search");
  return 0;
}

sub dataserv_unseen {
  my ($nick, $search) = (shift, shift);
  my $idx = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      if (!$search) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "UNSEEN <nick|hostmask>");
        return 0;
      }

      my $trace = $search = (split(/ /, $search, 0))[0];

      if ($trace !~ "\@") {
        $trace = "*\@" . $trace;
      } else {
        if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
      }

      my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
      my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
      my $searchrx = &plain2regexp($search);

      if (($search =~ /\*/) || ($search =~ /\?/)) {
        foreach (keys %{$irc{'seen'}}) {
          if (($irc{'seen'}->{$_}->{'nick'} =~ /$searchrx/i) || (($irc{'seen'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'seen'}->{$_}->{'host'} =~ /$hmaskrx/i)) || (($irc{'seen'}->{$_}->{'what'} eq "NICK") && ($irc{'seen'}->{$_}->{'xtra'} =~ /$searchrx/i))) {
            delete $irc{'seen'}->{$_};
            ++$idx;
          }
        }

        if (!$idx) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-match'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $search);
          return 0;
        }
      } else {
        if (exists $irc{'seen'}->{lc($search)}) {
          delete $irc{'seen'}->{lc($search)};
          ++$idx;
        } else {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-match'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $search);
          return 0;
        }
      }

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'seen-del'}", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $search);
      &serv_record($nick, $services{'DataServ'}->{'nick'}, "UNSEEN $search");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_seenstat {
  my $nick = shift;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      my ($nicks, $joins, $parts, $kicks, $quits) = (0, 0, 0, 0, 0);

      foreach (keys %{$irc{'seen'}}) {
        if ($irc{'seen'}->{$_}->{'what'} eq "NICK") {
          ++$nicks;
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "JOIN") {
          ++$joins;
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "PART") {
          ++$parts;
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "KICK") {
          ++$kicks;
        } elsif ($irc{'seen'}->{$_}->{'what'} eq "QUIT") {
          ++$quits;
        }
      }

      &putsock($sock{'irc'}, ":%s %s %s :Total number of nicks seen: %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, scalar keys %{$irc{'seen'}});
      &putsock($sock{'irc'}, ":%s %s %s :... of these, NICK msgs:    %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $nicks);
      &putsock($sock{'irc'}, ":%s %s %s :... of these, JOIN msgs:    %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $joins);
      &putsock($sock{'irc'}, ":%s %s %s :... of these, PART msgs:    %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $parts);
      &putsock($sock{'irc'}, ":%s %s %s :... of these, KICK msgs:    %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $kicks);
      &putsock($sock{'irc'}, ":%s %s %s :... of these, QUIT msgs:    %d", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $quits);

      &serv_record($nick, $services{'DataServ'}->{'nick'}, "SEENSTAT");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub dataserv_chanquery {
  my ($nick, $chan, $query) = (shift, shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      my $cmd = uc((split(/ /, $query, 0))[0]);
      my $arg = (split(/ /, $query, 2))[1];
      my ($what, $mode);

      if (($cmd eq "!OP") || ($cmd eq "!HALFOP") || ($cmd eq "!VOICE") || ($cmd eq "!DEOP") || ($cmd eq "!DEHALFOP") || ($cmd eq "!DEVOICE")) {
        if (($cmd eq "!OP") || ($cmd eq "!HALFOP") || ($cmd eq "!VOICE")) { $what = "+"; }
        if (($cmd eq "!DEOP") || ($cmd eq "!DEHALFOP") || ($cmd eq "!DEVOICE")) { $what = "-"; }

        if (($cmd eq "!OP") || ($cmd eq "!DEOP")) { $mode = "o"; }
        if (($cmd eq "!HALFOP") || ($cmd eq "!DEHALFOP")) { $mode = "h"; }
        if (($cmd eq "!VOICE") || ($cmd eq "!DEVOICE")) { $mode = "v"; }

        if (!$arg) {
          &putsock($sock{'irc'}, ":%s MODE %s %s%s %s", $services{'DataServ'}->{'nick'}, $chan, $what, $mode, $nick);
          &channel_mode($services{'DataServ'}->{'nick'}, $chan, $what . $mode, $nick);
          &serv_record($nick, $services{'DataServ'}->{'nick'}, "CHANMODE on $chan: $what$mode $nick");
          return 0;
        } else {
          my $inc    = 0;
          my $cnt    = 0;
          my $exec   = $what;
          my $count  = $arg =~ tr/ / /; ++$count;
          my ($target, $who);

          foreach (split(/ /, $arg)) {
            ++$inc;

            if (exists $irc{'channel'}->{lc($chan)} && exists $irc{'channel'}->{lc($chan)}->{'names'}->{lc($_)}) {
              $who    .= $_;
              $exec   .= $mode;
              $target .= $_ . " ";
              $cnt = $who =~ tr/ / /; ++$cnt;

              if (!($cnt % $cfg{'irc-simulmode'})) {
                &putsock($sock{'irc'}, ":%s MODE %s %s%s %s", $services{'DataServ'}->{'nick'}, $chan, $what, $mode x $cfg{'irc-simulmode'}, $who);
                undef $who;
              } elsif ($count == $inc) {
                &putsock($sock{'irc'}, ":%s MODE %s %s%s %s", $services{'DataServ'}->{'nick'}, $chan, $what, $mode x $cfg{'irc-simulmode'}, $who);
              } else {
                $who .= " ";
              }
            } else {
              if ($who && ($count == $inc)) {
                $who =~ s/\s+$//;
                &putsock($sock{'irc'}, ":%s MODE %s %s%s %s", $services{'DataServ'}->{'nick'}, $chan, $what, $mode x $cfg{'irc-simulmode'}, $who);
              }
            }
          }

          if ($target) {
            $target =~ s/\s+$//;
            &channel_mode($services{'DataServ'}->{'nick'}, $chan, $exec, $target);
            &serv_record($nick, $services{'DataServ'}->{'nick'}, "CHANMODE on $chan: $exec $target");
          }

          return 0;
        }
      } elsif ($cmd eq "!KICK") {
      } elsif ($cmd eq "!BAN") {
      } elsif ($cmd eq "!BANKICK") {
      } elsif ($cmd eq "!MODE") {
      }

      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  }
}

1;