# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: dcc.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %bot %cfg %flags %irc %msg %services %sock $ts);
} else {
  our (%access, %bot, %cfg, %flags, %irc, %msg, %services, %sock, $ts);
}

sub dcc_parse {
  my ($socket, $buffer) = (shift, shift);

  $sock{'list'}->{$socket}->{'lastcmd'} = $ts;

  if ($buffer =~ /^\./) {
    my $cmd = (split(/ /, $buffer, 0))[0]; $cmd = substr($cmd, 1);
    my $arg = (split(/ /, $buffer, 2))[1];

    if ((uc($cmd) eq "IDENTIFY") || (uc($cmd) eq "ID")) {
      &dcc_identify($socket, $arg);
    } elsif (uc($cmd) eq "HELP") {
      &dcc_help($socket, $arg);
    } elsif (uc($cmd) eq "SAVE") {
      &dcc_save($socket);
    } elsif ((uc($cmd) eq "QUIT") || (uc($cmd) eq "EXIT") || (uc($cmd) eq "BYE")) {
      &dcc_quit($socket, $arg);
    } elsif (uc($cmd) eq "UMODE") {
      &dcc_umode($socket, $arg);
    } elsif (uc($cmd) eq "STATS") {
      &dcc_stats($socket, $arg);
    } elsif (uc($cmd) eq "RAW") {
      &dcc_raw($socket, $arg);
    } elsif (uc($cmd) eq "BOOT") {
      &dcc_boot($socket, $arg);
    } elsif (uc($cmd) eq "DIE") {
      &dcc_die($socket, $arg);
    } elsif (uc($cmd) eq "WHO") {
      &dcc_who($socket);
    } elsif (uc($cmd) eq "CLONES") {
      &dcc_clones($socket);
    } elsif (uc($cmd) eq "TRACE") {
      &dcc_trace($socket, $arg);
    } elsif (uc($cmd) eq "MOTD") {
      &dcc_motd($socket);
    } elsif (uc($cmd) eq "RESTART") {
      &dcc_restart($socket);
    } elsif (uc($cmd) eq "REHASH") {
      &dcc_rehash($socket);
    } elsif (uc($cmd) eq "GLINE") {
      &dcc_gline($socket, $arg);
    } elsif (uc($cmd) eq "UNGLINE") {
      &dcc_ungline($socket, $arg);
    } elsif (uc($cmd) eq "KILL") {
      &dcc_kill($socket, $arg);
    } elsif (uc($cmd) eq "NOTICE") {
      &dcc_notice($socket, $arg);
    } elsif (uc($cmd) eq "OMODE") {
      &dcc_omode($socket, $arg);
    } elsif (uc($cmd) eq "LINK") {
      &dcc_link($socket, $arg);
    } elsif (uc($cmd) eq "UNLINK") {
      &dcc_unlink($socket, $arg);
    } else {
      &putsock($socket, $msg{'unkn-cmd'}, $cmd);
    }

    return 0;
  } else {
    foreach (keys %{$sock{'list'}}) {
      if ($sock{'list'}->{$_}->{'type'} eq "DCC") { &putsock($sock{'list'}->{$_}->{'sock'}, "<%s\@%s> %s", $sock{'list'}->{$socket}->{'nick'}, $services{'DataServ'}->{'nick'}, $buffer); }
    }

    &putbnet("<%s\@%s> %s", $sock{'list'}->{$socket}->{'nick'}, $services{'DataServ'}->{'nick'}, $buffer);
    return 0;
  }
}

sub dcc_identify {
  my ($socket, $passwd) = (shift, shift);
  my $prefix;

  if (!$passwd) {
    &putsock($socket, $msg{'syntax'}, "IDENTIFY <password>");
    return 0;
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-D") {
    $passwd = (split(/ /, $passwd, 0))[0];
    my $login = $sock{'list'}->{$socket}->{'nick'};

    if ($access{lc($login)}->{'passwd'} eq $passwd) {
      $sock{'list'}->{$socket}->{'type'}    = "DCC";
      $sock{'list'}->{$socket}->{'access'}  = $access{lc($login)}->{'login'};
      $sock{'list'}->{$socket}->{'lastcmd'} = $ts;
      &send_motd($socket, $cfg{'dcc-motd'});

      if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
        $prefix = "Admin";
      } elsif ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
        $prefix = "Oper";
      } else {
        $prefix = "User";
      }

      &putinfo(0, "%s %s (%s\@%s) has connected", $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
      &putbnet("(%s) %s %s (%s\@%s) has connected", $services{'DataServ'}->{'nick'}, $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
      return 0;
    } else {
      &putsock($socket, $msg{'bad-pass'});
      &putinfo("y", "Unauthorized DCC Chat attempt from %s (%s\@%s)", $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
      &del_socket($socket);
      return 0;
    }
  } else {
    &putsock($socket, $msg{'id-already'});
    return 0;
  }
}

sub dcc_umode {
  my ($socket, $mode) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};
  my $what  = 1;
  my $umodes; 

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if ($mode) {
      foreach (split(//, $mode)) {
        if ($_ eq "+") {
          $what = 1;
        } elsif ($_ eq "-") {
          $what = 0;
        } elsif ($flags{'event'}->{$_}) {
          if ($what) {
            $access{lc($login)}->{'event'} |= $flags{'event'}->{$_};
          } else {
            $access{lc($login)}->{'event'} &= ~$flags{'event'}->{$_};
          }
        }
      }
    }

    foreach (keys %{$flags{'event'}}) {
      if ($access{lc($login)}->{'event'} & $flags{'event'}->{$_}) { $umodes .= $_; }
    }

    if (!$mode) {
      &putsock($socket, "Your usermodes are currently [+%s]", $umodes);
    } else {
      &putsock($socket, "Your usermodes are now [+%s]", $umodes);
    }

    &dcc_record($socket, "UMODE $umodes");
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_who {
  my $socket = shift;
  my $connex = 0;
  my $idx    = 0;

  &putsock($socket, "Current users: (\@ = Admin, + = Oper)");

  foreach (keys %{$sock{'list'}}) {
    if ($sock{'list'}->{$_}->{'type'} eq "DCC") {
      my $prefix;
      my $login = $sock{'list'}->{$_}->{'access'};
      my $uhost = $sock{'list'}->{$_}->{'user'} . "\@" . $sock{'list'}->{$_}->{'host'};

      if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
        $prefix = "\@";
      } elsif ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
        $prefix = "+";
      } else {
        $prefix = " ";
      }

      &putsock($socket, "%s%-16s %-30s idle: %s", $prefix, $sock{'list'}->{$_}->{'nick'}, "(" . $uhost . ")", &format_time($sock{'list'}->{$_}->{'lastcmd'}));
    }
  }

  if (%bot) {
    foreach (keys %bot) {
      ++$idx;
      if ($bot{$_}->{'sock'}) { $connex = 1; last; }
    }

    if ($connex) {
      &putsock($socket, "Bot connection%s:", $idx > 1 ? "s" : "");

      foreach (keys %bot) {
        if ($bot{$_}->{'sock'}) { &putsock($socket, "  %s %-15s [%s:%d]", $bot{$_}->{'prefix'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $sock{'list'}->{$bot{$_}->{'sock'}}->{'port'}); }
      }
    }
  }

  &dcc_record($socket, "WHO");
  return 0;
}

sub dcc_help {
  my ($socket, $arg) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};

  if (!$arg) {
    &dcc_record($socket, "HELP");
    $arg = "index";
  } else {
    &dcc_record($socket, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/dataserv/dcc/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($socket, $msg{'no-help'}, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      $help =~ s/^\%\+a//;
      if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($socket, $help); }
    } elsif ($help =~ /^\%\+o/) {
      $help =~ s/^\%\+o//;
      if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($socket, $help); }
    } else {
      &putsock($socket, $help);
    }
  }

  close(FILE);
  return 0;
}

sub dcc_quit {
  my ($socket, $reason) = (shift, shift);

  &putsock($socket, "Closing connection: $reason");
  &putinfo(0, "%s (%s\@%s) has left the partyline %s", $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'}, $reason ? "[" . $reason . "]" : "");
  &putbnet("(%s) %s (%s\@%s) has left the partyline %s", $services{'DataServ'}->{'nick'}, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'}, $reason ? "[" . $reason . "]" : "");
  &dcc_record($socket, "QUIT $reason");
  &del_socket($socket);
  return 0;
}

sub dcc_boot {
  my ($socket, $arg) = (shift, shift);
  my $login  = $sock{'list'}->{$socket}->{'access'};
  my $target = (split(/ /, $arg, 0))[0];
  my $reason = (split(/ /, $arg, 2))[1];

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    if (!$target) {
      &putsock($socket, $msg{'syntax'}, "BOOT <nick> [reason]");
      return 0;
    }

    foreach (keys %{$sock{'list'}}) {
      if (($sock{'list'}->{$_}->{'type'} eq "DCC") && (lc($sock{'list'}->{$_}->{'nick'}) eq lc($target))) {
        &dcc_record($socket, "BOOT $sock{'list'}->{$_}->{'nick'} [$reason]");
        &putsock($sock{'list'}->{$_}->{'sock'}, $msg{'boot'}, $login, $reason);
        &putbnet("(%s) %s has been booted by %s %s", $services{'DataServ'}->{'nick'}, $sock{'list'}->{$_}->{'nick'}, $login, $reason ? "[" . $reason . "]" : "");
        &putsock($socket, $msg{'booted'}, $sock{'list'}->{$_}->{'nick'}, $reason);
        &del_socket($sock{'list'}->{$_}->{'sock'});
        return 0;
      }
    }

    &putsock($socket, $msg{'no-user'}, $target);
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_motd {
  my $socket = shift;

  &send_motd($socket, $cfg{'dcc-motd'});
  &dcc_record($socket, "MOTD");
  return 0;
}

sub dcc_stats {
  my ($socket, $switch) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};
  my $idx   = 0;

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$switch) {
      &putsock($socket, $msg{'syntax'}, "STATS <switch>");
      return 0;
    }

    $switch = (split(/ /, $switch, 0))[0];
    &dcc_record($socket, "STATS $switch");

    if (uc($switch) eq "C") {
      &putsock($socket, "-- Listing Channels --");
      &putsock($socket, "[Service Nick] [Channel                  ]");

      foreach (keys %services) {
        if (!exists $services{$_}->{'channel'}) { next; }
        $idx += scalar keys %{$services{$_}->{'channel'}};

        foreach my $chan (keys %{$services{$_}->{'channel'}}) {
          my $mode;

          if ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "o") {
            $mode = "\@";
          } elsif ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "h") {
            $mode = "\%";
          } elsif ($services{$_}->{'channel'}->{lc($chan)}->{'mode'} eq "v") {
            $mode = "+";
          }

          &putsock($socket, "[%-12s] [%-25s]", $services{$_}->{'nick'}, $mode . $services{$_}->{'channel'}->{$chan}->{'channel'});
        }
      }

      &putsock($socket, "-- End of List [%d channel%s] --", $idx, $idx > 1 ? "s" : "");
      return 0;

    } elsif (uc($switch) eq "E") {
      &putsock($socket, "-- Listing E-lines --");
      &putsock($socket, "[Hostmask                 ] [Server mask              ]");

      foreach (keys %{$irc{'noscan'}}) { &putsock($socket, "[%-25s] [%-25s]", $irc{'noscan'}->{$_}->{'hostmask'}, $irc{'noscan'}->{$_}->{'servmask'}); }

      $idx = scalar keys %{$irc{'noscan'}};
      &putsock($socket, "-- End of List [%d mask%s found] --", $idx, $idx > 1 ? "s" : "");
      return 0;

    } elsif (uc($switch) eq "I") {
      &putsock($socket, "-- Listing I-lines --");
      &putsock($socket, "[Hostmask                 ]");

      foreach (keys %{$irc{'ignore'}}) { &putsock($socket, "  %s\@%s", $irc{'ignore'}->{$_}->{'hostmask'}, $irc{'ignore'}->{$_}->{'usermask'}); }

      $idx = scalar keys %{$irc{'ignore'}};
      &putsock($socket, "-- End of List [%d mask%s found] --", $idx, $idx > 1 ? "s" : "");
      return 0;

    } elsif (uc($switch) eq "O") {
      my ($unick, $uhost, $udcc, $uflag);

      &putsock($socket, "-- Authorized Users --");
      &putsock($socket, "[Nickname       ] [Hostname                      ] [Flags] [DCC]");

      foreach (keys %access) {
        $unick = $access{$_}->{'login'};

        foreach my $client (keys %{$irc{'client'}}) {
          if ($irc{'client'}->{$client}->{'access'} && ($irc{'client'}->{$client}->{'access'} eq $access{$_}->{'login'})) {
            $unick = $irc{'client'}->{$client}->{'nick'};
            $uhost = $irc{'client'}->{$client}->{'host'};

            foreach my $usock (keys %{$sock{'list'}}) {
              if (($sock{'list'}->{$usock}->{'type'} eq "DCC") && ($sock{'list'}->{$usock}->{'access'} eq $access{$_}->{'login'})) { $udcc = " X "; }
            }
          }
        }

        foreach my $flag (keys %{$flags{'access'}}) {
          if ($access{$_}->{'flags'} & $flags{'access'}->{$flag}) { $uflag .= $flag; }
        }

        &putsock($socket, "[%-15s] [%-30s] [%-5s] [%-3s]", $unick, $uhost, $uflag, $udcc);
        undef($unick); undef($uhost); undef($udcc); undef($uflag);
      }

      $idx = scalar keys %access;
      &putsock($socket, "-- End of List [%d user%s] --", $idx, $idx > 1 ? "s" : "");
      return 0;

    } elsif (uc($switch) eq "U") {
      my $started = gmtime($cfg{'started'});
      &putsock($socket, "Uptime: %s (started: %s)", &format_time($cfg{'started'}), $started);
      return 0;

    } elsif (uc($switch) eq "G") {
      &putsock($socket, "-- Listing G-lines --");
      &putsock($socket, "[Mask                     ] [Reason                   ] [Expires       ] [Author         ]");

      foreach (keys %{$irc{'glined'}}) {
        my $mask   = $irc{'glined'}->{$_}->{'mask'};
        my $expire = $irc{'glined'}->{$_}->{'expire'};
        my $reason = $irc{'glined'}->{$_}->{'reason'};
        if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }

        &putsock($socket, "[%-25s] [%-25s] [%-14s] [%-15s]", $mask, $reason, $expire ? &format_time($irc{'glined'}->{$_}->{'set'} + $expire) : "never", $irc{'glined'}->{$_}->{'author'});
      }

      $idx = scalar keys %{$irc{'glined'}};
      &putsock($socket, "-- End of List [%d G-line%s found] --", $idx, $idx > 1 ? "s" : "");
      return 0;

    } elsif (uc($switch) eq "B") {
      &putsock($socket, "-- Bot Lines --");
      &putsock($socket, "[Bot Nickname      ] [Hostname                      ] [Port ]");

      foreach (keys %bot) {
        &putsock($socket, "[%-2s %-15s] [%-30s] [%-5s] [ %-2s ]", $bot{$_}->{'prefix'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $bot{$_}->{'port'});
      }

      &putsock($socket, "-- End of List --");
      return 0;

    } else {
      &putsock($socket, $msg{'unkn-switch'}, $switch);
      return 0;
    }
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_raw {
  my ($socket, $raw) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    if (!$raw) {
      &putsock($socket, $msg{'syntax'}, "RAW <command>");
      return 0;
    }

    &dcc_record($socket, "RAW $raw");
    &putsock($socket, $msg{'perform'}, $raw);
    &putsock($sock{'irc'}, $raw);
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_die {
  my ($socket, $reason) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    &dcc_record($socket, "DIE $reason");
    if ($cfg{'irc-connected'}) { &putsock($sock{'irc'}, ":%s WALLOPS :%s: DIE %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason ? "[" . $reason . "]" : ""); }
    &terminate(1, $reason);
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_save {
  my $socket = shift;
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    &dcc_record($socket, "SAVE");
    &save_db();
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_restart {
  my $socket = shift;
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    if ($cfg{'irc-connected'}) { &putsock($sock{'irc'}, ":%s WALLOPS :%s: RESTART", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}); }
    &dcc_record($socket, "RESTART");
    &restart();
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_rehash {
  my $socket = shift;
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    &rehash();
    if ($cfg{'irc-connected'}) { &putsock($sock{'irc'}, ":%s WALLOPS :%s: REHASH", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}); }
    &putsock($socket, $msg{'rehash-done'});
    &dcc_record($socket, "REHASH");
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_gline {
  my ($socket, $arg) = (shift, shift);
  my $count = $arg =~ tr/ / /; ++$count;
  my $idx   = 0;
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    if (!$arg) {
      &putsock($socket, $msg{'syntax'}, "GLINE [time] <user\@host> [reason]");
      return 0;
    }

    my ($mask, $reason);
    my $expire = &format_timestr((split(/ /, $arg, 0))[0]);

    if (($count > 1) && ($expire =~ /\d+/) && ($expire > 0)) {
      $mask   = (split(/ /, $arg, 0))[1];
      $reason = (split(/ /, $arg, 3))[2];
    } else {
      $mask   = (split(/ /, $arg, 0))[0];
      $reason = (split(/ /, $arg, 2))[1];
      $expire = 0;
    }

    if ($mask !~ "\@") {
      $mask = "*\@" . $mask;
    } else {
      if ((split(/\@/, $mask, 0))[0] eq "") { $mask = "*" . $mask; }
    }

    foreach (split(//, $mask)) {
      if (($_ eq "\@") || ($_ eq "*") || ($_ eq "?") || ($_ eq ".")) { next; }
      ++$idx;
    }

    if ($idx < $cfg{'nonspecial-chars'}) {
      &putsock($socket, $msg{'gline-bad'}, $mask);
      return 0;
    }

    if (exists $irc{'glined'}->{lc($mask)}) {
      &putsock($socket, $msg{'gline-in'}, $mask);
      return 0;
    }

    my $usermask = (split(/\@/, $mask, 0))[0]; my $umaskrx = &plain2regexp($usermask);
    my $hostmask = (split(/\@/, $mask, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);

    &gline($usermask, $hostmask, $login, $reason, $ts, $expire, 1);

    if ($cfg{'irc-connected'}) {
      foreach (keys %{$irc{'client'}}) {
        if (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i)) {
          if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
          if ($irc{'client'}->{$_}->{'access'} && ($access{lc($irc{'client'}->{$_}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) { next; }

          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'nick'} = $irc{'client'}->{$_}->{'nick'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'time'} = $ts;
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'user'} = $irc{'client'}->{$_}->{'user'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'host'} = $irc{'client'}->{$_}->{'host'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'what'} = "QUIT";
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'xtra'} = $reason;

          if ($cfg{'server-type'} eq "hybrid") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "bahamut") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "ultimate") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          }

          ++$irc{'stats'}->{'kills'}[0];
          &del_client($irc{'client'}->{$_}->{'nick'});
        }
      }
    }

    if ($expire > 0) {
      &putsock($socket, $msg{'gline-tmp'}, &format_time($ts + $expire), $mask);
      &dcc_record($socket, "GLINE [$expire] $mask [$reason]");
    } else {
      &putsock($socket, $msg{'gline-add'}, $mask);
      &dcc_record($socket, "GLINE $mask [$reason]");
    }

    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_ungline {
  my ($socket, $mask) = (shift, shift);
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$mask) {
      &putsock($socket, $msg{'syntax'}, "UNGLINE <user\@host>");
      return 0;
    }

    $mask = (split(/ /, $mask, 0))[0];

    if (exists $irc{'glined'}->{lc($mask)}) {
      &dcc_record($socket, "UNGLINE $mask");

      &putsock($socket, $msg{'gline-del'}, $mask);

      if ($cfg{'irc-connected'}) {
        delete $irc{'glined'}->{lc($mask)};

        if ($cfg{'server-type'} eq "bahamut") {
          &putsock($sock{'irc'}, ":%s UNSGLINE 0 :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{lc($mask)}->{'mask'});
        } elsif ($cfg{'server-type'} eq "ultimate") {
          &putsock($sock{'irc'}, ":%s REMGLINE :%s", $cfg{'irc-fakeserv'}, $irc{'glined'}->{lc($mask)}->{'mask'});
        }

        &putsock($sock{'irc'}, ":%s WALLOPS :%s: UNGLINE %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $mask);
      } else {
        $irc{'glined'}->{lc($mask)}->{'expire'} = $ts;
      }
    } else {
      &putsock($socket, $msg{'no-mask'}, $mask);
    }

    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_kill {
  my ($socket, $arg) = (shift, shift);
  my $idx   = 0;
  my $kill  = 0;
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    if (!$arg) {
      &putsock($socket, $msg{'syntax'}, "KILL <nick|user\@host> [reason]");
      return 0;
    }

    my $target = (split(/ /, $arg, 0))[0];
    my $reason = (split(/ /, $arg, 2))[1];
    if (!$reason) { $reason = "Killed"; }

    if (exists $irc{'client'}->{lc($target)}) {
      if (lc($irc{'client'}->{lc($target)}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) {
        putsock($socket, "Cannot kill Network Service Daemon [%s]", $irc{'client'}->{lc($target)}->{'nick'});
        return 0;
      }

      if ($irc{'client'}->{lc($target)}->{'access'} && ($access{lc($irc{'client'}->{lc($target)}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) {
        &putsock($socket, $msg{'kill-friend'}, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'});
        &putlog("y", "%s attempted to KILL protected user %s!%s\@%s %s", $login, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'}, $reason ? "[" . $reason . "]" : "");
        return 0;
      }

      $irc{'seen'}->{lc($target)}->{'nick'} = $irc{'client'}->{lc($target)}->{'nick'};
      $irc{'seen'}->{lc($target)}->{'time'} = $ts;
      $irc{'seen'}->{lc($target)}->{'user'} = $irc{'client'}->{lc($target)}->{'user'};
      $irc{'seen'}->{lc($target)}->{'host'} = $irc{'client'}->{lc($target)}->{'host'};
      $irc{'seen'}->{lc($target)}->{'what'} = "QUIT";
      $irc{'seen'}->{lc($target)}->{'xtra'} = $reason;

      if ($cfg{'server-type'} eq "hybrid") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
      } elsif ($cfg{'server-type'} eq "bahamut") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
      } elsif ($cfg{'server-type'} eq "ultimate") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
      }

      &putsock($socket, $msg{'kill-nick'}, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'}, $reason);
      &dcc_record($socket, "KILL $irc{'client'}->{lc($target)}->{'nick'} [$reason]");

      ++$irc{'stats'}->{'kills'}[0];
      &del_client($irc{'client'}->{lc($target)}->{'nick'});
      return 0;
    } else {
      if ($target !~ "\@") {
        $target = "*\@" . $target;
      } else {
        if ((split(/\@/, $target, 0))[0] eq "") { $target = "*" . $target; }
      }

      foreach (split(//, $target)) {
        if (($_ eq "\@") || ($_ eq "*") || ($_ eq "?") || ($_ eq ".")) { next; }
        ++$idx;
      }

      if ($idx < $cfg{'nonspecial-chars'}) {
        &putsock($socket, $msg{'gline-bad'}, $target);
        return 0;
      }

      my $usermask = (split(/\@/, $target, 0))[0]; my $umaskrx = &plain2regexp($usermask);
      my $hostmask = (split(/\@/, $target, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);

      foreach (keys %{$irc{'client'}}) {
        if (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i)) {
          if ($kill >= $cfg{'max-kill'}) {
            &putsock($socket, $msg{'kill-max'});
            last;
          }

          if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
          if ($irc{'client'}->{$_}->{'access'} && ($access{lc($irc{'client'}->{$_}->{'access'})}->{'flags'} & $flags{'access'}->{'f'})) { next; }

          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'nick'} = $irc{'client'}->{$_}->{'nick'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'time'} = $ts;
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'user'} = $irc{'client'}->{$_}->{'user'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'host'} = $irc{'client'}->{$_}->{'host'};
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'what'} = "QUIT";
          $irc{'seen'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'xtra'} = $reason;

          if ($cfg{'server-type'} eq "hybrid") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s (%s\@%s))", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "bahamut") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          } elsif ($cfg{'server-type'} eq "ultimate") {
            &putsock($sock{'irc'}, ":%s KILL %s :%s (%s\@%s)", $services{'DataServ'}->{'nick'}, $irc{'client'}->{$_}->{'nick'}, $reason, $login, $services{'DataServ'}->{'nick'});
          }

          ++$irc{'stats'}->{'kills'}[0]; ++$kill;
          &del_client($irc{'client'}->{$_}->{'nick'});
        }
      }

      &putsock($socket, $msg{'kill-host'}, $kill, $target);
      &dcc_record($socket, "KILL $target [$reason]");
      return 0;
    }
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_omode {
  my ($socket, $omode) = (shift, shift);
  my $count = $omode =~ tr/ / /; ++$count;
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    if ($count < 2) {
      &putsock($socket, $msg{'syntax'}, "OMODE <channel> <modes>");
      return 0;
    }

    my $chan = (split(/ /, $omode, 0))[0];
    my $mode = (split(/ /, $omode, 0))[1];
    my $args = (split(/ /, $omode, 3))[2];
    my @arg; foreach (split(/ /, $args)) { push @arg, $_; }

    if (!exists $irc{'channel'}->{lc($chan)}) {
      &putsock($socket, $msg{'no-chan'}, $chan);
      return 0;
    } else {
      my $idx  = 0;
      my $what = 1;
      my ($exec, $target);

      foreach (split(//, $mode)) {
        if ($_ eq "+") {
          $what = 1;
          $exec .= "+";
        } elsif ($_ eq "-") {
          $what = 0;
          $exec .= "-";
        } elsif (($_ eq "o") || ($_ eq "h") || ($_ eq "v")) {
          if (exists $irc{'channel'}->{lc($chan)}->{'names'}->{lc($arg[$idx])}) {
            $exec   .= $_;
            $target .= $arg[$idx] . " ";
          }
          ++$idx;
        } elsif (($_ eq "b") || ($_ eq "d") || ($_ eq "e")) {
          $exec   .= $_;
          $target .= $arg[$idx] . " ";
        } elsif (($_ =~ /[fIkl]/) && !(($_ eq "l") && !$what)) {
          $exec   .= $_;
          $target .= $arg[$idx] . " ";
          ++$idx;
        } elsif ($_ =~ /[acilmnpst]/) {
          $exec .= $_;
        }
      }

      $target =~ s/\s+$//;
      &putsock($sock{'irc'}, ":%s MODE %s %s %s", $services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
      &channel_mode($services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
      &putsock($sock{'irc'}, ":%s WALLOPS :%s: OMODE %s %s %s", $cfg{'irc-fakeserv'}, $services{'DataServ'}->{'nick'}, $irc{'channel'}->{lc($chan)}->{'channel'}, $exec, $target);
      &dcc_record($socket, "OMODE $irc{'channel'}->{lc($chan)}->{'channel'} $exec $target");
      return 0;
    }
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_link {
  my ($socket, $nick) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$nick) {
      &putsock($socket, $msg{'syntax'}, "LINK <botnick>");
      return 0;
    }

    $nick = (split(/ /, $nick, 0))[0];

    foreach (keys %bot) {
      if (lc($bot{$_}->{'nick'}) eq lc($nick)) {
        if ($bot{$_}->{'sock'}) {
          &putsock($socket, $msg{'bot-linked'}, $bot{$_}->{'nick'});
          return 0;
        } else {
          &dcc_record($socket, "LINK $bot{$_}->{'nick'}");
          &putsock($socket, $msg{'bot-attempt'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $bot{$_}->{'port'});
          &connect2bot($_);
          return 0;
        }
      }
    }

    &putsock($socket, $msg{'no-bot'}, $nick);
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_unlink {
  my ($socket, $arg) = (shift, shift);
  my $login = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$arg) {
      &putsock($socket, $msg{'syntax'}, "UNLINK <botnick> [reason]");
      return 0;
    }

    my $nick   = (split(/ /, $arg, 0))[0];
    my $reason = (split(/ /, $arg, 2))[1];

    foreach (keys %bot) {
      if (lc($bot{$_}->{'nick'}) eq lc($nick)) {
        if ($bot{$_}->{'sock'}) {
          &dcc_record($socket, "UNLINK $bot{$_}->{'nick'} $reason");

          if (!$reason) {
            $reason = $login;
          } else {
            $reason = $login . ": " . $reason;
          }

          &putsock($bot{$_}->{'sock'}, "(%s) Unlinked from %s (%s)", $services{'DataServ'}->{'nick'}, $bot{$_}->{'nick'}, $reason);
          &putinfo("b", "Unlinked from %s (%s)", $bot{$_}->{'nick'}, $reason);
          &del_socket($bot{$_}->{'sock'});
          return 0;
        } else {
          &putsock($socket, $msg{'no-bot'}, $nick);
          return 0;
        }
      }
    }

    &putsock($socket, $msg{'no-bot'}, $nick);
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_clones {
  my $socket = shift;
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    my $idx = 0;
    my $cln = 0;
    my %clones;

    &putsock($socket, "-- Listing Clones --");

    foreach (keys %{$irc{'client'}}) {
      if ($irc{'client'}->{$_}->{'serv'} ne $cfg{'irc-fakeserv'}) { push @{$clones{lc($irc{'client'}->{$_}->{'host'})}}, $irc{'client'}->{$_}->{'nick'}; }
    }

    foreach (keys %clones) {
      if (scalar @{$clones{$_}} > 1) {
        ++$cln;
        &putsock($socket, "*\@%s", $_);
        foreach my $clone (@{$clones{$_}}) {
          &putsock($socket, "  %02d) %-15s [%s]", ++$idx, $irc{'client'}->{lc($clone)}->{'nick'}, $irc{'client'}->{lc($clone)}->{'serv'});
        }

        $idx = 0;
      }
    }

    &putsock($socket, "-- End of List [%d clone%s found] --", $cln, $cln > 1 ? "s" : "");
    &dcc_record($socket, "CLONES");
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_trace {
  my ($socket, $arg) = (shift, shift);
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    my ($oper, $nonoper, $long) = (0, 0, 0);
    my ($i, $j, $s) = (0, 0, 0); my $server;
    my $search = my $trace = (split(/ /, $arg, 0))[0];
    my $option = (split(/ /, $arg, 2))[1];

    if (!$search) {
      &putsock($socket, $msg{'syntax'}, "TRACE <pattern> [options]");
      return 0;
    }

    if ($option) {
      foreach (split(/ /, $option)) {
        ++$i;

        if (lc($_) eq "-oper") {
          $oper = 1;
        } elsif (lc($_) eq "-nonoper") {
          $nonoper = 1;
        } elsif (lc($_) eq "-long") {
          $long = 1;
        } elsif (lc($_) eq "-server") {
          $s = $i + 1;
        } elsif ($s == $i) {
          $server = $_;
        }
      }

      if ($s && !$server) {
        &putsock($socket, $msg{'syntax'}, "TRACE <pattern> [options]");
        return 0;
      } elsif ($server && !exists $irc{'server'}->{lc($server)}) {
        &putsock($socket, $msg{'no-server'}, $server);
        return 0;
      }
    }

    if ($trace !~ "\@") {
      $trace = "*\@" . $trace;
    } else {
      if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
    }

    my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
    my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
    my $searchrx = &plain2regexp($search);

    &putsock($socket, "-- Listing Results --");

    foreach (keys %{$irc{'client'}}) {
      if (($irc{'client'}->{$_}->{'nick'} =~ /$searchrx/i) || (($irc{'client'}->{$_}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{$_}->{'host'} =~ /$hmaskrx/i))) {
        if ($nonoper && ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'})) { next; }
        if ($oper && !($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'})) { next; }
        if ($server && (lc($irc{'client'}->{$_}->{'serv'}) ne lc($server))) { next; }
        ++$j;

        if ($long) {
          my $modes;
          my $chanlist;

          foreach my $flag (keys %{$flags{'client'}}) {
            if ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{$flag}) { $modes .= $flag; }
          }

          foreach my $chan (keys %{$irc{'channel'}}) {
            if (exists $irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}) {
              if ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'o'}) {
                $chanlist .= "\@" . $irc{'channel'}->{$chan}->{'channel'} . " ";
              } elsif ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'h'}) {
                $chanlist .= "\%" . $irc{'channel'}->{$chan}->{'channel'} . " ";
              } elsif ($irc{'channel'}->{$chan}->{'names'}->{lc($irc{'client'}->{$_}->{'nick'})}->{'mode'} & $flags{'channel'}->{'v'}) {
                $chanlist .= "+" . $irc{'channel'}->{$chan}->{'channel'} . " ";
              } else {
                $chanlist .= $irc{'channel'}->{$chan}->{'channel'} . " ";
              }
            }
          }

          $chanlist =~ tr/ //s; $chanlist =~ s/\s+$//;

          if ($j > 1) { &putsock($socket, " "); }
          &putsock($socket, "User:     %s (%s\@%s)", $irc{'client'}->{$_}->{'nick'}, $irc{'client'}->{$_}->{'user'}, $irc{'client'}->{$_}->{'host'});
          &putsock($socket, "Realname: %s", $irc{'client'}->{$_}->{'real'});
          &putsock($socket, "Server:   %s", $irc{'client'}->{$_}->{'serv'});
          &putsock($socket, "Usermode: +%s", $modes);
          &putsock($socket, "Signon:   %s", &format_time($irc{'client'}->{$_}->{'time'}));
          if ($chanlist) { &putsock($socket, "Channel:  %s", $chanlist); }
        } else {
          &putsock($socket, "%s!%s\@%s (%s) [%s]", $irc{'client'}->{$_}->{'nick'}, $irc{'client'}->{$_}->{'user'}, $irc{'client'}->{$_}->{'host'}, $irc{'client'}->{$_}->{'real'}, $irc{'client'}->{$_}->{'serv'});
        }
      }
    }

    &putsock($socket, "-- End of List [%d match%s found] --", $j, $j > 1 ? "es" : "");
    &dcc_record($socket, "TRACE " . $search . ($option ? " [" . $option . "]" : ""));
    return 0;

  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_notice {
  my ($socket, $arg) = (shift, shift);
  my $login  = $sock{'list'}->{$socket}->{'access'};

  if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
    if (!$cfg{'irc-connected'}) {
      &putsock($socket, $msg{'no-irc'});
      return 0;
    }

    my $target = (split(/ /, $arg, 0))[0]; $target = lc($target);
    my $notice = (split(/ /, $arg, 2))[1];
    my $idx    = 0;
    my $domode;

    if (!$target || !$notice) {
      &putsock($socket, $msg{'syntax'}, "NOTICE <target> <message>");
      return 0;
    }

    if ($target eq "-all") {
      foreach (keys %{$irc{'client'}}) {
        if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }

        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{$_}->{'nick'}, $notice);
        ++$idx;
      }

    } elsif ($target eq "-oper") {
      foreach (keys %{$irc{'client'}}) {
        if (lc($irc{'client'}->{$_}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
        if ($irc{'client'}->{$_}->{'mode'} & $flags{'client'}->{'o'}) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{$_}->{'nick'}, $notice);
          ++$idx;
        }
      }
    } elsif ($target =~ /^-chan:(\S+)$/) {
      if (!$1) {
        &putsock($socket, $msg{'syntax'}, "NOTICE <target> <message>");
        return 0;
      } elsif (!exists $irc{'channel'}->{lc($1)}) {
        &putsock($socket, $msg{'no-chan'}, $1);
        return 0;
      }

      foreach (keys %{$irc{'channel'}->{lc($1)}->{'names'}}) {
        my $user = $irc{'channel'}->{lc($1)}->{'names'}->{$_}->{'nick'};
        if (lc($irc{'client'}->{lc($user)}->{'serv'}) eq lc($cfg{'irc-fakeserv'})) { next; }
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $user, $notice);
        ++$idx;
      }

    } elsif (($target =~ /^-serv:(\S+)$/) || ($target =~ /^-server:(\S+)$/)) {
      if (!$1) {
        &putsock($socket, $msg{'syntax'}, "NOTICE <target> <message>");
        return 0;
      } elsif (!exists $irc{'server'}->{lc($1)}) {
        &putsock($socket, $msg{'no-serv'}, $1);
        return 0;
      }

      foreach (keys %{$irc{'client'}}) {
        if ((lc($irc{'client'}->{lc($_)}->{'serv'}) eq $1) && (lc($cfg{'irc-fakeserv'}) ne $1)) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'DataServ'}->{'nick'}, $cfg{'irc-comm'}, $irc{'client'}->{lc($_)}->{'nick'}, $notice);
          ++$idx;
        }
      }

    } else {
      &putsock($socket, $msg{'syntax'}, "NOTICE <target> <message>");
      return 0;
    }

    &putsock($socket, $msg{'msg-sent'}, $idx);
    &dcc_record($socket, "NOTICE $target $notice");
    return 0;
  } else {
    &putsock($socket, $msg{'no-priv'});
    return 0;
  }
}

sub dcc_record {
  my ($socket, $arg) = (shift, shift);

  &putinfo("s", "#%s# %s", $sock{'list'}->{$socket}->{'nick'}, $arg);
  return 0;
}

1;