# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: fileserv.pl

use strict;
use Cwd;
use File::Spec;

if ($] < 5.006) {
  use vars qw(%access %cfg %flags %fserv %fsquery %irc %msg %services %sock $bit $ts $VERSION);
} else {
  our (%access, %cfg, %flags, %fserv, %fsquery, %irc, %msg, %services, %sock, $bit, $ts, $VERSION);
}

my %status;
my %search;

sub process_fileserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'FileServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'FileServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FileServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'FileServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FileServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'FileServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FileServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "DCC") {
      if ($query =~ /DCC CHAT chat (\d+) (\d+)/i) {
        &putsock($sock{'irc'}, ":%s NOTICE %s :Please use: /msg %s START", $services{'FileServ'}->{'nick'}, $nick, $services{'FileServ'}->{'nick'});
      } elsif ($query =~ /DCC SEND (\S+) (\d+) (\d+) (\d+)/i) {
        if ($cfg{'fserv-upload'}) { &fileserv_get($nick, $1, $2, $3, $4); }
      } elsif ($query =~ /DCC RESUME (\S+) (\d+) (\d+)/i) {
        foreach (keys %{$sock{'list'}}) {
          if (($sock{'list'}->{$_}->{'type'} eq "FSERV-B") && ($sock{'list'}->{$_}->{'port'} == $2)) {
            $sock{'list'}->{$_}->{'seek'} = $3;
            &putsock($sock{'irc'}, ":%s NOTICE %s :\cADCC ACCEPT %s %d %d\cA", $services{'FileServ'}->{'nick'}, $nick, $1, $2, $3);
            last;
          }
        }
      }
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FileServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (uc($cmd) eq "HELP") {
    &fileserv_help($nick, $arg);
  } elsif (uc($cmd) eq "IGNORE") {
    &fileserv_ignore($nick, $arg);
  } elsif (uc($cmd) eq "WHO") {
    &fileserv_who($nick);
  } elsif (uc($cmd) eq "BOOT") {
    &fileserv_boot($nick, $arg);
  } elsif (uc($cmd) eq "START") {
    &fileserv_start($nick);
  } elsif (uc($cmd) eq "STATUS") {
    &fileserv_status($nick);
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
  }

  return 0;
}

sub fileserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'FileServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'FileServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/fileserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub fileserv_boot {
  my ($nick, $target) = (shift, shift);

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {

      if (!$target) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "BOOT <nick>");
        return 0;
      }

      foreach (keys %{$sock{'list'}}) {
        if (($sock{'list'}->{$_}->{'type'} eq "FSERV") && (lc($sock{'list'}->{$_}->{'nick'}) eq lc($target))) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'booted'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $target);
          &serv_record($nick, $services{'FileServ'}->{'nick'}, "BOOT $sock{'list'}->{$_}->{'nick'}");
          &del_socket($_);
          return 0;
        }
      }

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-user'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $target);
      return 0;

    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub fileserv_start {
  my $nick = shift;
  my $idx  = 0;
  my $listen;
  my $port;

  foreach (keys %{$irc{'nofserv'}}) {
    my $umaskrx = &plain2regexp($irc{'nofserv'}->{$_}->{'usermask'});
    my $hmaskrx = &plain2regexp($irc{'nofserv'}->{$_}->{'hostmask'});

    if (($irc{'client'}->{lc($nick)}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{lc($nick)}->{'host'} =~ /$hmaskrx/i)) {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
      return 0;
    }
  }

  foreach (keys %fserv) {
    if ($fserv{$_}->{'ondcc'}) {
      if (lc($fserv{$_}->{'nick'}) eq lc($nick)) {
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'one-dcc'});
        return 0;
      } elsif (++$idx > $cfg{'fserv-maxsend'}) {
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-max'});
        return 0;
      }
    }
  }

  foreach (keys %fsquery) {
    if (($fsquery{$_}->{'nick'} eq $nick) && ($ts - $fsquery{$_}->{'time'} < 30)) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'fserv-wait'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, abs($ts - $fsquery{$_}->{'time'} - 30));
      return 0;
    }
  }

  my @rng = split(/:/, $cfg{'fserv-listen'});
  my @ip  = split(/\./, $cfg{'my-ip'});

  foreach ($rng[0]..$rng[1]) {
    $port = $_;

    $listen = IO::Socket::INET->new(
      Proto     => "tcp",
      LocalPort => $port,
      Listen    => 1,
      Reuse     => 1
    );

    if ($listen) { last; }
  }

  if (!$listen) {
    &putlog("y", "Unable to listen port %d", $port);
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-failed'});
    return 0;
  }

  $fsquery{lc($nick)}->{'nick'} = $nick;
  $fsquery{lc($nick)}->{'time'} = $ts;

  &putsock($sock{'irc'}, ":%s %s %s :DCC Chat (%s)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cfg{'my-ip'});
  &putsock($sock{'irc'}, ":%s PRIVMSG %s :\cADCC CHAT chat %d %d\cA", $services{'FileServ'}->{'nick'}, $nick, (($ip[0] * 16777216) + ($ip[1] * 65536) + ($ip[2] * 256) + $ip[3]), $port);

  $sock{'list'}->{$listen}->{'sock'}   = $listen;
  $sock{'list'}->{$listen}->{'type'}   = "FSERV-A";
  $sock{'list'}->{$listen}->{'nick'}   = $nick;
  $sock{'list'}->{$listen}->{'user'}   = $irc{'client'}->{lc($nick)}->{'user'};
  $sock{'list'}->{$listen}->{'host'}   = $irc{'client'}->{lc($nick)}->{'host'};
  $sock{'list'}->{$listen}->{'port'}   = $port;
  $sock{'list'}->{$listen}->{'fileno'} = $listen->fileno();
  $sock{'list'}->{$listen}->{'time'}   = $ts;
  vec($bit, $sock{'list'}->{$listen}->{'fileno'}, 1) = 1;

  &serv_record($nick, $services{'FileServ'}->{'nick'}, "START");
  return 0;
}

sub fileserv_who {
  my $nick = shift;
  my $idx  = 0;
  my $dcc;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      &putsock($sock{'irc'}, ":%s %s %s :-- Listing Users --", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      &putsock($sock{'irc'}, ":%s %s %s :[Nickname       ] [IP             ] [D/L] [Traffic   ] [DCC]", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
      foreach (keys %fserv) {
        ++$idx;
        if ($fserv{$_}->{'ondcc'}) { $dcc = " X "; }
        &putsock($sock{'irc'}, ":%s %s %s :[%-15s] [%-15s] [%-3s] [%-10s] [%-3s]", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $fserv{$_}->{'nick'}, $fserv{$_}->{'ip'}, $fserv{$_}->{'total'}, &format_size($fserv{$_}->{'traffic'}), $dcc);
        undef($dcc);
      }

      &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d user%s found] --", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
      &serv_record($nick, $services{'FileServ'}->{'nick'}, "WHO");
      return 0;

    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub fileserv_status {
  my $nick    = shift;
  my $ondcc   = 0;
  my $xfer    = 0;
  my $cps     = 0;

  $status{lc($nick)}->{'files'} = 0;
  $status{lc($nick)}->{'size'}  = 0;
  &fileserv_dumpdir($nick, $cfg{'fserv-root'});

  foreach (keys %fserv) {
    if ($fserv{$_}->{'ondcc'}) { ++$ondcc; }
    $xfer += $fserv{$_}->{'xfer'};
  }

  foreach (keys %{$sock{'list'}}) {
    if (($sock{'list'}->{$_}->{'type'} eq "FSERV-C") || ($sock{'list'}->{$_}->{'type'} eq "FSERV-X")) {
      if ($ts - $sock{'list'}->{$_}->{'start'} > 0) {
        $cps += $sock{'list'}->{$_}->{'seek'} / ($ts - $sock{'list'}->{$_}->{'start'});
      } else {
        $cps += $sock{'list'}->{$_}->{'seek'};
      }
    }
  }

  &putsock($sock{'irc'}, ":%s %s %s :-- Status Monitor --", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
  &putsock($sock{'irc'}, ":%s %s %s :Currently on DCC Chat: %d user%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $ondcc, $ondcc > 1 ? "s" : "");
  &putsock($sock{'irc'}, ":%s %s %s :Current Downloads:     %d", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $xfer);
  &putsock($sock{'irc'}, ":%s %s %s :Current Speed:         %s/sec", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, &format_size($cps));
  &putsock($sock{'irc'}, ":%s %s %s :Free Slots:            %d/%d", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, ($cfg{'fserv-maxsend'} - $ondcc), $cfg{'fserv-maxsend'});
  &putsock($sock{'irc'}, ":%s %s %s :Files Served:          %d (traffic: %s)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $services{'FileServ'}->{'total'}, &format_size($services{'FileServ'}->{'traffic'}));
  &putsock($sock{'irc'}, ":%s %s %s :Total Files:           %d (%s)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $status{lc($nick)}->{'files'}, &format_size($status{lc($nick)}->{'size'}));

  delete $status{lc($nick)};
  &serv_record($nick, $services{'FileServ'}->{'nick'}, "STATUS");
  return 0;
}

sub fileserv_dumpdir {
  my ($nick, $folder) = (shift, shift);

  if (!opendir(DIR, $folder)) {
    &putlog("y", "Unable to read folder: %s", $folder);
    return 0;
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $folder . "/" . $_;
    if (-d $absolute) {
      &fileserv_dumpdir($nick, $absolute);
    } else {
      ++$status{lc($nick)}->{'files'};
      $status{lc($nick)}->{'size'} += -s $absolute;
      push @{$status{lc($nick)}->{'list'}}, $absolute;
    }
  }

  return 0;
}

sub fserv_parse {
  my ($socket, $buffer) = (shift, shift);

  if (&fserv_flood($socket)) { return 0; }

  my $cmd = (split(/ /, $buffer, 0))[0];
  my $arg = (split(/ /, $buffer, 2))[1];

  if (uc($cmd) eq "HELP") {
    &fserv_help($socket);
  } elsif ((uc($cmd) eq "DIR") || (uc($cmd) eq "LS")) {
    &fserv_dir($socket);
  } elsif ((uc($cmd) eq "CD") || (uc($cmd) eq "CD..")) {
    if (uc($cmd) eq "CD..") { $arg = ".."; }
    &fserv_cd($socket, $arg);
  } elsif (uc($cmd) eq "GET") {
    &fserv_get($socket, $arg);
  } elsif (uc($cmd) eq "QUEUE") {
    &fserv_queue($socket);
  } elsif (uc($cmd) eq "FIND") {
    &fserv_find($socket, $arg);
  } elsif ((uc($cmd) eq "QUIT") || (uc($cmd) eq "EXIT") || (uc($cmd) eq "BYE")) {
    &fserv_quit($socket);
  } else {
    &putsock($socket, $msg{'unkn-cmd'}, $cmd);
  }

  return 0;
}

sub fserv_help {
  my $socket = shift;
  my $ip     = $sock{'list'}->{$socket}->{'ip'};
  my $nick   = $fserv{$ip}->{'nick'};

  &putsock($socket, "CD <folder>    Change directory");
  &putsock($socket, "DIR            List files");
  &putsock($socket, "FIND           Find files matching pattern");
  &putsock($socket, "GET <file>     Download a file");
  &putsock($socket, "QUEUE          List your queues");
  &putsock($socket, "QUIT           Quit the DCC Chat session");

  &serv_record($nick, $services{'FileServ'}->{'nick'}, "#HELP#");
  return 0;
}

sub fserv_queue {
  my $socket = shift;
  my $ip     = $sock{'list'}->{$socket}->{'ip'};
  my $idx    = 0;

  if (exists $fserv{$ip}->{'queue'} && (scalar @{$fserv{$ip}->{'queue'}} > 0)) {
    &putsock($socket, "-- Listing Files --");
    foreach (@{$fserv{$ip}->{'queue'}}) {
      my ($vol, $dir, $file) = File::Spec->splitpath($_);
      my $size = -s $_;
      &putsock($socket, "%02d) %-32s %s", ++$idx, $file, &format_size($size));
    }

    &putsock($socket, "-- End of List [%d file%s queued] --", $idx, $idx > 1 ? "s" : "");
  } else {
    &putsock($socket, "Your queue is empty");
  }

  &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#QUEUE#");
  return 0;
}

sub fserv_find {
  my ($socket, $pattern) = (shift, shift);
  my $ip = $sock{'list'}->{$socket}->{'ip'};
  my $folder;

  if (!$pattern) {
    &putsock($socket, $msg{'syntax'}, "FIND <pattern>");
    return 0;
  }

  $search{$socket}->{'files'} = 0;
  $search{$socket}->{'size'}  = 0;
  $search{$socket}->{'print'} = 0;

  &putsock($socket, "-- Listing files matching [%s] --", $pattern);
  &fserv_finddir($socket, $pattern, $sock{'list'}->{$socket}->{'folder'});
  if ($search{$socket}->{'print'} > $cfg{'fserv-maxwild'}) { &putsock($socket, "(%d other matches not displayed)", $search{$socket}->{'print'} - $cfg{'fserv-maxwild'}); }
  &putsock($socket, "-- End of List [%d file%s, %s] --", $search{$socket}->{'files'}, $search{$socket}->{'files'} > 1 ? "s" : "", &format_size($search{$socket}->{'size'}));
  delete $search{$socket};

  &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#FIND $pattern#");
  return 0;
}

sub fserv_finddir {
  my ($socket, $pattern, $folder) = (shift, shift, shift);
  my $patternrx = &plain2regexp($pattern);
  my $dir;

  if ($folder eq "/") {
    $dir = $cfg{'fserv-root'};
  } else {
    $dir = $cfg{'fserv-root'} . $folder;
  }

  if (!opendir(DIR, $dir)) {
    &putlog("y", "Unable to read folder: %s", $dir);
    return 0;
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $dir . "/" . $_;
    if (-d $absolute) {
      &fserv_finddir($socket, $pattern, $folder eq "/" ? "/" . $_ : $folder . "/" . $_);
    } elsif ((-e $absolute) && ($_ =~ /$patternrx/i)) {
      if (++$search{$socket}->{'print'} < $cfg{'fserv-maxwild'}) { &putsock($socket, "%-32s %s", $folder eq "/" ? $_ : $folder . "/" . $_, &format_size(-s $absolute)); }
      ++$search{$socket}->{'files'};
      $search{$socket}->{'size'} += -s $absolute;
    }
  }

  return 0;
}

sub fserv_cd {
  my ($socket, $folder) = (shift, shift);
  my $ip = $sock{'list'}->{$socket}->{'ip'};
  my $absolute;

  if (!$folder || ($folder eq ".")) {
    &putsock($socket, "Current folder: [%s]", $sock{'list'}->{$socket}->{'folder'});
    return 0;
  }

  if ($folder =~ /^\//) {
    if ($folder eq "/") {
      $absolute = $cfg{'fserv-root'};
    } else {
      $absolute = $cfg{'fserv-root'} . $folder;
    }
  } else {
    $absolute = $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'} . "/". $folder;
  }

  $absolute = File::Spec->canonpath($absolute);
  $absolute = Cwd::abs_path($absolute);

  if ($absolute eq "/") { $absolute = $cfg{'fserv-root'}; }

  if ($absolute !~ /^$cfg{'fserv-root'}/) {
    &putsock($socket, "Invalid directory [%s]", $folder);
    return 0;
  } elsif ((-d $absolute) && ($absolute !~ /\/\./)) {
    $folder = substr($absolute, length($cfg{'fserv-root'}));
    if (!$folder) { $folder = "/"; }
    &putsock($socket, "Entering directory: [%s]", $folder);
    $sock{'list'}->{$socket}->{'folder'} = $folder;

    if (-e $absolute . "/" . $cfg{'fserv-dirinfo'}) {
      &putsock($socket, "-------------------------------------");
      &send_motd($socket, $absolute . "/" . $cfg{'fserv-dirinfo'});
      &putsock($socket, "-------------------------------------");
    }

    &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#CD $folder#");
    return 0;
  } else {
    &putsock($socket, "Invalid directory [%s]", $folder);
    return 0;
  }
}

sub fserv_dir {
  my $socket  = shift;
  my $ip      = $sock{'list'}->{$socket}->{'ip'};
  my $tfiles  = 0;
  my $tsize   = 0;
  my @files   = ();
  my @folders = ();

  if (!opendir(DIR, $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'})) {
    &putlog("y", "Unable to read folder: %s", $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'});
    return 0;
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'} . $_;
    if (-d $absolute) {
      push @folders, $_;
    } else {
      push @files, $_;
    }
  }

  @folders = sort @folders;
  @files   = sort @files;

  if (!scalar @folders && !scalar @files) {
    &putsock($socket, "Directory is empty");
    return 0;
  }

  &putsock($socket, "-- Listing [%s] --", $sock{'list'}->{$socket}->{'folder'});
  foreach (@folders) { &putsock($socket, "<%s>", $_); }
  foreach (@files) {
    my $absolute = $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'} . "/" . $_; $absolute = File::Spec->canonpath($absolute);
    my $size = -s $absolute;
    ++$tfiles;
    $tsize += $size;
    &putsock($socket, "%-50s [%-5s]", $_, &format_size($size));
  }

  &putsock($socket, "-- End of List [%d file%s, %s] --", $tfiles, $tfiles > 1 ? "s" : "", &format_size($tsize));
  &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#DIR $sock{'list'}->{$socket}->{'folder'}#");
  return 0;
}

sub fserv_get {
  my ($socket, $filename) = (shift, shift);
  my $ip   = $sock{'list'}->{$socket}->{'ip'};
  my $absolute;

  if (!$filename) {
    &putsock($socket, $msg{'syntax'}, "GET <file>");
    return 0;
  }

  if ($filename =~ /^\//) {
    $absolute = $cfg{'fserv-root'} . $filename;
  } else {
    $absolute = $cfg{'fserv-root'} . $sock{'list'}->{$socket}->{'folder'} . "/" . $filename;
  }

  my ($vol, $dir, $file) = File::Spec->splitpath($absolute);
  $absolute  = File::Spec->canonpath($dir);
  $absolute  = Cwd::abs_path($absolute);
  $absolute .= "/" . $file;

  if ($absolute !~ /^$cfg{'fserv-root'}/) {
    &putsock($socket, "No such file [%s]", $filename);
    return 0;
  } elsif ((!-d $absolute) && (-e $absolute) && ($absolute !~ /\/\./)) {
    &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#GET $file#");
    my $size = -s $absolute;

    if ($fserv{$ip}->{'xfer'} >= $cfg{'fserv-send'}) {
      if (exists $fserv{$ip}->{'queue'} && (scalar @{$fserv{$ip}->{'queue'}} >= $cfg{'fserv-queue'})) {
        &putsock($socket, "Your queue is full, try later");
        return 0;
      } else {
        foreach (@{$fserv{$ip}->{'queue'}}) {
          if ($_ eq $absolute) {
            &putsock($socket, "This file is already queued");
            return 0;
          }
        }

        push @{$fserv{$ip}->{'queue'}}, $absolute;
        &putsock($socket, "File queued: %s (%s)", $file, &format_size($size));
        return 0;
      }
    } else {
      &fileserv_send($ip, $absolute);
      return 0;
    }

  } else {
    &putsock($socket, "No such file [%s]", $filename);
    return 0;
  }
}

sub fserv_quit {
  my $socket = shift;
  my $ip     = $sock{'list'}->{$socket}->{'ip'};

  &putsock($socket, "OK, bye!");
  &serv_record($fserv{$ip}->{'nick'}, $services{'FileServ'}->{'nick'}, "#QUIT#");
  &del_socket($socket);
  return 0;
}

sub fileserv_send {
  my ($ip, $absolute) = (shift, shift);

  my $size = -s $absolute;
  my ($vol, $dir, $file) = File::Spec->splitpath($absolute);

  my @rng  = split(/:/, $cfg{'fserv-listen'});
  my $port = $rng[1] - int(rand($rng[1] - $rng[0]));
  my @ipx  = split(/\./, $cfg{'my-ip'});

  my $listen = IO::Socket::INET->new(
    Proto     => "tcp",
    LocalPort => $port,
    Listen    => 1,
    Reuse     => 1
  );

  if (!$listen) {
    &putlog("y", "Unable to listen port %d", $port);
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $fserv{$ip}->{'nick'}, $cfg{'irc-comm'}, $msg{'dcc-failed'});
    &fileserv_queued($ip);
    return 0;
  }

  if ($fserv{$ip}->{'ondcc'}) {
    &putsock($fserv{$ip}->{'ondcc'}, "Now sending you: %s (%s)", $file, &format_size(-s $absolute));
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :Now sending you: %s (%s)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $fserv{$ip}->{'nick'}, $file, &format_size(-s $absolute));
  }

  ++$fserv{$ip}->{'xfer'};

  &putsock($sock{'irc'}, ":%s %s %s :DCC Send %s (%s)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $fserv{$ip}->{'nick'}, $file, $cfg{'my-ip'});
  $file =~ s/ /_/g;
  &putsock($sock{'irc'}, ":%s PRIVMSG %s :\cADCC SEND %s %d %d %d\cA", $services{'FileServ'}->{'nick'}, $fserv{$ip}->{'nick'}, $file, (($ipx[0] * 16777216) + ($ipx[1] * 65536) + ($ipx[2] * 256) + $ipx[3]), $port, $size);

  $sock{'list'}->{$listen}->{'sock'}     = $listen;
  $sock{'list'}->{$listen}->{'type'}     = "FSERV-B";
  $sock{'list'}->{$listen}->{'ip'}       = $ip;
  $sock{'list'}->{$listen}->{'port'}     = $port;
  $sock{'list'}->{$listen}->{'fileno'}   = $listen->fileno();
  $sock{'list'}->{$listen}->{'time'}     = $ts;
  $sock{'list'}->{$listen}->{'absolute'} = $absolute;
  $sock{'list'}->{$listen}->{'seek'}     = 0;
  vec($bit, $sock{'list'}->{$listen}->{'fileno'}, 1) = 1;

  return 0;
}

sub fileserv_ignore {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;
  my $mask;
  my $what;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "IGNORE <ADD|DEL|LIST> [mask]");
        return 0;
      }

      $what = (split(/ /, $arg, 0))[0];
      $mask = (split(/ /, $arg, 0))[1];

      if ($mask !~ "\@") {
        $mask = "*\@" . $mask;
      } else {
        if ((split(/\@/, $mask, 0))[0] eq "") { $mask = "*" . $mask; }
      }

      if (uc($what) eq "ADD") {
        if (!$mask) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "IGNORE <ADD|DEL|LIST> [mask]");
          return 0;
        } else {
          if (!exists $irc{'nofserv'}->{lc($mask)}) {
            &serv_record($nick, $services{'FileServ'}->{'nick'}, "IGNORE ADD $mask");
            $irc{'nofserv'}->{lc($mask)}->{'hostmask'} = lc((split(/\@/, $mask, 0))[0]);
            $irc{'nofserv'}->{lc($mask)}->{'usermask'} = lc((split(/\@/, $mask, 0))[1]);
            $irc{'nofserv'}->{lc($mask)}->{'new'}  = 1;
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'ignore-add'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'ignore-in'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          }
        }
      } elsif (uc($what) eq "DEL") {
        if (!$mask) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "IGNORE <ADD|DEL|LIST> [mask]");
          return 0;
        } else {
          if (exists $irc{'nofserv'}->{lc($mask)}) {
            &serv_record($nick, $services{'FileServ'}->{'nick'}, "IGNORE DEL $mask");
            delete $irc{'nofserv'}->{lc($mask)};
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'ignore-del'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-mask'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          }
        }
      } elsif (uc($what) eq "LIST") {
        &serv_record($nick, $services{'FileServ'}->{'nick'}, "IGNORE LIST");
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing Ignores --", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        foreach (keys %{$irc{'nofserv'}}) { &putsock($sock{'irc'}, ":%s %s %s :  %02d) %s\@%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, ++$idx, $irc{'nofserv'}->{$_}->{'usermask'}, $irc{'nofserv'}->{$_}->{'hostmask'}); }
        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d mask%s found] --", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "IGNORE <ADD|DEL|LIST> [mask]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub fileserv_queued {
  my $ip = shift;

  if (exists $fserv{$ip}->{'queue'} && scalar @{$fserv{$ip}->{'queue'}}) {
    my $absolute = shift @{$fserv{$ip}->{'queue'}};
    &fileserv_send($ip, $absolute);
    return 0;
  } else {
    if (!$fserv{$ip}->{'ondcc'} && !$fserv{$ip}->{'xfer'}) { delete $fserv{$ip}; }
    return 0;
  }
}

sub fserv_send {
  my ($socket, $seek) = (shift, shift);
  my $ip = $sock{'list'}->{$socket}->{'ip'};
  my $buffer;

  if (!open(FILE, "<" . $sock{'list'}->{$socket}->{'absolute'})) {
    &putlog("y", "Unable to read file: %s", $sock{'list'}->{$socket}->{'absolute'});
    &del_socket($socket);
    return 0;
  }

  $sock{'list'}->{$socket}->{'time'} = $ts;

  binmode FILE;
  seek(FILE, $seek, 0);
  read(FILE, $buffer, $cfg{'fserv-buffer'});

  if (!$buffer) {
    close(FILE);

    my ($vol, $dir, $file) = File::Spec->splitpath($sock{'list'}->{$socket}->{'absolute'});
    my $size = -s $sock{'list'}->{$socket}->{'absolute'};
    my $cps  = $size;

    if ($ts - $sock{'list'}->{$socket}->{'start'} > 0) { $cps = $size / ($ts - $sock{'list'}->{$socket}->{'start'}); }
    $sock{'list'}->{$socket}->{'type'} = "FSERV-D";

    --$fserv{$ip}->{'xfer'};
    ++$fserv{$ip}->{'total'};
    ++$services{'FileServ'}->{'total'};

    if ($fserv{$ip}->{'ondcc'}) {
      &putsock($fserv{$ip}->{'ondcc'}, "DCC Send of %s (%s) complete (%s, %s/sec)", $file, &format_size($size), &format_time($sock{'list'}->{$socket}->{'start'}), &format_size($cps));
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :DCC Send of %s (%s) complete (%s, %s/sec)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $fserv{$ip}->{'nick'}, $file, &format_size($size), &format_time($sock{'list'}->{$socket}->{'start'}), &format_size($cps));
    }

    &del_socket($socket);
    &fileserv_queued($ip);
    return 0;
  }

  $fserv{$ip}->{'traffic'} += $cfg{'fserv-buffer'};
  $services{'FileServ'}->{'traffic'} += $cfg{'fserv-buffer'};
  $sock{'list'}->{$socket}->{'seek'} += $cfg{'fserv-buffer'};
  print $socket $buffer;
  close(FILE);

  if ($cfg{'fserv-maxcps'} > 0) { $fserv{$ip}->{'limitcps'} += $cfg{'fserv-buffer'}; }
  return 0;
}

sub fileserv_get {
  my ($nick, $filename, $addr, $port, $size) = (shift, shift, shift, shift, shift);
  my $idx = 0;

  foreach (keys %{$sock{'list'}}) {
    if (($sock{'list'}->{$_}->{'type'} eq "FSERV-G") && (++$idx > $cfg{'fserv-maxget'})) {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-max'});
      return 0;
    }
  }

  if ($cfg{'fserv-maxup'} && ($size > $cfg{'fserv-maxup'})) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'fserv-toobig'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $filename, &format_size($size));
    return 0;
  }

  if (!opendir(DIR, $cfg{'fserv-upload'})) {
    &putlog("y", "Unable to open folder: %s", $cfg{'fserv-upload'});
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-failed'});
    return 0;
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ eq $filename) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'fserv-exists'}", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $filename);
      return 0;
    }
  }

  my $FILE;
  if (!open($FILE, ">" . $cfg{'fserv-upload'} . "/" . $filename)) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-failed'});
    return 0;
  }

  binmode $FILE;

  my $ip = inet_ntoa(pack("N", $addr));
  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => $ip,
    PeerPort => $port,
    Timeout  => 3
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'dcc-failed'});
    return 0;
  }

  $socket->autoflush(1);

  $sock{'list'}->{$socket}->{'sock'}   = $socket;
  $sock{'list'}->{$socket}->{'type'}   = "FSERV-G";
  $sock{'list'}->{$socket}->{'start'}  = $ts;
  $sock{'list'}->{$socket}->{'time'}   = $ts;
  $sock{'list'}->{$socket}->{'nick'}   = $nick;
  $sock{'list'}->{$socket}->{'user'}   = $irc{'client'}->{lc($nick)}->{'user'};
  $sock{'list'}->{$socket}->{'file'}   = $filename;
  $sock{'list'}->{$socket}->{'size'}   = $size;
  $sock{'list'}->{$socket}->{'read'}   = 0;
  $sock{'list'}->{$socket}->{'host'}   = $irc{'client'}->{lc($nick)}->{'host'};
  $sock{'list'}->{$socket}->{'port'}   = $socket->peerport();
  $sock{'list'}->{$socket}->{'filehd'} = $FILE;
  $sock{'list'}->{$socket}->{'fileno'} = $socket->fileno();
  vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 1;

  $size = &format_size($size);
  &serv_record($nick, $services{'FileServ'}->{'nick'}, "DCC Send $filename ($size) [starting...]");
  return 0;
}

1;