# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: findserv.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %cfg %flags %irc %msg %services %sock $VERSION);
} else {
  our (%access, %cfg, %flags, %irc, %msg, %services, %sock, $VERSION);
}

sub process_findserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'FindServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'FindServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FindServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'FindServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FindServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'FindServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FindServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'FindServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (my $pid = fork) {
    if (uc($cmd) eq "HELP") {
      &findserv_help($nick, $arg);
    } elsif (uc($cmd) eq "ALLTHEWEB") {
      &findserv_alltheweb($nick, $arg);
    } elsif (uc($cmd) eq "ALTAVISTA") {
      &findserv_altavista($nick, $arg);
    } elsif (uc($cmd) eq "CNET") {
      &findserv_cnet($nick, $arg);
    } elsif (uc($cmd) eq "FRESHMEAT") {
      &findserv_freshmeat($nick, $arg);
    } elsif (uc($cmd) eq "GOOGLE") {
      &findserv_google($nick, $arg);
    } elsif (uc($cmd) eq "LINUXDOC") {
      &findserv_linuxdoc($nick, $arg);
    } elsif (uc($cmd) eq "LYCOS") {
      &findserv_lycos($nick, $arg);
    } elsif (uc($cmd) eq "SLASHDOT") {
      &findserv_slashdot($nick, $arg);
    } elsif (uc($cmd) eq "YAHOO") {
      &findserv_yahoo($nick, $arg);
    } elsif (uc($cmd) eq "WHOIS") {
      &findserv_whois($nick, $arg);
    } elsif ((uc($cmd) eq "DNS") || (uc($cmd) eq "NSLOOKUP")) {
      &findserv_nslookup($nick, $arg);
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
    }
    exit(0);
  }

  return 0;
}

sub findserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'FindServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'FindServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/findserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub findserv_alltheweb {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ALLTHEWEB <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.alltheweb.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/search?cat=web&cs=utf-8&l=any&q=" . $query, "www.alltheweb.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "AllTheWeb", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (!$idx && ($buffer =~ /<p>No Web pages found that match your query.<\/p>/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($desc) = $buffer =~ /<a onmouseover="self.status='\S+';return true" onmouseout="self.status='';return true" href="\S+"><b>(.*)<\/b><\/a><br>$/; $desc =~ s/&lt;/</g; $desc =~ s/&gt;/>/g; }
          if (!$link) { ($link) = $buffer =~ /<a onmouseover="self.status='(\S+)';return true"/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "ALLTHEWEB $keyword");
  return 0;
}

sub findserv_altavista {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ALTAVISTA <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.altavista.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/web/results?q=" . $query . "&kgs=0&kls=0&avkw=qtrp", "www.altavista.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "AltaVista", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s; $buffer =~ s/<b>//g; $buffer =~ s/<\/b>//g;

        if (!$idx && ($buffer =~ /<span class=m><span class=y>We found 0 results. Suggestions:<\/span><\/span><br>/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($desc) = $buffer =~ /href="\S+" onMouseOver="status='\S+'; return true;" onMouseOut="status=''; return true;">(.*)<\/a>/; }
          if (!$link) { ($link) = $buffer =~ /href="\S+" onMouseOver="status='(\S+)'; return true;" onMouseOut="status=''; return true;">.*<\/a>/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "ALTAVISTA $keyword");
  return 0;
}

sub findserv_cnet {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "CNET <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.search.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/search?channel=1&tag=st.se.fd..sch&q=" . $query, "www.search.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "C|net", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (!$idx && ($buffer =~ /^<ul>There were no results in your initial search of Web Pages.<br>/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($desc) = $buffer =~ /^<li><a href="\S+">(.*)<\/a> &#150;/; }
          if (!$link) { ($link) = $buffer =~ /^<li><a href="\S+">.*<\/a> &#150; .*<br><font size="1">(\S+)<\/font>&nbsp;&nbsp;/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "CNET $keyword");
  return 0;
}

sub findserv_freshmeat {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "FRESHMEAT <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($proj, $desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "freshmeat.net",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/search-xml/?q=" . $query, "freshmeat.net", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "FreshMeat", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (!$idx && ($buffer =~ /<\/search-results>/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$proj) { ($proj) = $buffer =~ /<projectname_full>(.*)<\/projectname_full>/; }
          if (!$desc) { ($desc) = $buffer =~ /<desc_short>(.*)<\/desc_short>/; }
          if (!$link) { ($link) = $buffer =~ /<url_homepage>(\S+)<\/url_homepage>/; }
          if ($proj && $desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s: %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $proj, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($proj); undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "FRESHMEAT $keyword");
  return 0;
}

sub findserv_google {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "GOOGLE [options] <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link, $x);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.google.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
   );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/search?hl=en&ie=ISO-8859-1&q=" . $query, "www.google.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "Google", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s; $buffer =~ s/<b>//g; $buffer =~ s/<\/b>//g;

        if (!$idx && ($buffer =~ /^<br><br>Your search - $keyword - did not match any documents./)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($x, $desc) = $buffer =~ /(<\/blockquote>|<div>|<\/a><\/font> )<p class=g><a href=\S+>(.*?)<\/a>(<br>)?<font size=-1>([^<]| \- \[ | \.\.\.|<i>|<span class=f>)/; $desc =~ s/&amp;/&/g; $desc =~ s/&quot;/"/g; }
          if (!$link) { ($x, $link) = $buffer =~ /(<\/blockquote>|<div>|<\/a><\/font> )<p class=g><a href=(\S+)>(.*?)<\/a>(<br>)?<font size=-1>([^<]| \- \[ | \.\.\.|<i>|<span class=f>)/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "GOOGLE $keyword");
  return 0;
}

sub findserv_linuxdoc {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "LINUXDOC <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "tldp.org",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      my $post;
      my $words = $query; $words =~ tr/ /+/;

      $post .= "svr=";
      $post .= "\&rpt=20";
      $post .= "\&words=" . $words;
      $post .= "\&submit=go";

      $socket->autoflush(1);
      &http_post($socket, "/cgi-bin/ldpsrch.cgi", "tldp.org", undef, $post, undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "LinuxDoc", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (!$idx && ($buffer =~ /^<h4>No Entries Found...<\/h4><\/p><\/td><\/tr><\/table>$/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($desc) = $buffer =~ /^<td align=left valign=top><font face="Arial,Helvetica,sans-serif" size="-1"><a href="\S+"><b>(.*)<\/b><\/a>$/; }
          if (!$link) { ($link) = $buffer =~ /^<td align=left valign=top><font face="Arial,Helvetica,sans-serif" size="-1"><a href="(\S+)"><b>.*<\/b><\/a>$/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "LINUXDOC $keyword");
  return 0;
}

sub findserv_lycos {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "LYCOS <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "search.lycos.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/default.asp?loc=searchbox&tab=web&query=" . $query, "www.lycos.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "Lycos", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s; $buffer =~ s/<b>//g; $buffer =~ s/<\/b>//g;

        if (!$idx && ($buffer =~ /^<tr><td><a name="dir"><font face=verdana size=-1>There are no WEB results for this search term<\/font><\/a><\/td><\/tr>$/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$desc) { ($desc) = $buffer =~ /^<font face="verdana" size="-1"><a href="\S+;rsource=LCOSWF">(.*)<\/a><script language="JavaScript">/; }
          if (!$link) { ($link) = $buffer =~ /<font face="verdana" size="-2" color="#999999">(\S+)<\/font>$/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "LYCOS $keyword");
  return 0;
}

sub findserv_slashdot {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SLASHDOT <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($arti, $date, $link);
  my $idx   = 0;
  my $nodoc = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "slashdot.org",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/search.pl?query=" . $query, "slashdot.org", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SlashDot", $keyword);
      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (!$idx && ($buffer =~ /<\/FORM>No stories were found that match your query./)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if (!$arti) { ($arti) = $buffer =~ /<A HREF="\S+&amp;mode=thread&amp;tid=\d+">(.*)<\/A><\/B><BR>/; }
          if (!$date) { ($date) = $buffer =~ /<FONT SIZE="-1">On (.*)<\/FONT><BR>/; }
          if (!$link) { ($link) = $buffer =~ /<A HREF="(\S+)&amp;mode=thread&amp;tid=\d+">.*<\/A><\/B><BR>/; }
          if ($arti && $date && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            if ($link =~ /^\/\//) { $link = "http:" . $link; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s (%s)", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $arti, $date);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($arti); undef($date); undef($link);
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "SLASHDOT $keyword");
  return 0;
}

sub findserv_yahoo {
  my ($nick, $keyword) = (shift, shift);

  if (!$keyword) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YAHOO <keywords>");
    return 0;
  }

  my $query = $keyword; $query =~ tr/ /+/;
  my ($desc, $link);
  my $idx   = 0;
  my $nodoc = 0;
  my $line  = 0;
  my $dline = 0;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "search.yahoo.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/bin/search?p=" . $query, "search.yahoo.com", undef, undef);

      &putsock($sock{'irc'}, ":%s %s %s :$msg{'search-result'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "Yahoo!", $keyword);
      while (my $buffer = <$socket>) {
        ++$line;
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s; $buffer =~ s/<b>//; $buffer =~ s/<\/b>//g;

        if (!$idx && ($buffer =~ /<B><FONT face=arial><FONT color=red>No results were found for your search:<\/FONT>/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword);
          $nodoc = 1; last;
        } else {
          if ($dline == $line) { ($desc) = $buffer =~ /([^<>]+)<\/a>$/; $desc =~ s/&lt;/</g; $desc =~ s/&gt;/>/g; }
          if (!$desc && ($buffer =~ /^<big>$/)) { $dline = $line + 2; }
          if (!$link)  { ($link) = $buffer =~ /^<font color=746884>(\S+)(&nbsp;&nbsp;.*)?<\/font>$/; }
          if ($desc && $link) {
            if (++$idx > $cfg{'search-limit'}) { last; }
            &putsock($sock{'irc'}, ":%s %s %s :%d) %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $desc);
            &putsock($sock{'irc'}, ":%s %s %s :   %s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $link);
            undef($desc); undef($desc); undef($link); $dline = 0;
          }
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$idx && !$nodoc) { &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-doc'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $keyword); }
  &serv_record($nick, $services{'FindServ'}->{'nick'}, "YAHOO $keyword");
  return 0;
}

sub findserv_whois {
  my ($nick, $domain) = (shift, shift);
  my $lastdot = rindex($domain, ".");
  my $suffix  = uc(substr($domain, ++$lastdot));

  if (!$domain) { # || ($domain && (($suffix ne "COM") || ($suffix ne "NET") || ($suffix ne "ORG") || ($suffix ne "EDU")))) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "WHOIS <domain>");
    return 0;
  }

  my $nomatch = 0;
  my $skip    = 1;
  my $socket;
  my $server;

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "rs.internic.net",
    PeerPort => "whois(43)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &putsock($socket, "=%s\r", $domain);

      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;
        if ($buffer =~ /^No match for/) {
          $nomatch = 1;
          last;
        } elsif ($buffer =~ /^Domain Name: (\S+)$/) {
          if (uc($1) eq uc($domain)) { $skip = 0; }
        } elsif (!$skip && ($buffer =~ /^Whois Server: (\S+)$/)) {
          $server = $1;
          last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if ($nomatch) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-match'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $domain);
    return 0;
  } elsif (!$server) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => $server,
    PeerPort => "whois(43)",
    Timeout  => $cfg{'search-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  &putsock($sock{'irc'}, ":%s %s %s :-- WHOIS result for domain [%s] --", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $domain);

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'search-timeout'});

    eval {
      $socket->autoflush(1);
      &putsock($socket, "%s\r", $domain);

      while (<$socket>) {
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $_);
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  &serv_record($nick, $services{'FindServ'}->{'nick'}, "WHOIS $domain");
  return 0;
}

sub findserv_nslookup {
  my ($nick, $query) = (shift, shift);

  if (!$query) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "NSLOOKUP <domain|ip>");
    return 0;
  } else {
    my (@ip, $addr, $result);
 
    if ($query =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
      $addr   = pack("C4", split(/\./, $query));
      $result = gethostbyaddr($addr, &AF_INET);
    } else {
      @ip     = unpack("C4", (gethostbyname($query))[4]);
      $result = join("\.", @ip);
    }

    if ($result) {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'dns-resolv'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $query, $result);
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :$msg{'dns-unable'}", $services{'FindServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $query);
    }

    &serv_record($nick, $services{'FindServ'}->{'nick'}, "NSLOOKUP $query");
    return 0;
  }
}

1;