# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: irc.pl

use strict;

if ($] < 5.006) {
  use vars qw(%cfg %check %flags %fserv %fsquery %im %irc %services %sock $irc_bind $netmerge $server_bind $ts $TMP);
} else {
  our (%cfg, %check, %flags, %fserv, %fsquery, %im, %irc, %services, %sock, $irc_bind, $netmerge, $server_bind, $ts, $TMP);
}

$server_bind = {
  'PING'   => \&server_ping,
  'SERVER' => \&server_server,
  'SQUIT'  => \&server_squit,
  'ERROR'  => \&server_error,
  'NICK'   => \&server_nick
};

$irc_bind = {
  'PONG'    => \&server_pong,
  'PRIVMSG' => \&irc_privmsg,
  'NICK'    => \&irc_nick,
  'SERVER'  => \&irc_server,
  'MODE'    => \&irc_mode,
  'QUIT'    => \&irc_quit,
  'KICK'    => \&irc_kick,
  'TOPIC'   => \&irc_topic,
  'KILL'    => \&irc_kill,
  'SJOIN'   => \&irc_sjoin,
  'PART'    => \&irc_part,
  'PING'    => \&irc_ping,
  'WHOIS'   => \&irc_whois
};

sub server_ping {
  my $buffer = shift;

  &putsock($sock{'irc'}, ":%s PONG %s", $cfg{'irc-fakeserv'}, (split(/ /, $buffer, 2))[1]);
  $cfg{'irc-pinghub'} = $ts;

  if (!$cfg{'irc-connected'}) {
    $cfg{'irc-connected'} = $ts;

    foreach (keys %services) { if (exists $services{$_}->{'channel'}) { &check_channel($services{$_}->{'bot'}); } }
    if (!open($TMP, "+>>" . $cfg{'tmp-file'})) { &putlog("y", "Unable to write tmp file: %s", $cfg{'tmp-file'}); }
  }

  return 0;
}

sub server_pong {
  my $buffer = shift;
  my $server = (split(/:/, $buffer, 0))[1]; $server = (split(/ /, $server, 0))[0];

  $irc{'server'}->{lc($server)}->{'lag'}  = $ts - $irc{'server'}->{lc($server)}->{'ping'};
  $irc{'server'}->{lc($server)}->{'pong'} = 1;

  if (lc($irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hub'}) eq lc($server)) { $cfg{'irc-pinghub'} = $ts; }
  return 0;
}

sub server_server {
  my $buffer = shift;

  my $server = (split(/ /, $buffer, 0))[1];
  $irc{'server'}->{lc($server)}->{'name'} = $server;
  $irc{'server'}->{lc($server)}->{'info'} = (split(/:/, $buffer, 2))[1];
  $irc{'server'}->{lc($server)}->{'hub'}  = $server;
  $irc{'server'}->{lc($server)}->{'time'} = $ts;
  $irc{'server'}->{lc($server)}->{'lag'}  = 0;
  $irc{'server'}->{lc($server)}->{'ping'} = $ts;
  $irc{'server'}->{lc($server)}->{'pong'} = 1;
  $irc{'server'}->{lc($server)}->{'hopcount'} = (split(/ /, $buffer, 0))[2];
  @{$irc{'server'}->{lc($server)}->{'users'}} = (0, 0, 0);
  @{$irc{'server'}->{lc($server)}->{'opers'}} = (0, 0, 0);

  $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hub'} = $server;
  $irc{'server'}->{lc($server)}->{'leaf'}->{lc($cfg{'irc-fakeserv'})} = $cfg{'irc-fakeserv'};

  if (scalar keys %{$irc{'server'}} > $irc{'stats'}->{'servs'}[0]) {
    @{$irc{'stats'}->{'servs'}} = (scalar keys %{$irc{'server'}}, $ts);
    if ($cfg{'irc-connected'}) { &putinfo("x", "New max server count: %d", $irc{'stats'}->{'servs'}[0]); }
  }

  return 0;
}

sub server_squit {
  my $buffer = shift;

  my $server = (split(/ /, $buffer, 0))[1];
  foreach (keys %{$irc{'server'}}) {
    if (exists $irc{'server'}->{$_}->{'leaf'}->{lc($server)}) { delete $irc{'server'}->{$_}->{'leaf'}->{lc($server)}; }
  }

  &putinfo("l", "Netsplit: %s (hub: %s)", $server, $irc{'server'}->{lc($server)}->{'hub'});
  delete $irc{'server'}->{lc($server)};
  return 0;
}

sub server_error {
  &del_socket($sock{'irc'});
  return 0;
}

sub server_nick {
  my $buffer = shift;

  my $nick = (split(/ /, $buffer, 0))[1];
  my $serv = (split(/ /, $buffer, 0))[7];
  my $mode = (split(/ /, $buffer, 0))[4];
  my $what = 1;

  $irc{'client'}->{lc($nick)}->{'nick'} = $nick;
  $irc{'client'}->{lc($nick)}->{'user'} = (split(/ /, $buffer, 0))[5];
  $irc{'client'}->{lc($nick)}->{'host'} = (split(/ /, $buffer, 0))[6];
  $irc{'client'}->{lc($nick)}->{'real'} = (split(/:/, $buffer, 2))[1];
  $irc{'client'}->{lc($nick)}->{'serv'} = $serv;
  $irc{'client'}->{lc($nick)}->{'time'} = (split(/ /, $buffer, 0))[3];
  $irc{'client'}->{lc($nick)}->{'mode'} = '';
  $irc{'client'}->{lc($nick)}->{'access'}   = '';
  $irc{'client'}->{lc($nick)}->{'lastcmd'}  = 0;
  $irc{'client'}->{lc($nick)}->{'floodcnt'} = 0;
  $irc{'client'}->{lc($nick)}->{'ignored'}  = 0;

  if ($cfg{'irc-connected'} && !$netmerge) {
    &putinfo("C", "Client connection: %s!%s\@%s [%s]", $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $irc{'client'}->{lc($nick)}->{'serv'});

    if (exists $check{lc($nick)} && ($check{lc($nick)}->{'user'} eq $irc{'client'}->{lc($nick)}->{'user'}) && ($check{lc($nick)}->{'host'} eq $irc{'client'}->{lc($nick)}->{'host'})) {
      $check{lc($nick)}->{'kill'} = 0;
    } else {
      $check{lc($nick)}->{'nick'} = $nick;
      $check{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
      $check{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
      $check{lc($nick)}->{'kill'} = 0;
    }
  }


  foreach (split(//, $mode)) {
    if ($_ eq "+") {
      $what = 1;
    } elsif ($_ eq "-") {
      $what = 0;
    } elsif ($flags{'client'}->{$_}) {
      if ($what) {
        if ($_ eq "o") {
          if (++$irc{'stats'}->{'opers'}[0] > $irc{'stats'}->{'opers'}[1]) {
            if ($cfg{'irc-connected'} && !$netmerge) { &putinfo("x", "New max operator count: %d", $irc{'stats'}->{'opers'}[0]); }
            $irc{'stats'}->{'opers'}[1] = $irc{'stats'}->{'opers'}[0];
            $irc{'stats'}->{'opers'}[2] = $ts;
          }

          if (++$irc{'server'}->{lc($serv)}->{'opers'}[0] > $irc{'server'}->{lc($serv)}->{'opers'}[1]) {
            $irc{'server'}->{lc($serv)}->{'opers'}[1] = $irc{'server'}->{lc($serv)}->{'opers'}[0];
            $irc{'server'}->{lc($serv)}->{'opers'}[2] = $ts;
          }
        }

        $irc{'client'}->{lc($nick)}->{'mode'} |= $flags{'client'}->{$_};
      } else {
        if ($_ eq "o") {
          --$irc{'stats'}->{'opers'}[0];
          --$irc{'server'}->{lc($serv)}->{'opers'}[0];
        }

        $irc{'client'}->{lc($nick)}->{'mode'} &= ~$flags{'client'}->{$_};
      }
    }
  }

  &client_stats($nick);

  return 0;
}

sub irc_privmsg {
  my $buffer = shift;

  my $nick   = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $target = (split(/ /, $buffer, 0))[2];
  my $query  = (split(/:/, $buffer, 3))[2];

  if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

  foreach (keys %{$irc{'ignore'}}) {
    my $umaskrx = &plain2regexp($irc{'ignore'}->{$_}->{'usermask'});
    my $hmaskrx = &plain2regexp($irc{'ignore'}->{$_}->{'hostmask'});

    if (($irc{'client'}->{lc($nick)}->{'user'} =~ /$umaskrx/i) && ($irc{'client'}->{lc($nick)}->{'host'} =~ /$hmaskrx/i)) { return 0; }
  }

  if ($target =~ /^#/) {
    &process_channel($nick, $target, $query);
  } elsif (lc($target) eq lc($services{'DataServ'}->{'nick'})) {
    &process_dataserv($nick, $query);
  } elsif (exists $services{'FindServ'} && (lc($target) eq lc($services{'FindServ'}->{'nick'}))) {
    &process_findserv($nick, $query);
  } elsif (exists $services{'MobiServ'} && (lc($target) eq lc($services{'MobiServ'}->{'nick'}))) {
    &process_mobiserv($nick, $query);
  } elsif (exists $services{'ScanServ'} && (lc($target) eq lc($services{'ScanServ'}->{'nick'}))) {
    &process_scanserv($nick, $query);
  } elsif (exists $services{'FileServ'} && (lc($target) eq lc($services{'FileServ'}->{'nick'}))) {
    &process_fileserv($nick, $query);
  } elsif (exists $services{'PageServ'} && (lc($target) eq lc($services{'PageServ'}->{'nick'}))) {
    &process_pageserv($nick, $query);
  } else {
    if (exists $services{'PageServ'}) {
      if ($cfg{'icq-uin'}) {
        foreach (keys %{$im{'icq'}}) {
          if (lc($target) eq lc($im{'icq'}->{$_}->{'irc'})) {
            &icq_msguser($nick, $im{'icq'}->{$_}->{'uin'}, $query);
            return 0;
          }
        }
      }

      if ($cfg{'aim-uid'}) {
        foreach (keys %{$im{'aim'}}) {
          if (lc($target) eq lc($im{'aim'}->{$_}->{'irc'})) {
            &aim_msguser($nick, $im{'aim'}->{$_}->{'uid'}, $query);
            return 0;
          }
        }
      }

      if ($cfg{'ymsg-uid'}) {
        foreach (keys %{$im{'ymsg'}}) {
          if (lc($target) eq lc($im{'ymsg'}->{$_}->{'irc'})) {
            &ymsg_msguser($nick, $im{'ymsg'}->{$_}->{'uid'}, $query);
            return 0;
          }
        }
      }

      if ($cfg{'msn-uid'}) {
        foreach (keys %{$im{'msn'}}) {
          if (lc($target) eq lc($im{'msn'}->{$_}->{'irc'})) {
            &msn_msguser($nick, $im{'msn'}->{$_}->{'uid'}, $query);
            return 0;
          }
        }
      }

    }
  }

  return 0;
}

sub irc_nick {
  my $buffer = shift;

  my $oldnick = (split(/:/, $buffer, 0))[1]; $oldnick = (split(/ /, $oldnick, 0))[0];
  my $newnick = (split(/ /, $buffer, 0))[2];

  if (exists $check{lc($oldnick)} && $check{lc($oldnick)}->{'kill'}) { return 0; }

  $irc{'seen'}->{lc($oldnick)}->{'nick'} = $irc{'client'}->{lc($oldnick)}->{'nick'};
  $irc{'seen'}->{lc($oldnick)}->{'time'} = $ts;
  $irc{'seen'}->{lc($oldnick)}->{'user'} = $irc{'client'}->{lc($oldnick)}->{'user'};
  $irc{'seen'}->{lc($oldnick)}->{'host'} = $irc{'client'}->{lc($oldnick)}->{'host'};
  $irc{'seen'}->{lc($oldnick)}->{'what'} = "NICK";
  $irc{'seen'}->{lc($oldnick)}->{'xtra'} = $newnick;

  if (lc($newnick) ne lc($oldnick)) {
    foreach (keys %{$irc{'channel'}}) {
      if (exists $irc{'channel'}->{$_}->{'names'}->{lc($oldnick)}) {
        $irc{'channel'}->{$_}->{'names'}->{lc($newnick)}->{'nick'} = $newnick;
        $irc{'channel'}->{$_}->{'names'}->{lc($newnick)}->{'mode'} = $irc{'channel'}->{$_}->{'names'}->{lc($oldnick)}->{'mode'};
        delete $irc{'channel'}->{$_}->{'names'}->{lc($oldnick)};
      }
    }

    $irc{'client'}->{lc($newnick)}->{'nick'} = $newnick;
    $irc{'client'}->{lc($newnick)}->{'user'} = $irc{'client'}->{lc($oldnick)}->{'user'};
    $irc{'client'}->{lc($newnick)}->{'host'} = $irc{'client'}->{lc($oldnick)}->{'host'};
    $irc{'client'}->{lc($newnick)}->{'real'} = $irc{'client'}->{lc($oldnick)}->{'real'};
    $irc{'client'}->{lc($newnick)}->{'serv'} = $irc{'client'}->{lc($oldnick)}->{'serv'};
    $irc{'client'}->{lc($newnick)}->{'mode'} = $irc{'client'}->{lc($oldnick)}->{'mode'};
    $irc{'client'}->{lc($newnick)}->{'time'} = $irc{'client'}->{lc($oldnick)}->{'time'};
    $irc{'client'}->{lc($newnick)}->{'access'}   = $irc{'client'}->{lc($oldnick)}->{'access'};
    $irc{'client'}->{lc($newnick)}->{'lastcmd'}  = $irc{'client'}->{lc($oldnick)}->{'lastcmd'};
    $irc{'client'}->{lc($newnick)}->{'floodcnt'} = $irc{'client'}->{lc($oldnick)}->{'floodcnt'};
    $irc{'client'}->{lc($newnick)}->{'ignored'}  = $irc{'client'}->{lc($oldnick)}->{'ignored'};
    delete $irc{'client'}->{lc($oldnick)};
  } else {
    foreach (keys %{$irc{'channel'}}) {
      if (exists $irc{'channel'}->{$_}->{'names'}->{lc($oldnick)}) { $irc{'channel'}->{$_}->{'names'}->{lc($oldnick)}->{'nick'} = $newnick; }
    }

    $irc{'client'}->{lc($oldnick)}->{'nick'} = $newnick;
  }

  foreach (keys %fserv) {
    if ($fserv{$_}->{'nick'} eq $oldnick) {
      $fserv{$_}->{'nick'} = $newnick;
      last;
    }
  }

  if (exists $services{'PageServ'}) {
    foreach (keys %{$im{'request'}}) {
      if ($im{'request'}->{$_}->{'nick'} eq $oldnick) {
        $im{'request'}->{$_}->{'nick'} = $newnick;
        last;
      }
    }
  }

  foreach (keys %fsquery) {
    if ($fsquery{$_}->{'nick'} eq $oldnick) {
      $fsquery{lc($newnick)}->{'nick'} = $newnick;
      $fsquery{lc($newnick)}->{'time'} = $fsquery{$_}->{'time'};
      delete $fsquery{$_};
      last;
    }
  }

  foreach (keys %{$sock{'list'}}) {
    if (($sock{'list'}->{$_}->{'type'} eq "FSERV-G") && ($sock{'list'}->{$_}->{'nick'} eq $oldnick)) {
      $sock{'list'}->{$_}->{'nick'} = $newnick;
      last;
    }
  }

  &putinfo("n", "Nick change: %s -> %s", $oldnick, $newnick);
  return 0;
}

sub irc_server {
  my $buffer = shift;

  my $server = (split(/ /, $buffer, 0))[2];
  my $hub    = (split(/:/, $buffer, 0))[1]; $hub = (split(/ /, $hub, 0))[0];

  $irc{'server'}->{lc($server)}->{'name'} = $server;
  $irc{'server'}->{lc($server)}->{'info'} = (split(/:/, $buffer, 3))[2];
  $irc{'server'}->{lc($server)}->{'hub'}  = $hub;
  $irc{'server'}->{lc($server)}->{'time'} = $ts;
  $irc{'server'}->{lc($server)}->{'lag'}  = 0;
  $irc{'server'}->{lc($server)}->{'ping'} = $ts;
  $irc{'server'}->{lc($server)}->{'pong'} = 1;
  $irc{'server'}->{lc($server)}->{'hopcount'} = (split(/ /, $buffer, 0))[3];
  @{$irc{'server'}->{lc($server)}->{'users'}} = (0, 0, 0);
  @{$irc{'server'}->{lc($server)}->{'opers'}} = (0, 0, 0);

  $irc{'server'}->{lc($hub)}->{'leaf'}->{lc($server)} = $irc{'server'}->{lc($server)}->{'name'};

  if (scalar keys %{$irc{'server'}} > $irc{'stats'}->{'servs'}[0]) {
    @{$irc{'stats'}->{'servs'}} = (scalar keys %{$irc{'server'}}, $ts);
    if ($cfg{'irc-connected'}) { &putinfo("x", "New max server count: %d", $irc{'stats'}->{'servs'}[0]); }
  }

  if ($cfg{'irc-connected'}) {
    &putinfo("l", "Server %s has connected to %s", $server, $hub);
    $netmerge = $ts;
  }

  return 0;
}

sub irc_mode {
  my $buffer = shift;

  my $nick = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $targ = (split(/ /, $buffer, 0))[2];
  my $mode = (split(/:/, $buffer, 0))[2];
  my $cmod = (split(/ /, $buffer, 0))[3];
  my $args = (split(/ /, $buffer, 5))[4];
  my $what = 1;

  if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

  if ($targ =~ /^#/) {
    &channel_mode($nick, $targ, $cmod, $args);
    if ($cfg{'irc-connected'}) {
      if (exists $irc{'client'}->{lc($nick)}) {
        &putinfo("m", "Mode: %s [%s%s] by %s!%s\@%s", $targ, $cmod, $args ? " " . $args : "", $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
      } else {
        &putinfo("m", "Mode: %s [%s%s] by %s", $targ, $cmod, $args ? " " . $args : "", $irc{'server'}->{lc($nick)}->{'name'});
      }
    }
  } else {
    foreach (split(//, $mode)) {
      if ($_ eq "+") {
        $what = 1;
      } elsif ($_ eq "-") {
        $what = 0;
      } elsif ($flags{'client'}->{$_}) {
        my $serv = $irc{'client'}->{lc($targ)}->{'serv'};

        if ($what) {
          $irc{'client'}->{lc($targ)}->{'mode'} |= $flags{'client'}->{$_};
          if ($_ eq "o") {
            if ($cfg{'irc-connected'}) { &putinfo("o", "New Operator: %s (%s\@%s) [%s]", $irc{'client'}->{lc($targ)}->{'nick'}, $irc{'client'}->{lc($targ)}->{'user'}, $irc{'client'}->{lc($targ)}->{'host'},  $irc{'client'}->{lc($targ)}->{'serv'}); }

            if (++$irc{'stats'}->{'opers'}[0] > $irc{'stats'}->{'opers'}[1]) {
              if ($cfg{'irc-connected'}) { &putinfo("x", "New max operator count: %d", $irc{'stats'}->{'opers'}[0]); }
              $irc{'stats'}->{'opers'}[1] = $irc{'stats'}->{'opers'}[0];
              $irc{'stats'}->{'opers'}[2] = $ts;
            }

            if (++$irc{'server'}->{lc($serv)}->{'opers'}[0] > $irc{'server'}->{lc($serv)}->{'opers'}[1]) {
              $irc{'server'}->{lc($serv)}->{'opers'}[1] = $irc{'server'}->{lc($serv)}->{'opers'}[0];
              $irc{'server'}->{lc($serv)}->{'opers'}[2] = $ts;
            }
          }
        } else {
          if ($_ eq "o") {
            --$irc{'stats'}->{'opers'}[0];
            --$irc{'server'}->{lc($serv)}->{'opers'}[0];
          }

          $irc{'client'}->{lc($targ)}->{'mode'} &= ~$flags{'client'}->{$_};
        }
      }
    }
  }

  return 0;
}

sub irc_quit {
  my $buffer = shift;

  my $nick   = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $reason = (split(/:/, $buffer, 3))[2];

  $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
  $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
  $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
  $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
  $irc{'seen'}->{lc($nick)}->{'what'} = "QUIT";
  $irc{'seen'}->{lc($nick)}->{'xtra'} = $reason;

  &putinfo("E", "Client exit: %s!%s\@%s [%s]", $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $irc{'client'}->{lc($nick)}->{'serv'});

  if (exists $check{lc($nick)} && ($check{lc($nick)}->{'user'} eq $irc{'client'}->{lc($nick)}->{'user'}) && ($check{lc($nick)}->{'host'} eq $irc{'client'}->{lc($nick)}->{'host'})) { delete $check{lc($nick)}; }
  &del_client($nick);
  return 0;
}

sub irc_kick {
  my $buffer = shift;

  my $nick   = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $chan   = (split(/ /, $buffer, 0))[2];
  my $target = (split(/ /, $buffer, 0))[3];
  my $reason = (split(/:/, $buffer, 3))[2];

  if (exists $check{lc($target)} && $check{lc($target)}->{'kill'}) { return 0; }

  $irc{'seen'}->{lc($target)}->{'nick'} = $irc{'client'}->{lc($target)}->{'nick'};
  $irc{'seen'}->{lc($target)}->{'time'} = $ts;
  $irc{'seen'}->{lc($target)}->{'user'} = $irc{'client'}->{lc($target)}->{'user'};
  $irc{'seen'}->{lc($target)}->{'host'} = $irc{'client'}->{lc($target)}->{'host'};
  $irc{'seen'}->{lc($target)}->{'what'} = "KICK";
  $irc{'seen'}->{lc($target)}->{'xtra'} = $irc{'channel'}->{lc($chan)}->{'channel'};


  if (exists $irc{'client'}->{lc($nick)}) {
    &putinfo("k", "Channel kick: %s [%s] by %s!%s\@%s (%s)", $chan, $target, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $reason);
  } else {
    &putinfo("k", "Channel kick: %s [%s] by %s (%s)", $chan, $target, $irc{'server'}->{lc($nick)}->{'name'}, $reason);
  }

  if (exists $irc{'channel'}->{lc($chan)}->{'names'}->{lc($target)}) { delete $irc{'channel'}->{lc($chan)}->{'names'}->{lc($target)}; }
  if (scalar keys %{$irc{'channel'}->{lc($chan)}->{'names'}} == 0) { delete $irc{'channel'}->{lc($chan)}; }

  foreach (keys %services) {
    if ((lc($services{$_}->{'nick'}) eq lc($target)) && exists $services{$_}->{'channel'} && exists $services{$_}->{'channel'}->{lc($chan)}) { &serv_join($services{$_}->{'nick'}, $services{$_}->{'channel'}->{lc($chan)}->{'channel'}); last; }
  }

  if (exists $services{'PageServ'}) {
    if ($cfg{'icq-uin'}) {
      foreach (keys %{$im{'icq'}}) {
        if ((lc($chan) eq lc($cfg{'icq-chan'})) && (lc($im{'icq'}->{$_}->{'irc'}) eq lc($target))) {
          &icq_killuser($_, "Kicked");
          &icq_requser($_);
          last;
        }
      }
    }

    if ($cfg{'aim-uid'}) {
    }

    if ($cfg{'ymsg-uid'}) {
      foreach (keys %{$im{'ymsg'}}) {
        if ((lc($chan) eq lc($cfg{'ymsg-chan'})) && (lc($im{'ymsg'}->{$_}->{'irc'}) eq lc($target))) {
          &ymsg_killuser($_, "Kicked");
          &ymsg_requser($_);
          last;
        }
      }
    }

    if ($cfg{'msn-uid'}) {
      foreach (keys %{$im{'msn'}}) {
        if ((lc($chan) eq lc($cfg{'msn-chan'})) && (lc($im{'msn'}->{$_}->{'irc'}) eq lc($target))) {
          &msn_killuser($_, "Kicked");
          &msn_requser($_);
          last;
        }
      }
    }
  }

  return 0;
}

sub irc_topic {
  my $buffer = shift;

  my $nick  = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $chan  = (split(/ /, $buffer, 0))[2];
  my $topic = (split(/:/, $buffer, 3))[2];
  my $author;

  if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

  if (exists $irc{'client'}->{lc($nick)}) {
    $author = $irc{'client'}->{lc($nick)}->{'nick'} . "!" . $irc{'client'}->{lc($nick)}->{'user'} . "\@" . $irc{'client'}->{lc($nick)}->{'host'};
  } else {
    $author = $irc{'server'}->{lc($nick)}->{'name'};
  }

  @{$irc{'channel'}->{lc($chan)}->{'topic'}} = ($topic, $author, $ts);
  &putinfo("t", "Channel topic: %s (%s) %s", $chan, $author, $topic);
  return 0;
}

sub irc_kill {
  my $buffer = shift;

  my $nick   = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $reason = (split(/:/, $buffer, 3))[2];
  my $target = (split(/ /, $buffer, 0))[2];

  $irc{'seen'}->{lc($target)}->{'nick'} = $irc{'client'}->{lc($target)}->{'nick'};
  $irc{'seen'}->{lc($target)}->{'time'} = $ts;
  $irc{'seen'}->{lc($target)}->{'user'} = $irc{'client'}->{lc($target)}->{'user'};
  $irc{'seen'}->{lc($target)}->{'host'} = $irc{'client'}->{lc($target)}->{'host'};
  $irc{'seen'}->{lc($target)}->{'what'} = "QUIT";
  $irc{'seen'}->{lc($target)}->{'xtra'} = $reason;

  if (exists $irc{'client'}->{lc($nick)}) {
    ++$irc{'stats'}->{'kills'}[0];
    &putinfo("O", "Operator Kill: [%s] by %s!%s\@%s (%s)", $target, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $reason);
  } elsif (exists $irc{'server'}->{lc($nick)}) {
    ++$irc{'stats'}->{'kills'}[1];
    &putinfo("S", "Server Kill: [%s] by %s (%s)", $target, $irc{'server'}->{lc($nick)}->{'name'}, $reason);
  }

  if (exists $check{lc($target)} && ($check{lc($target)}->{'user'} eq $irc{'client'}->{lc($target)}->{'user'}) && ($check{lc($target)}->{'host'} eq $irc{'client'}->{lc($target)}->{'host'})) { delete $check{lc($target)}; }
  &del_client($target);

  foreach (keys %services) {
    if (lc($services{$_}->{'nick'}) eq lc($target)) { &introduce_bot($services{$_}->{'bot'}); last; }
  }

  if (exists $services{'PageServ'}) {
    if ($cfg{'icq-uin'}) {
      foreach (keys %{$im{'icq'}}) {
        if (lc($im{'icq'}->{$_}->{'irc'}) eq lc($target)) {
          &icq_killuser($_, "Killed");
          &icq_requser($_);
          last;
        }
      }
    }

    if ($cfg{'aim-uid'}) {
    }

    if ($cfg{'ymsg-uid'}) {
      foreach (keys %{$im{'ymsg'}}) {
        if (lc($im{'ymsg'}->{$_}->{'irc'}) eq lc($target)) {
          &ymsg_killuser($_, "Killed");
          &ymsg_requser($_);
          last;
        }
      }
    }

    if ($cfg{'msn-uid'}) {
      foreach (keys %{$im{'msn'}}) {
        if (lc($im{'msn'}->{$_}->{'irc'}) eq lc($target)) {
          &msn_killuser($_, "Killed");
          &msn_requser($_);
          last;
        }
      }
    }
  }

  return 0;
}

sub irc_sjoin {
  my $buffer = shift;
  my ($serv, $time, $chan, $mode, $args, $list);

  if ($buffer =~ /:(\S+) SJOIN (\d+) (\S+) +(\S+)(\s+.*)? :(.*)/) {
    $serv = $1;
    $time = $2;
    $chan = $3;
    $mode = $4;
    $list = $6;
    $args = $5; $args =~ s/^\s+//;
  } elsif ($buffer =~ /:(\S+) SJOIN (\d+) (\d+) (\S+) +(\S+)(\s+.*)? :(.*)/) {
    $serv = $1;
    $time = $3;
    $chan = $4;
    $mode = $5;
    $list = $7;
    $args = $6; $args =~ s/^\s+//;
  } elsif ($buffer =~ /:(\S+) SJOIN (\d+) :?(\S+)/) {
    $list = $1;
    $time = $2;
    $chan = $3;
  }

  my $nick = (split(/ /, $list, 0))[0]; $nick =~ s/^\@//; $nick =~ s/^\%//; $nick =~ s/^\+//;

  if (!$cfg{'irc-connected'}) {
    $irc{'channel'}->{lc($chan)}->{'channel'}  = $chan;
    $irc{'channel'}->{lc($chan)}->{'mode'}     = '';
    $irc{'channel'}->{lc($chan)}->{'time'}     = $time;
    @{$irc{'channel'}->{lc($chan)}->{'max'}}   = (1, $ts);
    @{$irc{'channel'}->{lc($chan)}->{'topic'}} = (undef, undef, 0);
    &channel_mode($nick, $chan, $mode, $args);
  } else {
    if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

    if (!exists $irc{'channel'}->{lc($chan)}) {
      &putinfo("j", "New channel: %s (created by %s!%s\@%s)", $chan, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
      $irc{'channel'}->{lc($chan)}->{'channel'}  = $chan;
      $irc{'channel'}->{lc($chan)}->{'mode'}     = '';
      $irc{'channel'}->{lc($chan)}->{'time'}     = $time;
      @{$irc{'channel'}->{lc($chan)}->{'max'}}   = (1, $ts);
      @{$irc{'channel'}->{lc($chan)}->{'topic'}} = (undef, undef, 0);
    } else {
      &putinfo("j", "Channel join: %s (%s!%s\@%s)", $chan, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    foreach (keys %services) {
      if (exists $services{$_}->{'channel'} && exists $services{$_}->{'channel'}->{lc($chan)} && $services{$_}->{'channel'}->{lc($chan)}->{'joinmsg'}) {
        my $joinmsg = $services{$_}->{'channel'}->{lc($chan)}->{'joinmsg'};
        $joinmsg =~ s/%C/$chan/g; $joinmsg =~ s/%N/$nick/g;
        &putsock($sock{'irc'}, ":%s %s %s :%s", $services{$_}->{'nick'}, $cfg{'irc-comm'}, $nick, $joinmsg);
      }
    }
  }

  foreach (split(/ /, $list)) {
    $nick = $_; $nick =~ s/^\@//; $nick =~ s/^\%//; $nick =~ s/^\+//;

    $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
    $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)}->{'mode'} = '';

    if (substr($_, 0, 1) eq "\@") {
      $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)}->{'mode'} |= $flags{'channel'}->{'o'};
    } elsif ((substr($_, 0, 1) eq "\%") || (substr($_, 0, 2) eq "\%")) {
      $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)}->{'mode'} |= $flags{'channel'}->{'h'};
    } elsif ((substr($_, 0, 1) eq "+") || (substr($_, 0, 2) eq "+") || (substr($_, 0, 3) eq "+")) {
      $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)}->{'mode'} |= $flags{'channel'}->{'v'};
    }

    $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
    $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
    $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
    $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
    $irc{'seen'}->{lc($nick)}->{'what'} = "JOIN";
    $irc{'seen'}->{lc($nick)}->{'xtra'} = $irc{'channel'}->{lc($chan)}->{'channel'};
  }

  &channel_stats($chan);

  return 0;
}

sub irc_part {
  my $buffer = shift;

  my $nick = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  my $chan = (split(/ /, $buffer, 0))[2];

  if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

  $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
  $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
  $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
  $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
  $irc{'seen'}->{lc($nick)}->{'what'} = "PART";
  $irc{'seen'}->{lc($nick)}->{'xtra'} = $irc{'channel'}->{lc($chan)}->{'channel'};

  foreach (keys %services) {
    if (exists $services{$_}->{'channel'} && exists $services{$_}->{'channel'}->{lc($chan)} && $services{$_}->{'channel'}->{lc($chan)}->{'partmsg'}) {
      my $partmsg = $services{$_}->{'channel'}->{lc($chan)}->{'partmsg'};
      $partmsg =~ s/%C/$chan/g; $partmsg =~ s/%N/$nick/g;
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{$_}->{'nick'}, $cfg{'irc-comm'}, $nick, $partmsg);
    }
  }

  &putinfo("p", "Channel part: %s (%s!%s\@%s)", $chan, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
  delete $irc{'channel'}->{lc($chan)}->{'names'}->{lc($nick)};
  if (scalar keys %{$irc{'channel'}->{lc($chan)}->{'names'}} == 0) { delete $irc{'channel'}->{lc($chan)}; }
  return 0;
}

sub irc_ping {
  my $buffer = shift;

  my $nick = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick, 0))[0];
  &putsock($sock{'irc'}, ":%s PONG %s :%s", $cfg{'irc-fakeserv'}, $cfg{'irc-fakeserv'}, $nick);
  return 0;
}

sub irc_whois {
  my $buffer = shift;

  my $nick   = (split(/:/, $buffer, 0))[1]; $nick = (split(/ /, $nick))[0];
  my $target = (split(/ /, $buffer, 0))[2];

  if (exists $check{lc($nick)} && $check{lc($nick)}->{'kill'}) { return 0; }

  &putsock($sock{'irc'}, ":%s 311 %s %s %s %s * :%s", $cfg{'irc-fakeserv'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($target)}->{'user'}, $irc{'client'}->{lc($target)}->{'host'}, $irc{'client'}->{lc($target)}->{'real'});
  &putsock($sock{'irc'}, ":%s 312 %s %s :%s", $cfg{'irc-fakeserv'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}, $cfg{'irc-fakeserv'}, $cfg{'irc-fakeinfo'});
  if ($irc{'client'}->{lc($target)}->{'mode'} & $flags{'client'}->{'o'}) { &putsock($sock{'irc'}, ":%s 313 %s %s :is a Network Service Daemon", $cfg{'irc-fakeserv'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}); }
  &putsock($sock{'irc'}, ":%s 317 %s %s %d %d :seconds idle, signon time", $cfg{'irc-fakeserv'}, $nick, $irc{'client'}->{lc($target)}->{'nick'}, 0, $irc{'client'}->{lc($target)}->{'time'});
  &putsock($sock{'irc'}, ":%s 318 %s %s :End of /WHOIS list", $cfg{'irc-fakeserv'}, $nick, $irc{'client'}->{lc($target)}->{'nick'});

  &putinfo("y", "Remote WHOIS [%s] requested by %s!%s\@%s", $irc{'client'}->{lc($target)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
  return 0;
}

1;