# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: mobiserv.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %cfg %flags %irc %msg %services %sock $VERSION);
} else {
  our (%access, %cfg, %flags, %irc, %msg, %services, %sock, $VERSION);
}

sub process_mobiserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'MobiServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'MobiServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'MobiServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'MobiServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'MobiServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'MobiServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'MobiServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'MobiServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (uc($cmd) eq "HELP") {
    &mobiserv_help($nick, $arg);
  } elsif ((uc($cmd) eq "SMS") || (uc($cmd) eq "SEND")) {
    &mobiserv_sms($nick, $arg);
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
  }

  return 0;
}

sub mobiserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'MobiServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'MobiServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/mobiserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub mobiserv_sms {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SMS <area> <gsm number> <message>");
    return 0;
  }

  my $area = (split(/ /, $arg, 0))[0];
  my $gsm  = (split(/ /, $arg, 0))[1];
  my $sms  = (split(/ /, $arg, 3))[2];

  if (!$area || !$gsm || ($gsm !~ /^\d+$/) || !$sms) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SMS <area> <gsm number> <message>");
    return 0;
  }

  if (my $pid = fork) {
    if (uc($area) eq "AT") {
      &sms_us_att($nick, $gsm, $sms);
    } elsif (uc($area) eq "CI") {
      &sms_us_cingular($nick, $gsm, $sms);
    } elsif (uc($area) eq "SP") {
      &sms_us_sprint($nick, $gsm, $sms);
    } elsif (uc($area) eq "NX") {
      &sms_us_nextel($nick, $gsm, $sms);
    } elsif (uc($area) eq "VZ") {
      &sms_us_verizon($nick, $gsm, $sms);
    } elsif ($area eq "+32") {
      if (substr($gsm, 0, 3) eq "047") {
        &sms_be_proximus($nick, $gsm, $sms);
      } elsif ((substr($gsm, 0, 4) eq "0485") || (substr($gsm, 0, 4) eq "0486")) {
        &sms_be_orange($nick, $gsm, $sms);
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
      }
    } elsif ($area eq "+33") {
      if ($gsm =~ /^06((03)|(09)|(55)|(1\d)|(2\d))\d{6}$/) {
        &sms_fr_sfr($nick, $gsm, $sms);
      } elsif ($gsm =~ /^06((53)|(98)|(99)|(6\d))\d{6}$/) {
        &sms_fr_6sens($nick, $gsm, $sms);
      #} elsif ($gsm =~ /^06((07)|(08)|(54)|(7\d)|(8\d))\d{6}$/) {
      # &sms_fr_orange($nick, $gsm, $sms);
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
      }
    } elsif ($area eq "+34") {
      &sms_es_vodafone($nick, $gsm, $sms);
    } elsif ($area eq "+39") {
      &sms_it_smsaffari($nick, $gsm, $sms);
    } elsif ($area eq "+41") {
      &sms_ch_bluewin($nick, $gsm, $sms);
    } elsif ($area eq "+49") {
      if ($gsm =~ /^017[69]\d{8}$/) {
        &sms_de_loop($nick, $gsm, $sms);
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'unkn-area'});
    }
    exit(0);
  }
  return 0;
}

sub sms_ch_bluewin {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result  = 0;
  my $confirm = 0;
  my $socket;

  if ($gsm !~ /^417[689]\d{7}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 130) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 130);
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "freesms.bluewin.ch",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "language=e";
      $post .= "\&number=" . substr($gsm, 4, 11);
      $post .= "\&textarea=" . $txt;
      $post .= "\&selectch=\%2B" . substr($gsm, 0, 4);
      $post .= "\&number2=";
      $post .= "\&checkbox=checkbox";
      $post .= "\&textarea2=";
      $post .= "\&leftchars=130";
      $post .= "\&submit=Envoyer";

      $socket->autoflush(1);
      &http_post($socket, "/cgi-bin/sendsms.pl", "freesms.bluewin.ch", "http://freesms.bluewin.ch/index_e.html", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Location: http:\/\/freesms.bluewin.ch\/wrongnumber_e.html/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^Location: http:\/\/freesms.bluewin.ch\/confirm_e.html/) {
          $confirm = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$confirm) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "freesms.bluewin.ch",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/confirm_e.html", "freesms.bluewin.ch", "http://freesms.bluewin.ch/index_e.html", undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /Your SMS message will arrive in the next few minutes./) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+32", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +41 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_be_proximus {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;
  my $socket;
  my ($cookie1, $cookie2);

  if ($gsm !~ /^047\d{7}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 130) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 130);
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "sms.proximus.be",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/WMS/ProxiText?laf=paygo&lan=en", "sms.proximus.be", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: BMB_WMS_COUNTER=(\S+); Max-Age=/) { $cookie1 = $1; }
        if ($buffer =~ /^Set-Cookie: sess=(\S+);path=\//) { $cookie2 = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$cookie1 || !$cookie2) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "sms.proximus.be",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "S45_=en";
      $post .= "&prefix=" . substr($gsm, 0, 4);
      $post .= "&suffix=" . substr($gsm, 4, 10);
      $post .= "&smsmessage=" . $txt;
 
      $socket->autoflush(1);
      &http_post($socket, "/WMS/ProxiText?smsaction=send&#38;lan=en&#38;laf=paygo", "sms.proximus.be", undef, $post, undef, sprintf("BMB_WMS_COUNTER=%s; sess=%s;", $cookie1, $cookie2));

      while (my $buffer = <$socket>) {
        if ($buffer =~ /<p>Please try again later.<\/p>/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /<p>Thank you for using SMS On Line.<\/p>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+32", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +32 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_be_orange {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if ($gsm !~ /^048[56]\d{6}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 140) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 140);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "smscenter.planetinternet.be",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "lang=fr";
      $post .= "\&prefix=" . substr($gsm, 0, 4);
      $post .= "\&number=" . substr($gsm, 4, 10);
      $post .= "\&msg=" . $txt;
      $post .= "\&carac=" . (140 - length($sms));
      $post .= "\&SUBMIT.x=0&SUBMIT.y=0";

      $socket->autoflush(1);
      &http_post($socket, "/cgi-bin/orange.cgi", "smscenter.planetinternet.be", "http://smscenter.planetinternet.be/fr/oranfr.html", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Location: http:\/\/smscenter.planetinternet.be\/hackdoc.html/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^Location: http:\/\/smscenter.planetinternet.be\/fr\/oranokfr.html/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+32", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +32 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_fr_sfr {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if ($gsm !~ /^06((03)|(09)|(55)|(1\d)|(2\d))\d{6}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 110) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 110);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "services.sfr.fr",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $request;
      my $txt = $sms; $txt =~ tr/ /+/;
      my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
      ++$mon; $year += 1900;

      $request .= "PAGE_COURANTE=/textoweb/saisieTextoSfr.jsp";
      $request .= "\&PAGE_SUIVANTE=/textoweb/confirmationEnvoiTextoSfr.jsp";
      $request .= "\&NOTIFICATION_FLAG=false";
      $request .= "\&LANGUAGE=FR";
      $request .= "\&NETWORK=smsc1";
      $request .= "\&DELIVERY_TIME=0";
      $request .= "\&VALIDITY_PERIOD=72";
      $request .= "\&DELIVERY_DATE=" . $mday;
      $request .= "\&DELIVERY_MONTH=" . $mon;
      $request .= "\&DELIVERY_YEAR=" . $year;
      $request .= "\&DELIVERY_HOUR=" . $hour;
      $request .= "\&DELIVERY_MIN=" . $min;
      $request .= "\&NOTIFICATION_ADDRESS=";
      $request .= "\&SENDER=" . $nick;
      $request .= "\&NUM_SENDER=";
      $request .= "\&RECIPIENT=" . $gsm;
      $request .= "\&MINI_TEXTO=0";
      $request .= "\&SHORT_MESSAGE=" . $txt;
      $request .= "\&caracteres=" . (110 - length($txt));
      $request .= "\&imageField2.x=0&imageField2.y=0";

      $socket->autoflush(1);
      &http_get($socket, "/FormulaireSMSSfrPageFind.servlet?" . $request, "services.sfr.fr", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /Service momentan&eacute;ment indisponible./) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /<td class="txt">Votre message a bien \t\ envoy\.<\/td>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+33", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +33 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_fr_6sens {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if ($gsm !~ /^06((53)|(98)|(99)|(6\d))\d{6}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 140) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 140);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.6sens.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "destinataire=" . $gsm;
      $post .= "\&msg=" . $txt;
      $post .= "\&indic=" . (length($sms) - 140);

      $socket->autoflush(1);
      &http_post($socket, "/telemessages/TelemessagesValidEnvoyeText.asp", "www.6sens.com", "http://www.6sens.com/telemessages/TelemessagesAnonymesProcess.asp", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if (($buffer =~ /n'a pas \t\ envoy\ au $gsm/) || ($buffer =~ /^HTTP\/1.1 302 Object moved/)) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /a bien \t\ envoy\ au $gsm/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+33", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +33 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_es_vodafone {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if ($gsm !~ /^6\d{8}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 160) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 160);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "212.73.32.54",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "mBody=" . $sms;
      $post .= "\&tNumber=" . $gsm;

      $socket->autoflush(1);
      &http_post($socket, "/cgi-bin/cgi-airt", "212.73.32.54", "http://sms.airtelvodafone.com/Vodafone/sms/IndexSMS", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /Error al procesar el mensaje/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^HTTP\/1.1 200 OK/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+34", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +34 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_it_smsaffari {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;
  my $socket;
  my $cookie;
  my $location;
  my $confirm;

  if ($gsm !~ /^3\d{9,10}$/) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 143) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 143);
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.smsaffari.it",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "msgCL2=" . (143 - length($txt));
      $post .= "\&lenSSig=0&lenLSig=0&lenSysSig=0";
      $post .= "\&Testo=" . $txt;
      $post .= "\&Pref=" . substr($gsm, 0, 3);
      $post .= "\&Telefono=" . substr($gsm, 3, 10);
      $post .= "\&I2.x=0&I2.y=0";

      $socket->autoflush(1);
      &http_post($socket, "/_sms/gate/cli_1/input.asp", "www.smsaffari.it", undef, $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: (\S+) path=\//) { $cookie = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$cookie) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.smsaffari.it",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "msgCL2=" . (143 - length($txt));
      $post .= "\&lenSSig=0&lenLSig=0&lenSysSig=0";
      $post .= "\&Testo=" . $txt;
      $post .= "\&Pref=" . substr($gsm, 0, 3);
      $post .= "\&Telefono=" . substr($gsm, 3, 10);
      $post .= "\&Action=INVIA";

      $socket->autoflush(1);
      &http_post($socket, "/_sms/gate/cli_1/input.asp", "www.smsaffari.it", "http://www.smsaffari.it", $post, undef, $cookie);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Location: http:\/\/smsaffari.acotel.it(\S+)/) { $location = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$location) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "smsaffari.acotel.it",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, $location, "smsaffari.acotel.it", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: (\S+) path=\//) { $confirm = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$confirm) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.smsaffari.it",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/_sms/conferma.asp", "www.smsaffari.it", undef, $confirm);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /no se est\ procesando correctamente<\/span>/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /con successo/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+39", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +39 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_us_att {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 110) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 110);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.mobile.att.net",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "pin=" . $gsm;
      $post .= "\&from=" . $nick;
      $post .= "\&subject=";
      $post .= "\&message=" . $txt;

      $socket->autoflush(1);
      &http_post($socket, "/messagecenter/pagersend.cgi", "www.mobile.att.net", "http://www.mobile.att.net/messagecenter/index.html", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if (($buffer =~ /^There has been an error sending your message.  Please check your message and try again.<br><br>/) || ($buffer =~ /^'$gsm' is not a valid PCS phone number or PocketNet alias./)) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^Thank you for using AT\&amp;T. Your message has been submitted<br><br>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+1", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +1 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_us_cingular {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if ((length($sms) + length($nick)) > 160) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 160);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "208.62.68.135",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $request;
      my $txt = $sms; $txt =~ tr/ /+/;

      $request .= "TO=" . $gsm;
      $request .= "\&FROM=" . $nick;
      $request .= "\&MESSAGE=" . $txt;
      $request .= "\&SUBMIT=Send+";
      $request .= "\&COUNT=" . length($sms) + length($nick);
      $request .= "\&SUBJECT=";
      $request .= "\&CALLBACK=";
      $request .= "\&min=" . $gsm;
      $request .= "\&msg=FRM:" . $nick . "+MSG:" . $txt;

      $socket->autoflush(1);
      &http_get($socket, "/msgresult.shtml?" . $request, "www.cingular.com", "http://www.cingular.com/messaging", undef);

      while (my $buffer = <$socket>) {
        if (($buffer =~ /^The Cingular Interactive Messaging address you entered was not valid./) || ($buffer =~ /^You must enter a valid Cingular Interactive Messaging address./)) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^Your message will be sent./) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+1", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +1 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_us_verizon {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if ((length($sms) + length($nick)) > 160) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 160);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "www.vtext.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "return=Messaging.jsp";
      $post .= "\&showDisclaimer=y";
      $post .= "\&Send.x=Yes";
      $post .= "\&TCS_MSG_TRACKING_IDS=0." . rand(100000000000000);
      $post .= "\&showgroup=n";
      $post .= "\&SRCADDR=";
      $post .= "\&min=" . $gsm;
      $post .= "\&mins=";
      $post .= "\&subject=";
      $post .= "\&message=" . $txt;
      $post .= "\&trackResponses=No";
      $post .= "\&senderName=" . $nick;
      $post .= "\&from=";
      $post .= "\&priority=normal";
      $post .= "\&tCharacterCount=" . (160 - (length($sms) + length($nick)));
      $post .= "\&Send.x=0&Send.y=0";

      $socket->autoflush(1);
      &http_post($socket, "/results.jsp", "www.vtext.com", "http://www.vtext.com", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /<td>Invalid Destination<\/td>/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /<td>Sending your message<\/td>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+1", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +1 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_us_nextel {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 280 + length($nick) + 2) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 280);
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "messaging.nextel.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "cmd=sendPage";
      $post .= "\&oneWayPTNs=" . $gsm;
      $post .= "\&twoWayPTNs=";
      $post .= "\&ignore=";
      $post .= "\&from=" . $nick;
      $post .= "\&subject=";
      $post .= "\&message=" . $txt;
      $post .= "\&count=" . (length($sms) + length($nick) + 2);
      $post .= "\&rep1=&rep2=&rep3=&rep4=&rep5=&rep6=";
      $post .= "\&replynumber=&replyemail=&confaddr=&schedule=";

      $socket->autoflush(1);
      &http_post($socket, "/cgi/iPageExt.dll", "messaging.nextel.com", undef, $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /<title>Nextel.com - An Error Has Occurred<\/title>/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /<title>Nextel.com - Your Message Has Been Received<\/title>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+1", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +1 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_us_sprint {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;
  my $socket;
  my ($cookie1, $cookie2);

  if (length($gsm) != 10) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 160) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 160);
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "messaging.sprintpcs.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/sml/guestCompose.do", "messaging.sprintpcs.com", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: ARPT=(\S+); path=\//) { $cookie1 = $1; }
        if ($buffer =~ /^Set-Cookie: JSESSIONID=(\S+); path=\//) { $cookie2 = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$cookie1 && !$cookie2) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "messaging.sprintpcs.com",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "post=Send";
      $post .= "\&recipientsAsString=" . $gsm;
      $post .= "\&mobileNumber=" . $gsm;
      $post .= "\&body=" . $txt;
      $post .= "\&counter=" . (160 - length($sms));
      $post .= "\&callback=";

      $socket->autoflush(1);
      &http_post($socket, "/sml/guestCompose.do;jsessionid=" . $cookie2, "messaging.sprintpcs.com", "http://messaging.sprintpcs.com/sml/guestCompose.do", $post, undef, sprintf("ARPT=%s; JSESSIONID=%s;", $cookie1, $cookie2));

      while (my $buffer = <$socket>) {
        if ($buffer =~ /One of the message recipients does not subscribe to this messaging service./) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /src="images\/clear.gif" width=5>Message sent<\/SPAN><\/td>/) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+1", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +1 $gsm (" . length($sms) . " chars)");
  return 0;
}

sub sms_de_loop {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;
  my $socket;
  my ($cookie1, $cookie2, $cookie3, $cookie4, $cookie5);

  if (length($gsm) != 12) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'bad-phone'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $gsm);
    return 0;
  }

  if (length($sms) > 120) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-long'});
    $sms = substr($sms, 0, 120);
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "web2sms.loop.de",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/sms_eingabe.asp", "web2sms.loop.de", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: SITESERVER=(\S+); expires/) { $cookie1 = $1; }
        if ($buffer =~ /^Set-Cookie: VISIT\%5FID=(\d+); path=\//) { $cookie2 = $1; }
        if ($buffer =~ /^Set-Cookie: USER\%5FID=(\d+); expires/) { $cookie3 = $1; }
        if ($buffer =~ /^Set-Cookie: ASPSESSION(\S+); path=\//) { $cookie4 = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$cookie1 || !$cookie2 || !$cookie3 || !$cookie4) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "web2sms.loop.de",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      my $post;
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "MSISDN=49" . substr($gsm, 1, 12);
      $post .= "\&MessageText=" . $txt;

      $socket->autoflush(1);
      &http_post($socket, "/sms_process.asp", "web2sms.loop.de", sprintf("http://web2sms.loop.de/sms_agbs.asp?prefix=49%d&MSISDN=%d&MessageText=%s&num=%d", substr($gsm, 1, 4), substr($gsm, 4, 12), $txt, 120 - length($txt)), $post, undef, sprintf("SITESERVER=%s; Page=2; VISIT\%5FSTAT=0; VISIT\%5FID=%d; USER\%5FID=%d; DB123=1; ASPSESSION%s; msg=test;", $cookie1, $cookie2, $cookie3, $cookie4));

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: SMS\%5FID=(\d+); expires/) { $cookie5 = $1; last; }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$cookie5) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
    return 0;
  }

  $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => "web2sms.loop.de",
    PeerPort => "http(80)",
    Timeout  => $cfg{'sms-timeout'}
  );

  if (!$socket) {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'});
    return 0;
  }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'sms-timeout'});

    eval {
      $socket->autoflush(1);
      &http_get($socket, "/sms_danke.asp?MSISDN=49" . substr($gsm, 1, 12), "web2sms.loop.de", "http://web2sms.loop.de/sms_result.asp?MSISDN=49" . substr($gsm, 1, 12), sprintf("SITESERVER=%s; Page=4; VISIT\%5FSTAT=0; VISIT\%5FID=%d; USER\%5FID=%d; DB123=1; ASPSESSION%s; msg=test; SMS\%5FID=%d;", $cookie1, $cookie2, $cookie3, $cookie4, $cookie5));

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^HTTP\/1.1 302 Object moved/) {
          &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'sms-error'});
          $result = 1; last;
        } elsif ($buffer =~ /^Set-Cookie: SMS\%5FID=; path=\//) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'sms-sent'}", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "+49", $gsm);
          $result = 1; last;
        }
      }
      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);

  if (!$result) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'MobiServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'serv-down'}); }
  $gsm = substr($gsm, 0, length($gsm) - 2) . "##";
  &serv_record($nick, $services{'MobiServ'}->{'nick'}, "SMS +49 $gsm (" . length($sms) . " chars)");
  return 0;
}

1;