# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: msn.pl

use strict;
use Digest::MD5 qw(md5_hex);

if ($] < 5.006) {
  use vars qw(%cfg %im %services %sock $ts);
} else {
  our (%cfg, %im, %services, %sock, $ts);
}

my %msn_error = (
  200 => 'Syntax error',
  201 => 'Invalid parameter',
  205 => 'Invalid user',
  206 => 'Domain name missing',
  207 => 'Already logged in',
  208 => 'Invalid username',
  209 => 'Invalid fusername',
  210 => 'User list full',
  215 => 'User already there',
  216 => 'User already on list',
  217 => 'User not online',
  218 => 'Already in mode',
  219 => 'User is in the opposit list',
  280 => 'Switchboard failed',
  281 => 'Transfer to switchboard failed',

  300 => 'Required field missing',
  302 => 'Not logged in',

  500 => 'Internal server error',
  501 => 'Database server error',
  510 => 'File operation failed',
  520 => 'Memory allocation failed',

  600 => 'Server is busy',
  601 => 'Server is unavaliable',
  602 => 'Peer nameserver is down',
  603 => 'Database connection failed',
  604 => 'Server is going down',

  707 => 'Could not create connection',
  711 => 'Write is blocking',
  712 => 'Session is overloaded',
  713 => 'Too many active users',
  714 => 'Too many sessions',
  715 => 'Not expected',
  717 => 'Bad friend file',

  911 => 'Authentication failed',
  913 => 'Not allowed when offline',
  920 => 'Not accepting new users'
);

sub connect2msn {
  my $reconnect = shift;
  $cfg{'msn-attempt'} = $ts;

  if (!$reconnect) {
    $im{'data'}->{'msn'}->{'server'} = "messenger.hotmail.com";
    $im{'data'}->{'msn'}->{'port'}   = 1863;
    $im{'data'}->{'msn'}->{'trid'}   = 0;
    @{$im{'data'}->{'msn'}->{'lst'}} = ();
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => $im{'data'}->{'msn'}->{'server'},
    PeerPort => $im{'data'}->{'msn'}->{'port'},
    Timeout  => 3
  );

  if (!$socket) {
    &putinfo("y", "Unable to connect to MSN server %s:%d", $im{'data'}->{'msn'}->{'server'}, $im{'data'}->{'msn'}->{'port'});
    return 0;
  }

  &putinfo("y", "Connected to MSN server %s:%d", $im{'data'}->{'msn'}->{'server'}, $im{'data'}->{'msn'}->{'port'});
  $socket->autoflush(1);

  $sock{'msn'} = $socket;
  $sock{'list'}->{$socket}->{'sock'}   = $socket;
  $sock{'list'}->{$socket}->{'type'}   = "MSN";
  $sock{'list'}->{$socket}->{'host'}   = $socket->peerhost();
  $sock{'list'}->{$socket}->{'port'}   = $socket->peerport();
  $sock{'list'}->{$socket}->{'fileno'} = $socket->fileno();
  vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 1;

  &putsock($socket, "VER %d MSNP7 MSNP6 MSNP5 MSNP4 CVRO", $im{'data'}->{'msn'}->{'trid'}++);

  return 0;
}

sub msn_incoming {
  my $buffer = shift; $buffer =~ s/\s+$//;

putlog(0, "MSN: $buffer");

  my $cmd = (split(/ /, $buffer, 0))[0];
  my $key = (split(/ /, $buffer, 0))[2];

putlog(0, "MSN cmd: $cmd / key: $key");

  if (($cmd =~ /^\d{3}$/) && (exists $msn_error{$cmd})) {
    my $error = (split(/ /, $buffer, 0))[1];
    if (($error == 5) || ($error == 6)) {
      &putinfo("y", "Could not log into MSN service. Please verify that your username and password are correctly typed.");
    } else {
      &putinfo("y", "MSN error: $msn_error{$cmd}");
    }

    &del_socket($sock{'msn'});
    return 0;

  } elsif ($cmd eq "OUT") {
    my $error = (split(/ /, $buffer, 0))[1];
    if ($error eq "OTH") { &putinfo("y", "You have logged in from another computer"); }
    if ($error eq "SSD") { &putinfo("y", "Server going down for an upgrade"); }
    &del_socket($sock{'msn'});
    return 0;
  } elsif ($cmd eq "VER") {
    &putsock($sock{'msn'}, "INF %d", $im{'data'}->{'msn'}->{'trid'}++);
    return 0;
  } elsif ($cmd eq "INF") {
    &putsock($sock{'msn'}, "USR %d MD5 I %s", $im{'data'}->{'msn'}->{'trid'}++, $cfg{'msn-uid'});
    return 0;
  } elsif ($cmd eq "XFR") {
    if ($key eq "NS") {
      &del_socket($sock{'msn'});
      my $server = (split(/ /, $buffer, 0))[3];
      ($im{'data'}->{'msn'}->{'server'}, $im{'data'}->{'msn'}->{'port'}) = ((split(/:/, $server, 0))[0], (split(/:/, $server, 0))[1]);
      &connect2msn(1);
      return 0;
    }
  } elsif ($cmd eq "USR") {
    if ($key eq "MD5") {
      my $passwd = (split(/ /, $buffer, 0))[4] . $cfg{'msn-passwd'};
      &putsock($sock{'msn'}, "USR %d MD5 S %s\r", $im{'data'}->{'msn'}->{'trid'}++, md5_hex($passwd));
      return 0;
    } elsif ($key eq "OK") {
      &putsock($sock{'msn'}, "CHG %d NLN", $im{'data'}->{'msn'}->{'trid'}++);
      return 0;
    }

  } elsif ($cmd eq "ILN") {
    my $uid  = (split(/ /, $buffer, 0))[3];
    my $nick = (split(/\@/, $uid, 0))[0];

    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
      if (($im{'data'}->{'msn'}->{'list'}[$sub][0] eq $uid) && (!exists $im{'msn'}->{$uid})) {
        if ($im{'data'}->{'msn'}->{'list'}[$sub][1] ne "") { $nick = $im{'data'}->{'msn'}->{'list'}[$sub][1]; }
        &msn_adduser($uid, $nick, (split(/ /, $buffer, 0))[4], (split(/ /, $buffer, 0))[2]);
        last;
      }
    }

    return 0;
  } elsif ($cmd eq "NLN") {
    my $uid  = (split(/ /, $buffer, 0))[2];
    my $nick = (split(/\@/, $uid, 0))[0];

    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
      if (($im{'data'}->{'msn'}->{'list'}[$sub][0] eq $uid) && (!exists $im{'msn'}->{$uid})) {
        if ($im{'data'}->{'msn'}->{'list'}[$sub][1] ne "") { $nick = $im{'data'}->{'msn'}->{'list'}[$sub][1]; }
        &msn_adduser($uid, $nick, (split(/ /, $buffer, 0))[3], (split(/ /, $buffer, 0))[1]);
        last;
      }
    }

    return 0;
  } elsif ($cmd eq "FLN") {
    my $uid = (split(/ /, $buffer, 0))[1];

    if (exists $im{'msn'}->{$uid}) { &msn_killuser($uid, "Signed off"); }
    return 0;

  } elsif ($cmd eq "CHG") {
    &putsock($sock{'msn'}, "LST %d FL", $im{'data'}->{'msn'}->{'trid'}++);

    return 0;

  } elsif ($cmd eq "LST") {
    push @{$im{'data'}->{'msn'}->{'lst'}}, (split(/ /, $buffer, 0))[6];

    if ((split(/ /, $buffer, 0))[4] == (split(/ /, $buffer, 0))[5]) {
      if (scalar @{$im{'data'}->{'msn'}->{'list'}}) {
        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
          my $req = 1;
          foreach my $uid (@{$im{'data'}->{'msn'}->{'lst'}}) {
            if ($uid eq $im{'data'}->{'msn'}->{'list'}[$sub][0]) { $req = 0; }
          }
          if ($req) { &msn_requser($im{'data'}->{'msn'}->{'list'}[$sub][0]); }
        }
      }
    }

    return 0;

  } elsif ($cmd eq "CHL") {
    my $chl = md5_hex($key . "Q1P7W2E4J9R8U3S5");
    &putsock($sock{'msn'}, "QRY 10 msmsgs\@msnmsgr.com 32");
    &msn_output($sock{'msn'}, $chl);
    return 0;
  }

  return 0;
}

sub msn_output {
  my ($socket, $raw) = (shift, shift);

  print $socket $raw;
  return 0;
}

sub msn_requser {
  my $uid = shift;

  &putsock($sock{'msn'}, "ADD %d FL %s %s", $im{'data'}->{'msn'}->{'trid'}++, $uid, $uid);
  &putsock($sock{'msn'}, "ADD %d AL %s %s", $im{'data'}->{'msn'}->{'trid'}++, $uid, $uid);
  push @{$im{'data'}->{'msn'}->{'lst'}}, $uid;
  return 0;
}

sub msn_killuser {
  my ($uid, $reason) = (shift, shift);
  my $nick = $im{'msn'}->{$uid}->{'irc'};

  delete $im{'msn'}->{$uid};
  if (exists $irc{'client'}->{lc($nick)}) {
    &putsock($sock{'irc'}, ":%s QUIT :%s", $nick, $reason);
    &del_client($nick);
  }

  return 0;
}

sub msn_eraseuser {
  my $uid = shift;
  my @list;
  foreach my $uid (@{$im{'data'}->{'msn'}->{'lst'}}) {
    if ($_ eq $uid) { next; }
    push @list, $uid;
  }

  @{$im{'data'}->{'msn'}->{'lst'}} = @list;

  &msn_killuser($uid, "Deleted");

  if ($sock{'msn'}) {
    &putsock($sock{'msn'}, "REM %d FL %s", $im{'data'}->{'msn'}->{'trid'}++, $uid);
    &putsock($sock{'msn'}, "REM %d AL %s", $im{'data'}->{'msn'}->{'trid'}++, $uid);
  }

  return 0;
}

sub msn_adduser {
  my ($uid, $nick, $screen, $status) = (shift, shift, shift, shift);
  my $user = (split(/\@/, $uid, 0))[0];

  if (!$nick) { $nick = (split(/\@/, $uid, 0))[0]; }

  my $irc_nick = $cfg{'msn-nick'}; $irc_nick =~ s/\%/$nick/;
  $screen =~ s/\%20/ /g;

  $im{'msn'}->{$uid}->{'uid'}    = $uid;
  $im{'msn'}->{$uid}->{'nick'}   = $nick;
  $im{'msn'}->{$uid}->{'screen'} = $screen;
  $im{'msn'}->{$uid}->{'status'} = $status;
  $im{'msn'}->{$uid}->{'irc'}    = $irc_nick;

  if ($cfg{'server-type'} eq "hybrid") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s :%s", $irc_nick, $ts, "+i", $user, $user . ".messenger.msn.com", $cfg{'irc-fakeserv'}, $screen);
    &putsock($sock{'irc'}, ":%s MODE :%s", $irc_nick, "+i");
  } elsif ($cfg{'server-type'} eq "bahamut") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s 0 0 :%s", $irc_nick, $ts, "+i", $user, $user . ".messenger.msn.com", $cfg{'irc-fakeserv'}, $screen);
  } elsif ($cfg{'server-type'} eq "ultimate") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s 0 :%s", $irc_nick, $ts, $user, $user . ".messenger.msn.com", $cfg{'irc-fakeserv'}, $screen);
    &putsock($sock{'irc'}, "MODE %s %s", $irc_nick, "+i");
  }

  $irc{'client'}->{lc($irc_nick)}->{'nick'} = $irc_nick;
  $irc{'client'}->{lc($irc_nick)}->{'user'} = $user;
  $irc{'client'}->{lc($irc_nick)}->{'host'} = $user . ".messenger.msn.com";
  $irc{'client'}->{lc($irc_nick)}->{'real'} = $screen;
  $irc{'client'}->{lc($irc_nick)}->{'serv'} = $cfg{'irc-fakeserv'};
  $irc{'client'}->{lc($irc_nick)}->{'time'} = $ts;
  $irc{'client'}->{lc($irc_nick)}->{'mode'} |= $flags{'client'}->{'i'};

  &client_stats($irc_nick);

  if ($cfg{'msn-chan'}) {
    my $chan_ts = $ts;
    my $chan_mode;
    my $mode;

    if (exists $irc{'channel'}->{lc($cfg{'msn-chan'})}) {
      $chan_ts = $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'time'};
    } else {
      $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'channel'}  = $cfg{'msn-chan'};
      $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'time'}     = $ts;
      @{$irc{'channel'}->{lc($cfg{'msn-chan'})}->{'max'}}   = (1, $ts);
      @{$irc{'channel'}->{lc($cfg{'msn-chan'})}->{'topic'}} = (undef, undef, 0);
      $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'mode'} |= $flags{'channel'}->{'n'};
      $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'mode'} |= $flags{'channel'}->{'t'};
      $chan_mode = "nt";
    }

    $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'names'}->{lc($irc_nick)}->{'nick'} = $irc_nick;
    $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'names'}->{lc($irc_nick)}->{'mode'} = '';

    &putsock($sock{'irc'}, ":%s SJOIN %d %s +%s :%s%s", $cfg{'irc-fakeserv'}, $chan_ts, $irc{'channel'}->{lc($cfg{'msn-chan'})}->{'channel'}, $chan_mode, $mode, $irc_nick);
    &channel_stats($cfg{'msn-chan'});
  }

  return 0;
}

sub msn_msguser {
  my ($nick, $uid, $query) = (shift, shift, shift);

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $im{'msn'}->{$uid}->{'irc'}, $nick, "FXserv IRC services $VERSION");
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $im{'msn'}->{$uid}->{'irc'}, $nick, (split(/ /, $1, 0))[1]);
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $im{'msn'}->{$uid}->{'irc'}, $nick, scalar localtime());
    } elsif ($ctcp eq "DCC") {
    }

    return 0;
  }

  $query = "<" . $nick . "> " . $query;

  return 0;
}

sub msn_newnick {
  my ($uid, $nick) = (shift, shift);

  if (exists $im{'msn'}->{$uid}) {
    my $newnick = $cfg{'msn-nick'}; $newnick =~ s/\%/$nick/;
    my $oldnick = $im{'msn'}->{$uid}->{'irc'};

    &putsock($sock{'irc'}, ":%s NICK %s", $oldnick, $newnick);

    $irc{'client'}->{lc($newnick)}->{'nick'} = $newnick;
    $irc{'client'}->{lc($newnick)}->{'user'} = $irc{'client'}->{lc($oldnick)}->{'user'};
    $irc{'client'}->{lc($newnick)}->{'host'} = $irc{'client'}->{lc($oldnick)}->{'host'};
    $irc{'client'}->{lc($newnick)}->{'real'} = $irc{'client'}->{lc($oldnick)}->{'real'};
    $irc{'client'}->{lc($newnick)}->{'serv'} = $irc{'client'}->{lc($oldnick)}->{'serv'};
    $irc{'client'}->{lc($newnick)}->{'time'} = $irc{'client'}->{lc($oldnick)}->{'time'};
    $irc{'client'}->{lc($newnick)}->{'mode'} = $irc{'client'}->{lc($oldnick)}->{'mode'};;
    delete $irc{'client'}->{lc($oldnick)};

    $im{'msn'}->{$uid}->{'nick'} = $nick;
    $im{'msn'}->{$uid}->{'irc'}  = $newnick;
    return 0;
  }

  return 0;
}

1;