# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: pageserv.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %cfg %flags %im %irc %msg %services %sock $VERSION);
} else {
  our (%access, %cfg, %flags, %im, %irc, %msg, %services, %sock, $VERSION);
}

sub process_pageserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'PageServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'PageServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'PageServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'PageServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'PageServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'PageServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'PageServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'PageServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (uc($cmd) eq "HELP") {
    &pageserv_help($nick, $arg);
  } elsif (uc($cmd) eq "ICQ") {
    &pageserv_icq($nick, $arg);
  } elsif (uc($cmd) eq "YMSG") {
    &pageserv_ymsg($nick, $arg);
  } elsif (uc($cmd) eq "MSN") {
    &pageserv_msn($nick, $arg);
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
  }

  return 0;
}

sub pageserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'PageServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'PageServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/pageserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub pageserv_icq {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;
  my $online = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ICQ <ADD|DEL|LIST> [uin]");
        return 0;
      }

      my $what = (split(/ /, $arg, 0))[0];
      my $uin  = (split(/ /, $arg, 0))[1];
      my $new  = (split(/ /, $arg, 0))[2];

      if (uc($what) eq "NICK") {
        if ((!$uin) || ($uin && !$new)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ICQ NICK <uin> <nick>");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'icq'}->{'list'}[$sub][0] eq $uin) {
            $im{'data'}->{'icq'}->{'list'}[$sub][1] = $new;
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-nick'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin, $new);
            &serv_record($nick, $services{'PageServ'}->{'nick'}, "ICQ NICK $uin $new");
            &icq_newnick($uin, $new);
            return 0;
          }
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin);
        return 0;

      } elsif (uc($what) eq "ADD") {
        if ((!$uin) || ($uin !~ /^\d+$/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ICQ <ADD|DEL|LIST> [uin]");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'icq'}->{'list'}[$sub][0] eq $uin) {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-in'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin);
            return 0;
          }
        }

        push @{$im{'data'}->{'icq'}->{'list'}}, [$uin, undef];
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-add'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin);
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "ICQ ADD $uin");
        &icq_newuser($uin);
        return 0;

      } elsif (uc($what) eq "DEL") {
        if ((!$uin) || ($uin !~ /^\d+$/)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ICQ <ADD|DEL|LIST> [uin]");
          return 0;
        }

        my @list;
        my $found = 0;

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'icq'}->{'list'}[$sub][0] eq $uin) {
            $found = 1;
          } else {
            push @list, [$im{'data'}->{'icq'}->{'list'}[$sub][0], $im{'data'}->{'icq'}->{'list'}[$sub][1]];
          }
        }

        if ($found) {
          &serv_record($nick, $services{'PageServ'}->{'nick'}, "ICQ DEL $uin");
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-del'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin);
          @{$im{'data'}->{'icq'}->{'list'}} = @list;
          &icq_eraseuser($uin);
          return 0;
        } else {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uin);
          return 0;
        }
      } elsif (uc($what) eq "LIST") {
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "ICQ LIST");
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing ICQ Contacts --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[User ICQ #] [Nickname       ] [Status]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
          my $status;
          if (exists $im{'icq'}->{$im{'data'}->{'icq'}->{'list'}[$sub][0]}) { ++$online; $status = "ONLINE"; }

          &putsock($sock{'irc'}, ":%s %s %s :[%-10s] [%-15s] [%-6s]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $im{'data'}->{'icq'}->{'list'}[$sub][0], $im{'data'}->{'icq'}->{'list'}[$sub][1], $status);
          ++$idx;
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d contact%s found, %d online] --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "", $online);
        return 0;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "ICQ <ADD|DEL|LIST> [uin]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub pageserv_ymsg {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;
  my $online = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YMSG <ADD|DEL|LIST> [uid]");
        return 0;
      }

      my $what = (split(/ /, $arg, 0))[0];
      my $uid  = lc((split(/ /, $arg, 0))[1]);
      my $new  = (split(/ /, $arg, 0))[2];

      if (uc($what) eq "NICK") {
        if ((!$uid) || ($uid && !$new)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YMSG NICK <uid> <nick>");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $uid) {
            $im{'data'}->{'ymsg'}->{'list'}[$sub][1] = $new;
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-nick'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid, $new);
            &serv_record($nick, $services{'PageServ'}->{'nick'}, "YMSG NICK $uid $new");
            &ymsg_newnick($uid, $new);
            return 0;
          }
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
        return 0;

      } elsif (uc($what) eq "ADD") {
        if (!$uid) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YMSG <ADD|DEL|LIST> [uid]");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $uid) {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-in'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
            return 0;
          }
        }

        push @{$im{'data'}->{'ymsg'}->{'list'}}, [$uid, undef];
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-add'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "YMSG ADD $uid");
        &ymsg_requser($uid);
        return 0;

      } elsif (uc($what) eq "DEL") {
        if (!$uid) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YMSG <ADD|DEL|LIST> [uid]");
          return 0;
        }

        my @list;
        my $found = 0;

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $uid) {
            $found = 1;
          } else {
            push @list, [$im{'data'}->{'ymsg'}->{'list'}[$sub][0], $im{'data'}->{'ymsg'}->{'list'}[$sub][1]];
          }
        }

        if ($found) {
          &serv_record($nick, $services{'PageServ'}->{'nick'}, "YMSG DEL $uid");
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-del'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
          @{$im{'data'}->{'ymsg'}->{'list'}} = @list;
          &ymsg_eraseuser($uid);
          return 0;
        } else {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
          return 0;
        }
      } elsif (uc($what) eq "LIST") {
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "YMSG LIST");
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing YMSG Contacts --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[User ID                  ] [Nickname       ] [Status]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          my $status;
          if (exists $im{'ymsg'}->{$im{'data'}->{'ymsg'}->{'list'}[$sub][0]}) { ++$online; $status = "ONLINE"; }

          &putsock($sock{'irc'}, ":%s %s %s :[%-25s] [%-15s] [%-6s]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $im{'data'}->{'ymsg'}->{'list'}[$sub][0], $im{'data'}->{'ymsg'}->{'list'}[$sub][1], $status);
          ++$idx;
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d contact%s found, %d online] --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "", $online);
        return 0;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "YMSG <ADD|DEL|LIST> [uid]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub pageserv_msn {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;
  my $online = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "MSN <ADD|DEL|LIST> [uid]");
        return 0;
      }

      my $what = (split(/ /, $arg, 0))[0];
      my $uid  = lc((split(/ /, $arg, 0))[1]);
      my $new  = (split(/ /, $arg, 0))[2];

      if (uc($what) eq "NICK") {
        if ((!$uid) || ($uid && !$new)) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "MSN NICK <uid> <nick>");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'msn'}->{'list'}[$sub][0] eq $uid) {
            $im{'data'}->{'msn'}->{'list'}[$sub][1] = $new;
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-nick'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid, $new);
            &serv_record($nick, $services{'PageServ'}->{'nick'}, "MSN NICK $uid $new");
            &msn_newnick($uid, $new);
            return 0;
          }
        }

        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
        return 0;

      } elsif (uc($what) eq "ADD") {
        if (!$uid) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "MSN <ADD|DEL|LIST> [uid]");
          return 0;
        }

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'msn'}->{'list'}[$sub][0] eq $uid) {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-in'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
            return 0;
          }
        }

        push @{$im{'data'}->{'msn'}->{'list'}}, [$uid, undef];
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-add'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "MSN ADD $uid");
        &msn_requser($uid);
        return 0;

      } elsif (uc($what) eq "DEL") {
        if (!$uid) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "MSN <ADD|DEL|LIST> [uid]");
          return 0;
        }

        my @list;
        my $found = 0;

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'msn'}->{'list'}[$sub][0] eq $uid) {
            $found = 1;
          } else {
            push @list, [$im{'data'}->{'msn'}->{'list'}[$sub][0], $im{'data'}->{'msn'}->{'list'}[$sub][1]];
          }
        }

        if ($found) {
          &serv_record($nick, $services{'PageServ'}->{'nick'}, "MSN DEL $uid");
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-del'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
          @{$im{'data'}->{'msn'}->{'list'}} = @list;
          &msn_eraseuser($uid);
          return 0;
        } else {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'contact-unkn'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $uid);
          return 0;
        }

      } elsif (uc($what) eq "LIST") {
        &serv_record($nick, $services{'PageServ'}->{'nick'}, "MSN LIST");
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing MSN Contacts --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[User ID                  ] [Nickname       ] [Status]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
          my $status;
          if (exists $im{'msn'}->{$im{'data'}->{'msn'}->{'list'}[$sub][0]}) { ++$online; $status = "ONLINE"; }

          &putsock($sock{'irc'}, ":%s %s %s :[%-25s] [%-15s] [%-6s]", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $im{'data'}->{'msn'}->{'list'}[$sub][0], $im{'data'}->{'msn'}->{'list'}[$sub][1], $status);
          ++$idx;
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d contact%s found, %d online] --", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "", $online);
        return 0;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "MSN <ADD|DEL|LIST> [uid]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'PageServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub pageserv_refresh {
  if ($cfg{'icq-uin'}) {
    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
      if ((!exists $im{'icq'}->{$im{'data'}->{'icq'}->{'list'}[$sub][0]}) && ($sock{'icq'})) { &icq_newuser($im{'data'}->{'icq'}->{'list'}[$sub][0]); }

      if ((exists $im{'icq'}->{$im{'data'}->{'icq'}->{'list'}[$sub][0]}) && ($im{'data'}->{'icq'}->{'list'}[$sub][1] ne $im{'icq'}->{$im{'data'}->{'icq'}->{'list'}[$sub][0]}->{'nick'})) { &icq_newnick($im{'data'}->{'icq'}->{'list'}[$sub][0], $im{'data'}->{'icq'}->{'list'}[$sub][1]); }
    }

    foreach my $uin (keys %{$im{'icq'}}) {
      my $delete = 1;
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) {
        if ($uin eq $im{'data'}->{'icq'}->{'list'}[$sub][0]) { $delete = 0; }
      }

      if ($delete) {
        &putinfo("y", "ICQ: deleted $uin");
        &icq_eraseuser($uin);
      }
    }
  }

  if ($cfg{'aim-uid'}) {
  }

  if ($cfg{'ymsg-uid'}) {
    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
      if ((!exists $im{'ymsg'}->{$im{'data'}->{'ymsg'}->{'list'}[$sub][0]}) && ($sock{'ymsg'})) { &ymsg_requser($im{'data'}->{'ymsg'}->{'list'}[$sub][0]); }

      if ((exists $im{'ymsg'}->{$im{'data'}->{'ymsg'}->{'list'}[$sub][0]}) && ($im{'data'}->{'ymsg'}->{'list'}[$sub][1] ne $im{'ymsg'}->{$im{'data'}->{'ymsg'}->{'list'}[$sub][0]}->{'nick'})) { &ymsg_newnick($im{'data'}->{'ymsg'}->{'list'}[$sub][0], $im{'data'}->{'ymsg'}->{'list'}[$sub][1]); }
    }

    foreach my $uid (keys %{$im{'ymsg'}}) {
      my $delete = 1;
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
        if ($uid eq $im{'data'}->{'ymsg'}->{'list'}[$sub][0]) { $delete = 0; }
      }

      if ($delete) {
        &putinfo("y", "YMSG: deleted $uid");
        &ymsg_eraseuser($uid);
      }
    }
  }

  if ($cfg{'msn-uid'}) {
    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
      my $req = 1;
      foreach my $uid (@{$im{'data'}->{'msn'}->{'lst'}}) {
        if ($uid eq $im{'data'}->{'msn'}->{'list'}[$sub][0]) { $req = 0; }
      }
      if ($req && ($sock{'msn'})) { &msn_requser($im{'data'}->{'msn'}->{'list'}[$sub][0]); }

      if ((exists $im{'msn'}->{$im{'data'}->{'msn'}->{'list'}[$sub][0]}) && ($im{'data'}->{'msn'}->{'list'}[$sub][1] ne $im{'msn'}->{$im{'data'}->{'msn'}->{'list'}[$sub][0]}->{'nick'})) { &msn_newnick($im{'data'}->{'msn'}->{'list'}[$sub][0], $im{'data'}->{'msn'}->{'list'}[$sub][1]); }
    }

    foreach my $uid (keys %{$im{'msn'}}) {
      my $delete = 1;
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) {
        if ($uid eq $im{'data'}->{'msn'}->{'list'}[$sub][0]) { $delete = 0; }
      }

      if ($delete) {
        &putinfo("y", "MSN: deleted $uid");
        &msn_eraseuser($uid);
      }
    }
  }

  return 0;
}

1;