# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: routines.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %bot %cfg %flags %im %irc %msg %services %sock $bit $localtime $ts $VERSION);
} else {
  our (%access, %bot, %cfg, %flags, %im, %irc, %msg, %services, %sock, $bit, $localtime, $ts, $VERSION);
}

sub read_settings {
  my $rehash = shift;

  if (!open(FILE, "<" . $cfg{'settings-file'})) {
    &putlog("y", "Unable to read configuration file: %s (FATAL)", $cfg{'settings-file'});
    exit(1);
  }

  while (<FILE>) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;

    if (($data =~ /^#/) || ($data !~ /\S/)) { next; }
    my ($key, $value) = split(/ /, $data, 2);

    if ($key eq "log-event") {
      if ($rehash) { delete $cfg{'log-event'}; }
      foreach (split(//, $value)) {
        if ($flags{'event'}->{$_}) { $cfg{'log-event'} |= $flags{'event'}->{$_}; }
      }
    } elsif (($key eq "DataServ") && !$rehash) {
      $services{'DataServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'DataServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'DataServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'DataServ'}->{'bot'}  = "DataServ";
    } elsif (($key eq "FindServ") && !$rehash) {
      $services{'FindServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'FindServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'FindServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'FindServ'}->{'bot'}  = "FindServ";
    } elsif (($key eq "MobiServ") && !$rehash) {
      $services{'MobiServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'MobiServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'MobiServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'MobiServ'}->{'bot'}  = "MobiServ";
    } elsif (($key eq "ScanServ") && !$rehash) {
      $services{'ScanServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'ScanServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'ScanServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'ScanServ'}->{'bot'}  = "ScanServ";
    } elsif (($key eq "FileServ") && !$rehash) {
      $services{'FileServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'FileServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'FileServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'FileServ'}->{'bot'}  = "FileServ";
      $services{'FileServ'}->{'traffic'} = 0;
      $services{'FileServ'}->{'total'}   = 0;
    } elsif (($key eq "PageServ") && !$rehash) {
      $services{'PageServ'}->{'nick'} = (split(/ /, $value, 0))[0];
      $services{'PageServ'}->{'user'} = (split(/ /, $value, 0))[1];
      $services{'PageServ'}->{'real'} = (split(/ /, $value, 3))[2];
      $services{'PageServ'}->{'bot'}  = "PageServ";

    } elsif (($key eq "fserv-root") || ($key eq "help-path")) {
      if ($value =~ /\/$/) { $value = substr($value, 0, length($value) - 1); }
      $cfg{$key} = $value;

    } elsif (($key eq "dcc-listen") || ($key eq "bot-listen")) {
      my $listen = IO::Socket::INET->new(
        Proto     => "tcp",
        LocalPort => $value,
        Listen    => 1,
        Reuse     => 1
      );

      if (!$listen) {
        &putlog("y", "Unable to listen port %d", $value);
      } else {
        $sock{'list'}->{$listen}->{'sock'} = $listen;

        if ($key eq "bot-listen") {
          $sock{'list'}->{$listen}->{'type'} = "BOT-A";
        } elsif ($key eq "dcc-listen") {
          $sock{'list'}->{$listen}->{'type'} = "DCC-A";
        }

        $sock{'list'}->{$listen}->{'fileno'} = $listen->fileno();
        vec($bit, $sock{'list'}->{$listen}->{'fileno'}, 1) = 1;
      }

      $cfg{$key} = $value;

    } elsif ($rehash && $cfg{'irc-connected'} && (($key eq "hub-server") || ($key eq "irc-fakeserv") || ($key eq "irc-fakeinfo"))) {
      next;
    } else {
      $cfg{$key} = $value;
    }
  }

  close(FILE);
  return 0;
}

sub read_access {
  my $rehash = shift;

  if (!open(FILE, "<" . $cfg{'access-file'})) {
    &putlog("y", "Unable to read access file: %s (FATAL)", $cfg{'access-file'});
    exit(1);
  }

  if ($rehash) { undef(%access); }

  while (<FILE>) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;
    if (($data =~ /^#/) || ($data !~ /\S/)) { next; }

    if ($data =~ /^O:([^:]+):([^:]+):([^:]+)$/) {
      $access{lc($1)}->{'login'}  = $1;
      $access{lc($1)}->{'passwd'} = $2;
      $access{lc($1)}->{'event'}  = '';
      foreach (split(//, $3)) {
        if ($flags{'access'}->{$_}) { $access{lc($1)}->{'flags'} |= $flags{'access'}->{$_}; }
      }

    } elsif ($data =~ /^B:([^:]+):([^:]+):([^:]+):(\d+)$/) {
      $bot{lc($2)}->{'host'}   = $1;
      $bot{lc($2)}->{'nick'}   = $2;
      $bot{lc($2)}->{'passwd'} = $3;
      $bot{lc($2)}->{'port'}   = $4;

    } elsif ($data =~ /^E:([^:]+):([^:]+)$/) {
      $irc{'noscan'}->{lc($1)}->{'hostmask'} = $1;
      $irc{'noscan'}->{lc($1)}->{'servmask'} = $2;
      $irc{'noscan'}->{lc($1)}->{'new'}      = 0;

    } elsif ($data =~ /^I:([^:]+)$/) {
      $irc{'ignore'}->{lc($1)}->{'hostmask'} = (split(/\@/, $1, 0))[0];
      $irc{'ignore'}->{lc($1)}->{'usermask'} = (split(/\@/, $1, 0))[1];
      $irc{'ignore'}->{lc($1)}->{'new'}      = 0;
    } elsif ($data =~ /^F:([^:]+)$/) {
      $irc{'nofserv'}->{lc($1)}->{'hostmask'} = (split(/\@/, $1, 0))[0];
      $irc{'nofserv'}->{lc($1)}->{'usermask'} = (split(/\@/, $1, 0))[1];
      $irc{'nofserv'}->{lc($1)}->{'new'}      = 0;
    } elsif ($data =~ /^C:([^:]+):([^:]+):([^:]+)?:([^:]+)?$/) {
      foreach (keys %services) {
        if ($services{$_}->{'nick'} eq $1) {
          my $chan = $2;
          my $joinmsg = $3;
          my $partmsg = $4;
          my $mode;

          if (substr($chan, 0, 1) eq "\@") {
            $mode = "o";
          } elsif (substr($chan, 0, 1) eq "\%") {
            $mode = "h";
          } elsif (substr($chan, 0, 1) eq "+") {
            $mode = "v";
          }

          $chan =~ s/^\@//; $chan =~ s/^\%//; $chan =~ s/^\+//;

          $services{$_}->{'channel'}->{lc($chan)}->{'channel'} = $chan;
          $services{$_}->{'channel'}->{lc($chan)}->{'mode'}    = $mode;
          $services{$_}->{'channel'}->{lc($chan)}->{'joinmsg'} = $joinmsg;
          $services{$_}->{'channel'}->{lc($chan)}->{'partmsg'} = $partmsg;
        }
      }
    }
  }

  close(FILE);
  return 0;
}

sub read_glines {
  my $rehash = shift;

  if (!open(FILE, "<" . $cfg{'gline-file'})) {
    &putlog("y", "Unable to read G-lines file: %s", $cfg{'gline-file'});
    return 0;
  }

  while (<FILE>) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;
    if (($data =~ /^#/) || ($data !~ /\S/)) { next; }

    if ($data =~ /^G:([^:]+):(\d+):(\d+):([^:]+):([^:]+)?$/) {
      $irc{'glined'}->{lc($1)}->{'mask'}     = $1;
      $irc{'glined'}->{lc($1)}->{'usermask'} = (split(/\@/, $1, 0))[0];
      $irc{'glined'}->{lc($1)}->{'hostmask'} = (split(/\@/, $1, 0))[1];
      $irc{'glined'}->{lc($1)}->{'set'}      = $2;
      $irc{'glined'}->{lc($1)}->{'expire'}   = $3;
      $irc{'glined'}->{lc($1)}->{'author'}   = $4;
      $irc{'glined'}->{lc($1)}->{'reason'}   = $5;
      $irc{'glined'}->{lc($1)}->{'new'}      = 0;

      if (($irc{'glined'}->{lc($1)}->{'expire'} > 0) && ($ts - $irc{'glined'}->{lc($1)}->{'set'} > $irc{'glined'}->{lc($1)}->{'expire'})) { delete $irc{'glined'}->{lc($1)}; }
    }
  }

  close(FILE);
  return 0;
}

sub read_db {
  if (!open(FILE, "<" . $cfg{'dataserv-db'})) {
    &putlog("y", "Unable to read db file: %s", $cfg{'dataserv-db'});
    return 0;
  }

  while (<FILE>) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;
    if (($data =~ /^#/) || ($data !~ /\S/)) { next; }

    if ($data =~ /^O:([^:]+):([^:]+)$/) {
      if (exists $access{lc($1)}) {
        foreach (split(//, $2)) {
          if ($flags{'event'}->{$_}) { $access{lc($1)}->{'event'} |= $flags{'event'}->{$_}; }
        }
      }
    }
  }

  close(FILE);

  if (exists $services{'PageServ'}) {
    if (!open(FILE, "<" . $cfg{'pageserv-db'})) {
      &putlog("y", "Unable to read db file: %s", $cfg{'pageserv-db'});
      return 0;
    }

    if ($cfg{'icq-uin'}) { @{$im{'data'}->{'icq'}->{'list'}} = (); }
    if ($cfg{'ymsg-uid'}) { @{$im{'data'}->{'ymsg'}->{'list'}} = (); }
    if ($cfg{'aim-uid'}) { @{$im{'data'}->{'aim'}->{'list'}} = (); }
    if ($cfg{'msn-uid'}) { @{$im{'data'}->{'msn'}->{'list'}} = (); }

    while (<FILE>) {
      my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;
      if (($data =~ /^#/) || ($data !~ /\S/)) { next; }

      if ($data =~ /^ICQ:(\d+):(\S+)?$/) {
        push @{$im{'data'}->{'icq'}->{'list'}}, [$1, $2];
      } elsif ($data =~ /^AIM:(\S+):(\S+)?$/) {
        push @{$im{'data'}->{'aim'}->{'list'}}, [$1, $2];
      } elsif ($data =~ /^YMSG:(\S+):(\S+)?$/) {
        push @{$im{'data'}->{'ymsg'}->{'list'}}, [$1, $2];
      } elsif ($data =~ /^MSN:(\S+):(\S+)?$/) {
        push @{$im{'data'}->{'msn'}->{'list'}}, [$1, $2];
      }
    }

    close(FILE);
  }

  if (!open(FILE, "<" . $cfg{'seen-file'})) {
    &putlog("y", "Unable to read seen file: %s", $cfg{'seen-file'});
    return 0;
  }

  while (<FILE>) {
    my $data = $_; $data =~ tr/ //s; $data =~ s/\s+$//; $data =~ s/^\s+//;
    if (($data =~ /^\#/) || ($data !~ /\S/)) { next; }

    my ($nick, $time, $user, $host, $what, $xtra) = split(/ /, $data, 6); 

    if ($nick && $time && $user && $host && $what) {
      $irc{'seen'}->{lc($nick)}->{'nick'} = $nick;
      $irc{'seen'}->{lc($nick)}->{'time'} = $time;
      $irc{'seen'}->{lc($nick)}->{'user'} = $user;
      $irc{'seen'}->{lc($nick)}->{'host'} = $host;
      $irc{'seen'}->{lc($nick)}->{'what'} = $what;
      $irc{'seen'}->{lc($nick)}->{'xtra'} = $xtra;
    }
  }

  close(FILE);
  return 0;
}

sub save_db {
  my $idx  = 0;
  my @seen = ();
  my @file = ();
  my $umodes;

  if (!open(FILE, ">" . $cfg{'dataserv-db'})) {
    &putinfo("y", "Unable to write database file: %s", $cfg{'dataserv-db'});
    return 0;
  }

  print FILE "# This file is automatically generated.\n";
  print FILE "# Do not modify the contents.\n";
  print FILE "\n";

  foreach (keys %access) {
    foreach my $umode (keys %{$flags{'event'}}) {
      if ($access{$_}->{'event'} & $flags{'event'}->{$umode}) { $umodes .= $umode; }
    }

    if ($umodes) { print FILE sprintf("O:%s:%s\n", $access{$_}->{'login'}, $umodes); }
    undef($umodes);
  }

  print FILE "\n";
  print FILE "# EOF\n";
  close(FILE);


  if (exists $services{'PageServ'}) {
    if (!open(FILE, ">" . $cfg{'pageserv-db'})) {
      &putinfo("y", "Unable to write database file: %s", $cfg{'pageserv-db'});
      return 0;
    }

    print FILE "# This file is automatically generated.\n";
    print FILE "# Do not modify the contents.\n";
    print FILE "\n";

    if ($cfg{'icq-uin'}) {
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'icq'}->{'list'}}; $sub++) { print FILE sprintf("ICQ:%d:%s\n", $im{'data'}->{'icq'}->{'list'}[$sub][0], $im{'data'}->{'icq'}->{'list'}[$sub][1]); }
    }

    if ($cfg{'aim-uid'}) {
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'aim'}->{'list'}}; $sub++) { print FILE sprintf("AIM:%s:%s\n", $im{'data'}->{'aim'}->{'list'}[$sub][0], $im{'data'}->{'aim'}->{'list'}[$sub][1]); }
    }

    if ($cfg{'ymsg-uid'}) {
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) { print FILE sprintf("YMSG:%s:%s\n", $im{'data'}->{'ymsg'}->{'list'}[$sub][0], $im{'data'}->{'ymsg'}->{'list'}[$sub][1]); }
    }

    if ($cfg{'msn-uid'}) {
      for (my $sub = 0; $sub < scalar @{$im{'data'}->{'msn'}->{'list'}}; $sub++) { print FILE sprintf("MSN:%s:%s\n", $im{'data'}->{'msn'}->{'list'}[$sub][0], $im{'data'}->{'msn'}->{'list'}[$sub][1]); }
    }

    print FILE "\n";
    print FILE "# EOF\n";
    close(FILE);
  }



  if (!open(FILE, ">" . $cfg{'seen-file'})) {
    &putinfo("y", "Unable to write seen file: %s", $cfg{'seen-file'});
    return 0;
  }

  print FILE "# This file is automatically generated.\n";
  print FILE "# Do not modify the contents.\n";
  print FILE "\n";

  if (scalar keys %{$irc{'seen'}} > $cfg{'seen-maxrec'}) {
    @seen = sort { $irc{'seen'}->{$b}->{'time'} <=> $irc{'seen'}->{$a}->{'time'} } keys %{$irc{'seen'}};
    foreach (@seen) {
      print FILE sprintf("%s %d %s %s %s %s\n", $irc{'seen'}->{$_}->{'nick'}, $irc{'seen'}->{$_}->{'time'}, $irc{'seen'}->{$_}->{'user'}, $irc{'seen'}->{$_}->{'host'}, $irc{'seen'}->{$_}->{'what'}, $irc{'seen'}->{$_}->{'xtra'});
      if (++$idx > $cfg{'seen-maxrecs'}) { last; }
    }

  } else {
    foreach (keys %{$irc{'seen'}}) { print FILE sprintf("%s %d %s %s %s %s\n", $irc{'seen'}->{$_}->{'nick'}, $irc{'seen'}->{$_}->{'time'}, $irc{'seen'}->{$_}->{'user'}, $irc{'seen'}->{$_}->{'host'}, $irc{'seen'}->{$_}->{'what'}, $irc{'seen'}->{$_}->{'xtra'}); }
  }

  print FILE "\n";
  print FILE "# EOF\n";
  close(FILE);

  if (!open(FILE, ">>" . $cfg{'access-file'})) {
    &putinfo("y", "Unable to write access file: %s", $cfg{'access-file'});
    return 0;
  }

  foreach (keys %{$irc{'ignore'}}) {
    if ($irc{'ignore'}->{$_}->{'new'}) {
      print FILE sprintf("I:%s\@%s\n", $irc{'ignore'}->{$_}->{'usermask'}, $irc{'ignore'}->{$_}->{'hostmask'});
      $irc{'ignore'}->{$_}->{'new'} = 0;
    }
  }

  foreach (keys %{$irc{'noscan'}}) {
    if ($irc{'noscan'}->{$_}->{'new'}) {
      print FILE sprintf("E:%s:%s\n", $irc{'noscan'}->{$_}->{'hostmask'}, $irc{'noscan'}->{$_}->{'servmask'});
      $irc{'noscan'}->{$_}->{'new'} = 0;
    }
  }

  foreach (keys %{$irc{'nofserv'}}) {
    if ($irc{'nofserv'}->{$_}->{'new'}) {
      print FILE sprintf("F:%s\@%s\n", $irc{'nofserv'}->{$_}->{'usermask'}, $irc{'nofserv'}->{$_}->{'hostmask'});
      $irc{'nofserv'}->{$_}->{'new'} = 0;
    }
  }

  close(FILE);

  if (!open(FILE, ">" . $cfg{'gline-file'})) {
    &putinfo("y", "Unable to write G-lines file: %s", $cfg{'gline-file'});
    return 0;
  }

  print FILE "# This file is automatically generated.\n";
  print FILE "# Do not modify the contents.\n";
  print FILE "\n";

  foreach (keys %{$irc{'glined'}}) {
    if ($irc{'glined'}->{$_}->{'new'}) {
      print FILE sprintf("G:%s:%d:%d:%s:%s\n", $irc{'glined'}->{$_}->{'mask'}, $irc{'glined'}->{$_}->{'set'}, $irc{'glined'}->{$_}->{'expire'}, $irc{'glined'}->{$_}->{'author'}, $irc{'glined'}->{$_}->{'reason'});
      $irc{'glined'}->{$_}->{'new'} = 0;
    }
  }

  print FILE "\n";
  print FILE "# EOF\n";
  close(FILE);

  &putinfo("y", "Database has been saved");
  $cfg{'last-sync'} = $ts;
  return 0;
}

sub send_motd {
  my ($socket, $file) = (shift, shift);

  if (!open(FILE, "<" . $file)) {
    &putlog("y", "Unable to open MOTD file: %s", $file);
    &putsock($socket, "MOTD is missing...");
    return 0;
  }

  while (<FILE>) {
    my $buffer = &substitute($_); $buffer =~ s/\s+$//;
    &putsock($socket, $buffer);
  }

  close(FILE);
  return 0;
}

sub serv_record {
  my ($nick, $serv, $arg) = (shift, shift, shift);

  &putinfo("s", "%s: %s!%s\@%s %s", $serv, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $arg);
  return 0;
}

sub putlog {
  my ($flag, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  if ($flags{'event'}->{$flag} && !($cfg{'log-event'} & $flags{'event'}->{$flag})) { return 0; }

  if (!open(FILE, ">>" . $cfg{'log-file'})) { return 0; }

  print FILE sprintf("[%s] %s\n", $localtime, $buffer);
  close(FILE);
  return 0;
}

sub substitute {
  my $string = shift;

  $string =~ s/\%D/$services{'DataServ'}->{'nick'}/g;
  if (exists $services{'FileServ'}) { $string =~ s/\%X/$services{'FileServ'}->{'nick'}/g; }
  if (exists $services{'FindServ'}) { $string =~ s/\%F/$services{'FindServ'}->{'nick'}/g; }
  if (exists $services{'MobiServ'}) { $string =~ s/\%M/$services{'MobiServ'}->{'nick'}/g; }
  if (exists $services{'PageServ'}) { $string =~ s/\%P/$services{'PageServ'}->{'nick'}/g; }
  if (exists $services{'ScanServ'}) { $string =~ s/\%S/$services{'ScanServ'}->{'nick'}/g; }
  $string =~ s/\%a/$cfg{'admin-info'}/g;
  $string =~ s/\%b/\002/g;
  $string =~ s/\%t/$localtime/g;
  $string =~ s/\%v/$VERSION/g;
  return $string;
}

sub restart {
  &save_db();
  &putinfo("y", "Restarting services...");
  &disconnect("Restarting services...");
  if (!unlink($cfg{'pid-file'})) { &putlog("y", "Unable to delete pid file: %s", $cfg{'pid-file'}); }
  exec("perl $0");
  exit(0);
}

sub rehash {
  if (scalar keys %bot) {
    foreach (keys %bot) {
      if (!$bot{$_}->{'sock'}) { delete $bot{$_}; }
    }
  }

  if (exists $irc{'ignore'}) {
    foreach (keys %{$irc{'ignore'}}) {
      if (!$irc{'ignore'}->{$_}->{'new'}) { delete $irc{'ignore'}->{$_}; }
    }
  }

  if (exists $irc{'glined'}) {
    foreach (keys %{$irc{'glined'}}) {
      if (!$irc{'glined'}->{$_}->{'new'}) { delete $irc{'glined'}->{$_}; }
    }
  }

  if (exists $irc{'noscan'}) {
    foreach (keys %{$irc{'noscan'}}) {
      if (!$irc{'noscan'}->{$_}->{'new'}) { delete $irc{'noscan'}->{$_}; }
    }
  }

  if (exists $irc{'nofserv'}) {
    foreach (keys %{$irc{'nofserv'}}) {
      if (!$irc{'nofserv'}->{$_}->{'new'}) { delete $irc{'nofserv'}->{$_}; }
    }
  }

  if (exists $sock{'list'}) {
    foreach (keys %{$sock{'list'}}) {
      if (($sock{'list'}->{$_}->{'type'} eq "DCC-A") || ($sock{'list'}->{$_}->{'type'} eq "BOT-A")) { &del_socket($sock{'list'}->{$_}->{'sock'}); }
    }
  }

  foreach (keys %services) {
    if (exists $services{$_}->{'channel'}) { delete $services{$_}->{'channel'}; }
  }

  if (exists $services{'PageServ'}) {
    if ($cfg{'icq-uin'}) { @{$im{'data'}->{'icq'}->{'list'}} = (); }
    if ($cfg{'aim-uid'}) { @{$im{'data'}->{'aim'}->{'list'}} = (); }
    if ($cfg{'ymsg-uid'}) { @{$im{'data'}->{'ymsg'}->{'list'}} = (); }
    if ($cfg{'msn-uid'}) { @{$im{'data'}->{'msn'}->{'list'}} = (); }
  }

  require "lang/" . $cfg{'lang-pack'};

  &read_settings(1);
  &read_access(1);
  &read_glines();
  &read_db();

  if (exists $services{'PageServ'}) { &pageserv_refresh(); }

  foreach (keys %services) { &check_channel($services{$_}->{'bot'}); }

  if (exists $irc{'client'}) {
    foreach (keys %{$irc{'client'}}) {
      if ($irc{'client'}->{$_}->{'access'} && !exists $access{lc($irc{'client'}->{$_}->{'access'})}) { $irc{'client'}->{$_}->{'access'} = ''; }
    }
  }

  if (exists $sock{'list'}) {
    foreach (keys %{$sock{'list'}}) {
      if (($sock{'list'}->{$_}->{'type'} eq "DCC") && $sock{'list'}->{$_}->{'access'} && !exists $access{lc($sock{'list'}->{$_}->{'access'})}) { &del_socket($sock{'list'}->{$_}->{'sock'}); }
    }
  }

  return 1;
}

sub terminate {
  my ($save, $reason) = (shift, shift);

  &disconnect($reason);

  foreach (keys %{$sock{'list'}}) {
    vec($bit, $sock{'list'}->{$_}->{'fileno'}, 1) = 0;
    $sock{'list'}->{$_}->{'sock'}->close();
    delete $sock{'list'}->{$_};
  }

  if ($save) { &save_db(); }
  &putinfo("y", "Shutting down services %s", $reason ? "\(" . $reason . "\)" : "...");
  if (!unlink($cfg{'pid-file'})) { &putlog("y", "Unable to delete pid file: %s", $cfg{'pid-file'}); }
  exit(0);
}

sub format_timestr {
  my $str = shift;
  my $sec = 0;
  my $tmp = 0;
  my $output;

  foreach (split(//, $str)) {
    if ($_ =~ /\d/) {
      $tmp .= $_;
    } else {
      if (lc($_) eq "w") {
        $sec += $tmp * 604800;
      } elsif (lc($_) eq "d") {
        $sec += $tmp * 86400;
      } elsif (lc($_) eq "h") {
        $sec += $tmp * 3600;
      } elsif (lc($_) eq "m") {
        $sec += $tmp * 60;
      } elsif (lc($_) eq "s") {
        $sec += $tmp;
      } else {
        $sec = 0;
      }

      undef($tmp);
    }
  }

  $output = $sec;
  return $output;
}

sub format_date {
  my $date = shift;
  my $output;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($date);
  $year = sprintf("%02d", $year % 100);

  $output = sprintf("%02d/%02d/%02d %02d:%02d", $mday, ++$mon, $year, $hour, $min);
  return $output;
}

sub format_size {
  my $bytes = shift;
  my $output;

  if ($bytes > 1073741824) {
    $output  = sprintf("%.2f", $bytes / 1073741824);
    $output .= "Gb";
  } elsif ($bytes > 1048576) {
    $output  = sprintf("%.2f", $bytes / 1048576);
    $output .= "Mb";
  } elsif ($bytes > 1024) {
    $output  = sprintf("%.2f", $bytes / 1024);
    $output .= "Kb";
  } else {
    $output  = sprintf("%.2f", $bytes);
    $output .= "b";
  }

  return $output;
}

sub format_time {
  my $time = shift;
  my $output;

  my $diff = abs($ts - $time);

  my $wks = int($diff / 604800); $diff %= 604800;
  my $day = int($diff / 86400);  $diff %= 86400;
  my $hrs = int($diff / 3600);   $diff %= 3600;
  my $min = int($diff / 60);     $diff %= 60;

  if ($wks) { $output .= $wks . "w "; }
  if ($day) { $output .= $day . "d "; }
  if ($hrs) { $output .= $hrs . "h "; }
  if ($min) { $output .= $min . "m "; }
  if (!$output) { $output .= $diff . "s"; }
  $output =~ s/\s+$//;

  return $output;
}

sub plain2regexp {
  my $plain = shift;

  $plain = quotemeta($plain);
  $plain =~ s/\\\*/.*/g;
  $plain =~ s/\\\?/./g;

  return '^' . $plain . '$';
}

sub int2bytes {
  my ($bytes, $value) = (shift, shift);
  my @result;

  if (!$value) { $value = 0; }

  for (my $idx = 0; $idx < $bytes; $idx++) { unshift @result, ($value >> ($idx * 8) & 0xFF); }
  return @result;
}

sub int2endian {
  my ($bytes, $value) = (shift, shift);
  my @result;

  for (my $idx = 0; $idx < $bytes; $idx++) { push @result, ($value >> ($idx * 8) & 0xFF); }
  return @result;
}

sub str2bytes {
  my $string = shift;
  my @result;

  foreach (split(//, $string)) { push @result, ord($_); }
  return @result;
}

sub bytes2int {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result = 0;

  for (my $idx = $start; $idx < $start + $bytes; $idx++) {
    $result <<= 8;
    $result |= ($array->[$idx] || 0);
  }

  return $result;
}

sub endian2int {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result = 0;

  for (my $idx = $start + $bytes - 1; $idx >= $start; $idx--) {
    $result <<= 8;
    $result |= ($array->[$idx] || 0);
  }

  return $result;
}

sub bytes2str {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result;

  for (my $idx = $start; $idx < $start + $bytes; $idx++) { $result .= $array->[$idx] ? chr($array->[$idx]) : ''; }
  return $result;
}

sub base64 {
  my ($v, $n) = (shift, shift);
  my $result = '';
  my $salt   = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

  while (--$n >= 0) {
    $result .= substr($salt, $v & 0x3f, 1);
    $v >>= 6;
  }

  return $result;
}

1;