# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: scanserv.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %cfg %flags %irc %msg %services %sock $ts $TMP $VERSION);
} else {
  our (%access, %cfg, %flags, %irc, %msg, %services, %sock, $ts, $TMP, $VERSION);
}

sub process_scanserv {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return 0; }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'ScanServ'}->{'nick'}, $nick, "FXserv IRC services $VERSION");
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $services{'ScanServ'}->{'nick'}, $nick, "Administrator: $cfg{'admin-info'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'ScanServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $services{'ScanServ'}->{'nick'}, $nick, (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'ScanServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $services{'ScanServ'}->{'nick'}, $nick, scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'ScanServ'}->{'nick'}, $ctcp, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $services{'ScanServ'}->{'nick'}, $1, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'});
    }

    return 0;
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if (uc($cmd) eq "HELP") {
    &scanserv_help($nick, $arg);
  } elsif (uc($cmd) eq "SCAN") {
    &scanserv_scan($nick, $arg);
  } elsif (uc($cmd) eq "EXCEPT") {
    &scanserv_except($nick, $arg);
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'unkn-cmd'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $cmd);
  }

  return 0;
}

sub scanserv_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &serv_record($nick, $services{'ScanServ'}->{'nick'}, "HELP");
    $arg = "index";
  } else {
    &serv_record($nick, $services{'ScanServ'}->{'nick'}, "HELP $arg");
  }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = $cfg{'help-path'} . "/scanserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-help'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $arg);
    return 0;
  }

  while (<FILE>) {
    my $help = &substitute($_);

    if ($help =~ /^\%\+a/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+a//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } elsif ($help =~ /^\%\+o/) {
      if ($irc{'client'}->{lc($nick)}->{'access'}) {
        $help =~ s/^\%\+o//;
        my $login = $irc{'client'}->{lc($nick)}->{'access'};
        if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) { &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help); }
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $help);
    }
  }

  close(FILE);
  return 0;
}

sub scanserv_check {
  my $nick   = shift;
  my $clones = 0;
  my $type;

  if (!exists $irc{'client'}->{lc($nick)}) { return 0; }

  foreach (keys %{$irc{'noscan'}}) {
    my $hmaskrx = &plain2regexp($irc{'noscan'}->{$_}->{'hostmask'});
    my $smaskrx = &plain2regexp($irc{'noscan'}->{$_}->{'servmask'});

    if (($irc{'client'}->{lc($nick)}->{'host'} =~ /$hmaskrx/i) && ($irc{'client'}->{lc($nick)}->{'serv'} =~ /$smaskrx/i)) {
      &putlog("y", "%s: skipping scan for %s!%s\@%s [%s] (E-lined)", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $irc{'noscan'}->{$_}->{'servmask'});
      return 0;
    }
  }

  foreach (keys %{$irc{'client'}}) {
    if ($irc{'client'}->{$_}->{'host'} eq $irc{'client'}->{lc($nick)}->{'host'}) { ++$clones; }
  }

  if ($clones > $cfg{'max-clones'}) {
    &putinfo("y", "Clone detected: %s (%s\@%s) [%s]", $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $irc{'client'}->{lc($nick)}->{'serv'});
    if ($cfg{'kill-clones'} && (lc($cfg{'kill-clones'}) eq "yes")) {

      if ($cfg{'server-type'} eq "hybrid") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (No more than %d users with the same host)", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'ScanServ'}->{'nick'}, $cfg{'max-clones'});
      } elsif ($cfg{'server-type'} eq "bahamut") {
        &putsock($sock{'irc'}, ":%s KILL %s :No more than %d users with the same host", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $cfg{'max-clones'});
      } elsif ($cfg{'server-type'} eq "ultimate") {
        &putsock($sock{'irc'}, ":%s KILL %s :No more than %d users with the same host", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $cfg{'max-clones'});
      }

      ++$irc{'stats'}->{'kills'}[0];
      &del_client($nick);
      return 1;
    }
  }


  if (my $pid = fork) {
    if ($msg{'scan-1'}) { &putsock($sock{'irc'}, ":%s NOTICE %s :%s", $services{'ScanServ'}->{'nick'}, $nick, $msg{'scan-1'}); }
    if ($msg{'scan-2'}) { &putsock($sock{'irc'}, ":%s NOTICE %s :%s", $services{'ScanServ'}->{'nick'}, $nick, $msg{'scan-2'}); }

    foreach (split(/ /, $cfg{'scan-wingate'})) {
      if (&scanserv_sock($irc{'client'}->{lc($nick)}->{'host'}, $_, "WGATE")) { $type = "Wingate"; }
    }

    foreach (split(/ /, $cfg{'scan-http'})) {
      if (&scanserv_sock($irc{'client'}->{lc($nick)}->{'host'}, $_, "HTTP")) { $type = "HTTP"; }
    }

    foreach (split(/ /, $cfg{'scan-socks4'})) {
      if (&scanserv_sock($irc{'client'}->{lc($nick)}->{'host'}, $_, "SOCKS4")) { $type = "SOCKS4"; }
    }

    foreach (split(/ /, $cfg{'scan-socks5'})) {
      if (&scanserv_sock($irc{'client'}->{lc($nick)}->{'host'}, $_, "SOCKS5")) { $type = "SOCKS5"; }
    }

    if ($type) {
      &putinfo("y", "Proxy/Wingate detected (%s): %s (%s\@%s) [%s]", $type, $irc{'client'}->{lc($nick)}->{'nick'}, $irc{'client'}->{lc($nick)}->{'user'}, $irc{'client'}->{lc($nick)}->{'host'}, $irc{'client'}->{lc($nick)}->{'serv'});
      print $TMP sprintf("%s %s %s %s %d %d %s\n", "*", $irc{'client'}->{lc($nick)}->{'host'}, $irc{'client'}->{lc($nick)}->{'nick'}, $services{'ScanServ'}->{'nick'}, $ts, 0, $msg{'scan-gline'});

      if ($cfg{'server-type'} eq "hybrid") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s!%s (%s)", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $cfg{'irc-fakeserv'}, $services{'ScanServ'}->{'nick'}, $msg{'scan-gline'});
      } elsif ($cfg{'server-type'} eq "bahamut") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $msg{'scan-gline'});
      } elsif ($cfg{'server-type'} eq "ultimate") {
        &putsock($sock{'irc'}, ":%s KILL %s :%s", $services{'ScanServ'}->{'nick'}, $irc{'client'}->{lc($nick)}->{'nick'}, $msg{'scan-gline'});
      }

    } else {
      &putsock($sock{'irc'}, ":%s NOTICE %s :%s", $services{'ScanServ'}->{'nick'}, $nick, $msg{'scan-passed'});
    }

    exit(0);
  }

  return 0;
}

sub scanserv_scan {
  my ($nick, $target) = (shift, shift);
  my $type;
  my $host;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (exists $irc{'client'}->{lc($target)}) {
        $host = $irc{'client'}->{lc($target)}->{'host'};
      } elsif ($target =~ /\./) {
        $host = $target;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "SCAN <nick|host>");
        return 0;
      }

      if (my $pid = fork) {
        foreach (split(/ /, $cfg{'scan-wgate'})) {
          if (&scanserv_sock($host, $_, "WGATE")) { $type = "Wingate"; }
        }

        foreach (split(/ /, $cfg{'scan-http'})) {
          if (&scanserv_sock($host, $_, "HTTP")) { $type = "HTTP"; }
        }

        foreach (split(/ /, $cfg{'scan-socks4'})) {
          if (&scanserv_sock($host, $_, "SOCKS4")) { $type = "SOCKS4"; }
        }

        foreach (split(/ /, $cfg{'scan-socks5'})) {
          if (&scanserv_sock($host, $_, "SOCKS5")) { $type = "SOCKS5"; }
        }

        if ($type) {
          if (exists $irc{'client'}->{lc($target)}) {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'scan-unsecure'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "$irc{'client'}->{lc($target)}->{'nick'} ($irc{'client'}->{lc($target)}->{'user'}\@$irc{'client'}->{lc($target)}->{'host'})", $type);
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'scan-unsecure'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $target, $type);
          }
        } else {
          if (exists $irc{'client'}->{lc($target)}) {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'scan-clean'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "$irc{'client'}->{lc($target)}->{'nick'} ($irc{'client'}->{lc($target)}->{'user'}\@$irc{'client'}->{lc($target)}->{'host'})");
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'scan-clean'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $target);
          }
        }
        exit(0);
      }

      &serv_record($nick, $services{'ScanServ'}->{'nick'}, "SCAN $target");
      return 0;
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub scanserv_except {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;

  if ($irc{'client'}->{lc($nick)}->{'access'}) {
    my $login = $irc{'client'}->{lc($nick)}->{'access'};
    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      if (!$arg) {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "EXCEPT <ADD|DEL|LIST> [mask] [server]");
        return 0;
      }

      my $what = (split(/ /, $arg, 0))[0];
      my $mask = (split(/ /, $arg, 0))[1];
      my $serv = (split(/ /, $arg, 0))[2];

      if (uc($what) eq "ADD") {
        if (!$mask) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "EXCEPT <ADD|DEL|LIST> [mask] [server]");
          return 0;
        } else {
          if (!exists $irc{'noscan'}->{lc($mask)}) {
            $irc{'noscan'}->{lc($mask)}->{'hostmask'} = $mask;
            $irc{'noscan'}->{lc($mask)}->{'servmask'} = $serv ? $serv : "*";
            $irc{'noscan'}->{lc($mask)}->{'new'}      = 1;
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'except-add'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask, $serv);
            &serv_record($nick, $services{'ScanServ'}->{'nick'}, "EXCEPT ADD $mask [$serv]");
            return 0;
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'except-in'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask, $serv);
            return 0;
          }
        }
      } elsif (uc($what) eq "DEL") {
        if (!$mask) {
          &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "EXCEPT <ADD|DEL|LIST> [mask]");
          return 0;
        } else {
          if (exists $irc{'noscan'}->{lc($mask)}) {
            &serv_record($nick, $services{'ScanServ'}->{'nick'}, "EXCEPT DEL $mask");
            delete $irc{'noscan'}->{lc($mask)};
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'except-del'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          } else {
            &putsock($sock{'irc'}, ":%s %s %s :$msg{'no-mask'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $mask);
            return 0;
          }
        }
      } elsif (uc($what) eq "LIST") {
        &serv_record($nick, $services{'ScanServ'}->{'nick'}, "EXCEPT LIST");
        &putsock($sock{'irc'}, ":%s %s %s :-- Listing E-lines --", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);
        &putsock($sock{'irc'}, ":%s %s %s :[Hostmask                      ] [Server mask              ]", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick);

        foreach (keys %{$irc{'noscan'}}) {
          ++$idx;
          &putsock($sock{'irc'}, ":%s %s %s :[%-30s] [%-25s]", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $irc{'noscan'}->{$_}->{'hostmask'}, $irc{'noscan'}->{$_}->{'servmask'});
        }

        &putsock($sock{'irc'}, ":%s %s %s :-- End of List [%d mask%s found] --", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $idx, $idx > 1 ? "s" : "");
        return 0;
      } else {
        &putsock($sock{'irc'}, ":%s %s %s :$msg{'syntax'}", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, "EXCEPT <ADD|DEL|LIST> [mask]");
        return 0;
      }
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'no-priv'});
      return 0;
    }
  } else {
    &putsock($sock{'irc'}, ":%s %s %s :%s", $services{'ScanServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $msg{'denied'});
    return 0;
  }
}

sub scanserv_sock {
  my ($host, $port, $proto) = (shift, shift, shift);
  my $test = 0;
  my $buffer;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => $host,
    PeerPort => $port,
    Timeout  => $cfg{'scan-timeout'}
  );

  if (!$socket) { return 0; }

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($cfg{'scan-timeout'});

    eval {
      $socket->autoflush(1);

      if ($proto eq "WGATE") {
        &putsock($socket, "%s\r", $cfg{'scan-test'});
        my $line = $socket->sysread($buffer, 512);
        if ($buffer =~ /$cfg{'scan-string'}/) { $test = 1; }
      } elsif ($proto eq "HTTP") {
        &putsock($socket, "CONNECT %s HTTP/1.0\r\n\r", $cfg{'scan-test'});
        my $line = $socket->sysread($buffer, 512);
        if ($buffer =~ /^HTTP\/1.[01] 200 [Cc]onnection/) { $test = 1; }
      } elsif ($proto eq "SOCKS4") {
        my @x  = split(/:/, $cfg{'scan-test'});
        my @ip = split(/\./, $x[0]);
        print $socket sprintf("%c%c%c%c%c%c%c%c%c", 4, 1, int($x[1] / 256), $x[1] % 256, $ip[0], $ip[1], $ip[2], $ip[3], 0);
        my $line = $socket->sysread($buffer, 512);
        if ((ord(substr($buffer, 0, 1)) == 0) && (ord(substr($buffer, 1, 1)) == 90)) { $test = 1; }
      } elsif ($proto eq "SOCKS5") {
        my @x  = split(/:/, $cfg{'scan-test'});
        my @ip = split(/\./, $x[0]);
        print $socket sprintf("%c%c%c", 5, 1, 0);
        print $socket sprintf("%c%c%c%c%c%c%c%c%c%c", 5, 1, 0, 1, $ip[0], $ip[1], $ip[2], $ip[3], int($x[1] / 256), $x[1] % 256);
        my $line = $socket->sysread($buffer, 512);
        if ((ord(substr($buffer, 0, 1)) == 5) && (ord(substr($buffer, 1, 1)) == 0)) { $test = 1; }
      }

      return 0;
    };
    alarm(0);
  };

  alarm(0);
  close($socket);
  return $test;
}

1;