# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: socket.pl

use strict;

if ($] < 5.006) {
  use vars qw(%access %bot %cfg %flags %fserv %im %irc %msg %services %sock $bit $irc_bind $localtime $server_bind $ts $TMP $VERSION);
} else {
  our (%access, %bot, %cfg, %flags, %fserv, %im, %irc, %msg, %services, %sock, $bit, $irc_bind, $localtime, $server_bind, $ts, $TMP, $VERSION);
}

sub connect2irc {
  $cfg{'irc-attempt'} = $ts;

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => (split(/:/, $cfg{'hub-server'}, 0))[0],
    PeerPort => (split(/:/, $cfg{'hub-server'}, 0))[1],
    Timeout  => 3
  );

  if (!$socket) {
    &putinfo("y", "Unable to connect to IRC server %s:%d", (split(/:/, $cfg{'hub-server'}, 0))[0], (split(/:/, $cfg{'hub-server'}, 0))[1]);
    return 0;
  }

  $socket->autoflush(1);

  $sock{'irc'} = $socket;
  $sock{'list'}->{$socket}->{'sock'}   = $socket;
  $sock{'list'}->{$socket}->{'type'}   = "IRC";
  $sock{'list'}->{$socket}->{'host'}   = $socket->peerhost();
  $sock{'list'}->{$socket}->{'port'}   = $socket->peerport();
  $sock{'list'}->{$socket}->{'fileno'} = $socket->fileno();
  vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 1;

  &putinfo("y", "Connected to IRC server %s:%d", (split(/:/, $cfg{'hub-server'}, 0))[0], (split(/:/, $cfg{'hub-server'}, 0))[1]);

  if (exists $irc{'server'}->{lc($cfg{'irc-fakeserv'})}) {
    delete $irc{'server'}->{lc($irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hub'})};
    delete $irc{'server'}->{lc($cfg{'irc-fakeserv'})};
  }

  $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'name'} = $cfg{'irc-fakeserv'};
  $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'info'} = $cfg{'irc-fakeinfo'};
  $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'time'} = $ts;
  $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'hopcount'} = 2;
  @{$irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'users'}} = (0, 0, 0);
  @{$irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}} = (0, 0, 0);

  if ($cfg{'server-type'} eq "hybrid") {
    &putsock($sock{'irc'}, "PASS %s :TS", $cfg{'hub-passwd'});
    &putsock($sock{'irc'}, "CAPAB :EX DE");
  } elsif ($cfg{'server-type'} eq "bahamut") {
    &putsock($sock{'irc'}, "PASS %s :TS", $cfg{'hub-passwd'});
    &putsock($sock{'irc'}, "CAPAB NICKIP SSJOIN TS3");
  } elsif ($cfg{'server-type'} eq "ultimate") {
    &putsock($sock{'irc'}, "PASS :%s", $cfg{'hub-passwd'});
  }

  &putsock($sock{'irc'}, "SERVER %s 1 :%s", $cfg{'irc-fakeserv'}, $cfg{'irc-fakeinfo'});

  if (($cfg{'server-type'} eq "hybrid") || ($cfg{'server-type'} eq "bahamut")) { &putsock($sock{'irc'}, "SVINFO 3 1 0 :%d", $ts); }

  foreach (keys %services) { &introduce_bot($services{$_}->{'bot'}); }

  if (exists $services{'PageServ'}) {
    if ($cfg{'icq-uin'}) {
      $im{'icq'} = undef;
      &connect2icq(0);
    }

    if ($cfg{'ymsg-uid'}) {
      $im{'ymsg'} = undef;
      &connect2ymsg(0);
    }

    if ($cfg{'aim-uid'}) {
      $im{'aim'} = undef;
      &connect2aim(0);
    }

    if ($cfg{'msn-uid'}) {
      $im{'msn'} = undef;
      &connect2msn(0);
    }
  }

  return 0;
}

sub connect2bot {
  my $nick = shift;

  if (!$bot{$nick}->{'sock'}) {
    $bot{$nick}->{'attempt'} = $ts;

    my $socket = IO::Socket::INET->new(
      Proto    => "tcp",
      PeerAddr => $bot{$nick}->{'host'},
      PeerPort => $bot{$nick}->{'port'},
      Timeout  => 3
    );

    if (!$socket) {
      &putinfo("b", "Unable to connect to bot [%s] (%s:%d)", $bot{$nick}->{'nick'}, $bot{$nick}->{'host'}, $bot{$nick}->{'port'});
      return 0;
    }

    $socket->autoflush(1);
    $bot{$nick}->{'sock'}   = $socket;
    $bot{$nick}->{'prefix'} = "->";

    $sock{'list'}->{$socket}->{'sock'}   = $socket;
    $sock{'list'}->{$socket}->{'type'}   = "BOT";
    $sock{'list'}->{$socket}->{'time'}   = $ts;
    $sock{'list'}->{$socket}->{'host'}   = $socket->peerhost();
    $sock{'list'}->{$socket}->{'port'}   = $socket->peerport();
    $sock{'list'}->{$socket}->{'fileno'} = $socket->fileno();
    vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 1;

    &putsock($socket, "%s %s %s", $bot{$nick}->{'nick'}, $services{'DataServ'}->{'nick'}, $bot{$nick}->{'passwd'});
    &putinfo("b", "Linked to %s [%s:%d]", $bot{$nick}->{'nick'}, $bot{$nick}->{'host'}, $bot{$nick}->{'port'});
    return 0;
  }
}

sub introduce_bot {
  my $bot   = shift;
  my $nick  = $services{$bot}->{'nick'};
  my $what  = 1;
  my $mode  = $cfg{'irc-umodes'};
  my $umode = '';

  if ((($bot eq "DataServ") && ($cfg{'irc-umodes'} !~ /o/)) || (exists $services{'ScanServ'} && ($bot eq "ScanServ") && ($cfg{'irc-umodes'} !~ /o/))) { $mode .= "o"; }

  if ($cfg{'server-type'} eq "bahamut") {
    if ($cfg{'irc-umodes'} !~ /a/) { $mode .= "a"; }
  } elsif ($cfg{'server-type'} eq "ultimate") {
    if ($cfg{'irc-umodes'} !~ /b/) { $mode .= "b"; }
    if ($cfg{'irc-umodes'} !~ /S/) { $mode .= "S"; }
  }

  foreach (split(//, $mode)) {
    if ($_ eq "+") {
      $what = 1;
    } elsif ($_ eq "-") {
      $what = 0;
    } elsif ($flags{'client'}->{$_}) {
      if ($what) {
        $umode |= $flags{'client'}->{$_};
      } else {
        $umode &= ~$flags{'client'}->{$_};
      }
    }
  }

  if ($cfg{'server-type'} eq "hybrid") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s :%s", $services{$bot}->{'nick'}, $ts, $mode, $services{$bot}->{'user'}, $cfg{'irc-fakeserv'}, $cfg{'irc-fakeserv'}, $services{$bot}->{'real'});
    &putsock($sock{'irc'}, ":%s MODE :%s", $services{$bot}->{'nick'}, $mode);
  } elsif ($cfg{'server-type'} eq "bahamut") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s 0 0 :%s", $services{$bot}->{'nick'}, $ts, $mode, $services{$bot}->{'user'}, $cfg{'irc-fakeserv'}, $cfg{'irc-fakeserv'}, $services{$bot}->{'real'});
    &putsock($sock{'irc'}, "SQLINE %s :Reserved for services", $services{$bot}->{'nick'});
  } elsif ($cfg{'server-type'} eq "ultimate") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s 0 :%s", $services{$bot}->{'nick'}, $ts, $services{$bot}->{'user'}, $cfg{'irc-fakeserv'}, $cfg{'irc-fakeserv'}, $services{$bot}->{'real'});
    &putsock($sock{'irc'}, "MODE %s %s", $services{$bot}->{'nick'}, $mode);
    &putsock($sock{'irc'}, "SQLINE %s :Reserved for services", $services{$bot}->{'nick'});
  }

  $irc{'client'}->{lc($nick)}->{'nick'} = $services{$bot}->{'nick'};
  $irc{'client'}->{lc($nick)}->{'user'} = $services{$bot}->{'user'};
  $irc{'client'}->{lc($nick)}->{'host'} = $cfg{'irc-fakeserv'};
  $irc{'client'}->{lc($nick)}->{'real'} = $services{$bot}->{'real'};
  $irc{'client'}->{lc($nick)}->{'serv'} = $cfg{'irc-fakeserv'};
  $irc{'client'}->{lc($nick)}->{'time'} = $ts;
  $irc{'client'}->{lc($nick)}->{'mode'} = $umode;

  if ($umode & $flags{'client'}->{'o'}) {
    if (++$irc{'stats'}->{'opers'}[0] > $irc{'stats'}->{'opers'}[1]) {
      &putinfo("x", "New max operator count: %d", $irc{'stats'}->{'opers'}[0]);
      $irc{'stats'}->{'opers'}[1] = $irc{'stats'}->{'opers'}[0];
      $irc{'stats'}->{'opers'}[2] = $ts;
    }

    if (++$irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}[0] > $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}[1]) {
      $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}[1] = $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}[0];
      $irc{'server'}->{lc($cfg{'irc-fakeserv'})}->{'opers'}[2] = $ts;
    }
  }

  &client_stats($nick);
  return 0;
}

sub disconnect {
  my $reason = shift;

  if (exists $services{'PageServ'}) {
    if ($sock{'icq'}) { &del_socket($sock{'icq'}); }
    if ($sock{'ymsg'}) { &del_socket($sock{'ymsg'}); }
    if ($sock{'aim'}) { &del_socket($sock{'aim'}); }
    if ($sock{'msn'}) { &del_socket($sock{'msn'}); }
  }

  foreach (keys %services) {
    &putsock($sock{'irc'}, ":%s QUIT :%s", $services{$_}->{'nick'}, $reason);
    if (($cfg{'server-type'} eq "bahamut") || ($cfg{'server-type'} eq "ultimate")) { &putsock($sock{'irc'}, ":%s UNSQLINE %s", $cfg{'irc-fakeserv'}, $services{$_}->{'nick'}); }
  }

  &putsock($sock{'irc'}, ":%s ERROR :%s", $cfg{'irc-fakeserv'}, $reason);
  &putsock($sock{'irc'}, ":%s QUIT", $cfg{'irc-fakeserv'});
  &del_socket($sock{'irc'});
  return 0;
}

sub putsock {
  my ($socket, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  print $socket $buffer . "\n";
  return 0;
}

sub putbnet {
  my ($format, @arg) = (shift, @_);
  my $buffer = sprintf($format, @arg);

  if (!%bot) { return 0; }

  foreach (keys %bot) {
    if ($bot{$_}->{'sock'}) { &putsock($bot{$_}->{'sock'}, $buffer); }
  }

  return 0;
}

sub putinfo {
  my ($flag, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  foreach (keys %{$sock{'list'}}) {
    if ($sock{'list'}->{$_}->{'type'} eq "DCC") {
      if (!$flag || ($access{lc($sock{'list'}->{$_}->{'access'})}->{'event'} & $flags{'event'}->{$flag})) { &putsock($sock{'list'}->{$_}->{'sock'}, "[%s] %s%s", $localtime, !$flag ? "" : "*** ", $buffer); }
    }
  }

  if ($cfg{'log-event'} & $flags{'event'}->{$flag}) { &putlog($flag, $buffer); }
  return 0;
}

sub del_socket {
  my $socket = shift;

  if ($sock{'list'}->{$socket}->{'type'} eq "IRC") {
    &putinfo("y", "Disconnected from IRC server %s:%d", (split(/:/, $cfg{'hub-server'}, 0))[0], (split(/:/, $cfg{'hub-server'}, 0))[1]);
    $cfg{'irc-connected'} = 0;
    delete $sock{'irc'};
    delete $irc{'channel'};
    delete $irc{'client'};
    delete $irc{'server'};

    $irc{'stats'}->{'opers'}[0] = 0;

    foreach (keys %{$irc{'stats'}->{'tld'}}) { $irc{'stats'}->{'tld'}->{$_}[0] = 0; }
    foreach (keys %{$irc{'stats'}->{'domain'}}) { $irc{'stats'}->{'domain'}->{$_}[0] = 0; }
    if ($TMP) { close($TMP); }

    if (exists $services{'PageServ'}) {
      if ($sock{'icq'}) { &del_socket($sock{'icq'}); }
      if ($sock{'ymsg'}) { &del_socket($sock{'ymsg'}); }
      if ($sock{'aim'}) { &del_socket($sock{'aim'}); }
      if ($sock{'msn'}) { &del_socket($sock{'msn'}); }
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "ICQ") {
    &putinfo("y", "Disconnected from ICQ server %s:%d", $im{'data'}->{'icq'}->{'server'}, $im{'data'}->{'icq'}->{'port'});
    delete $sock{'icq'};

    foreach (keys %{$im{'icq'}}) {
      my $nick = $im{'icq'}->{$_}->{'irc'};

      $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
      $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
      $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
      $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
      $irc{'seen'}->{lc($nick)}->{'what'} = "QUIT";
      $irc{'seen'}->{lc($nick)}->{'xtra'} = "Disconnected";

      &icq_killuser($_, "Disconnected");
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "YMSG") {
    &putinfo("y", "Disconnected from YMSG server %s:%d", $im{'data'}->{'ymsg'}->{'server'}, $im{'data'}->{'ymsg'}->{'port'});
    delete $sock{'ymsg'};

    foreach (keys %{$im{'ymsg'}}) {
      my $nick = $im{'ymsg'}->{$_}->{'irc'};

      $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
      $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
      $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
      $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
      $irc{'seen'}->{lc($nick)}->{'what'} = "QUIT";
      $irc{'seen'}->{lc($nick)}->{'xtra'} = "Disconnected";

      &ymsg_killuser($_, "Disconnected");
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "MSN") {
    &putinfo("y", "Disconnected from MSN server %s:%d", $im{'data'}->{'msn'}->{'server'}, $im{'data'}->{'msn'}->{'port'});
    delete $sock{'msn'};

    foreach (keys %{$im{'msn'}}) {
      my $nick = $im{'msn'}->{$_}->{'irc'};

      $irc{'seen'}->{lc($nick)}->{'nick'} = $irc{'client'}->{lc($nick)}->{'nick'};
      $irc{'seen'}->{lc($nick)}->{'time'} = $ts;
      $irc{'seen'}->{lc($nick)}->{'user'} = $irc{'client'}->{lc($nick)}->{'user'};
      $irc{'seen'}->{lc($nick)}->{'host'} = $irc{'client'}->{lc($nick)}->{'host'};
      $irc{'seen'}->{lc($nick)}->{'what'} = "QUIT";
      $irc{'seen'}->{lc($nick)}->{'xtra'} = "Disconnected";

      &msn_killuser($_, "Disconnected");
    }
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT-B") {
    &putinfo("b", "Lost bot connection to [%s:%d]", $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'});
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT") {
    foreach (keys %bot) {
      if ($sock{'list'}->{$socket}->{'sock'} eq $bot{$_}->{'sock'}) {
        &putinfo("b", "Bot %s [%s:%d] has disconnected", $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $sock{'list'}->{$socket}->{'port'});
        &putbnet("(%s) Bot %s (%s:%d) has disconnected", $services{'DataServ'}->{'nick'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $sock{'list'}->{$socket}->{'port'});
        delete $bot{$_}->{'sock'};
        delete $bot{$_}->{'prefix'};
        $bot{$_}->{'attempt'} = $ts;
      }
    }
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT-C") {
    &putinfo("b", "Unauthorized bot connection received from [%s:%d] on port %d", $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'}, $cfg{'bot-listen'});
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC") {
    my $login = $sock{'list'}->{$socket}->{'access'};
    my $prefix;

    if ($access{lc($login)}->{'flags'} & $flags{'access'}->{'a'}) {
      $prefix = "Admin";
    } elsif ($access{lc($login)}->{'flags'} & $flags{'access'}->{'o'}) {
      $prefix = "Oper";
    } else {
      $prefix = "User";
    }

    &putinfo(0, "%s %s (%s\@%s) has disconnected", $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
    &putbnet("(%s) %s %s (%s\@%s) has disconnected", $services{'DataServ'}->{'nick'}, $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
  } elsif (($sock{'list'}->{$socket}->{'type'} eq "DCC-B") || ($sock{'list'}->{$socket}->{'type'} eq "DCC-C")) {
    &putinfo("y", "Lost connection to [%s\@%s:%d]", $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'});
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-D") {
    &putinfo("y", "Lost DCC connection to %s [%s\@%s]", $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});

  } elsif (($sock{'list'}->{$socket}->{'type'} eq "FSERV-C") || ($sock{'list'}->{$socket}->{'type'} eq "FSERV-X")) {
    my $ip = $sock{'list'}->{$socket}->{'ip'};
    my ($vol, $dir, $file) = File::Spec->splitpath($sock{'list'}->{$socket}->{'absolute'});

    --$fserv{$ip}->{'xfer'};

    if ($fserv{$ip}->{'ondcc'}) {
      &putsock($fserv{$ip}->{'ondcc'}, "DCC Send of %s (%s) incomplete", $file, &format_size(-s $sock{'list'}->{$socket}->{'absolute'}));
    } else {
      &putsock($sock{'irc'}, ":%s %s %s :DCC Send of %s (%s) incomplete", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $fserv{$ip}->{'nick'}, $file, &format_size(-s $sock{'list'}->{$socket}->{'absolute'}));
    }

    &fileserv_queued($ip);
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV-G") {
    my $nick = $sock{'list'}->{$socket}->{'nick'};
    if (exists $irc{'client'}->{lc($nick)}) { &putsock($sock{'irc'}, ":%s %s %s :DCC Get of %s (%s) incomplete", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $sock{'list'}->{$socket}->{'file'}, &format_size($sock{'list'}->{$socket}->{'size'})); }
    &putinfo("s", "%s: %s!%s\@%s DCC Send %s (%s) [failed]", $services{'FileServ'}->{'nick'}, $nick, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'file'}, &format_size($sock{'list'}->{$socket}->{'size'}));
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV") {
    my $ip = $sock{'list'}->{$socket}->{'ip'};

    &putinfo("s", "%s: %s (%s\@%s) has disconnected", $services{'FileServ'}->{'nick'}, $fserv{$ip}->{'nick'}, $fserv{$ip}->{'user'}, $fserv{$ip}->{'host'});
    delete $fserv{$ip}->{'ondcc'};
    if (!$fserv{$ip}->{'xfer'}) { delete $fserv{$ip}; }
  }

  vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 0;
  $sock{'list'}->{$socket}->{'sock'}->close();
  delete $sock{'list'}->{$socket};
  return 0;
}

sub process_socket {
  my ($socket, $buffer) = (shift, shift);

  if ($sock{'list'}->{$socket}->{'type'} eq "IRC") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;
    my $cmd = (split(/ /, $buffer, 2))[0]; $cmd = uc($cmd);

    if (exists $server_bind->{$cmd}) {
      &{$server_bind->{$cmd}}($buffer);
      return 0;
    } else {
      $cmd = (split(/ /, $buffer, 3))[1];
      $cmd = uc($cmd);
      if (exists $irc_bind->{$cmd}) { &{$irc_bind->{$cmd}}($buffer); }
      return 0;
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "ICQ") {
    my @packet;

    foreach (split(//, $buffer)) { push @packet, ord($_); }

    for (my $idx = 0; $idx < scalar @packet; $idx++) {
      if ($im{'data'}->{'icq'}->{'flapbytes'} > 0) {
        push @{$im{'data'}->{'icq'}->{'flaprecv'}->{'data'}}, $packet[$idx];

        if (--$im{'data'}->{'icq'}->{'flapbytes'} <= 0) {
          &icq_incoming(\%{$im{'data'}->{'icq'}->{'flaprecv'}});
          $im{'data'}->{'icq'}->{'flaprecv'}  = undef;
          $im{'data'}->{'icq'}->{'flapbytes'} = 0;
        }
        next;
      }

      if ($packet[$idx] != 42) {
        &putinfo("y", "ICQ: recieved data missaligned!");
        &del_socket($sock{'icq'});
        return 0;
      }

      $im{'data'}->{'icq'}->{'flaprecv'}->{'chan_id'}   = &bytes2int(\@packet, $idx + 1, 1);
      $im{'data'}->{'icq'}->{'flaprecv'}->{'seq_id'}    = &bytes2int(\@packet, $idx + 2, 2);
      $im{'data'}->{'icq'}->{'flaprecv'}->{'size'}      = $im{'data'}->{'icq'}->{'flapbytes'} = &bytes2int(\@packet, $idx + 4, 2);
      $im{'data'}->{'icq'}->{'flaprecv'}->{'family_id'} = &bytes2int(\@packet, $idx + 6, 2);
      $im{'data'}->{'icq'}->{'flaprecv'}->{'sub_id'}    = &bytes2int(\@packet, $idx + 8, 2);
      $idx += 5;
    }

    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "YMSG") {
    my @packet;

my $test;
foreach (split(//, $buffer)) { $test .= sprintf("%02X ", ord($_)); }
putlog(0, "ymsg: $test");

    foreach (split(//, $buffer)) { push @packet, ord($_); }

    if (scalar @packet < 20) {
      &putinfo("y", "YMSG: recieved data missaligned!");
      &del_socket($sock{'ymsg'});
      return 0;
    }

    $im{'data'}->{'ymsg'}->{'flaprecv'}->{'size'}    = $im{'data'}->{'ymsg'}->{'flapbytes'} = &bytes2int(\@packet, 8, 2) + 20;
    $im{'data'}->{'ymsg'}->{'flaprecv'}->{'packlen'} = &bytes2int(\@packet, 8, 2);
    $im{'data'}->{'ymsg'}->{'flaprecv'}->{'service'} = &bytes2str(\@packet, 10, 2);
    $im{'data'}->{'ymsg'}->{'flaprecv'}->{'status'}  = &bytes2int(\@packet, 12, 4);
    $im{'data'}->{'ymsg'}->{'sessid'} = &bytes2int(\@packet, 16, 4);

putlog(0, "ymsg size: $im{'data'}->{'ymsg'}->{'flaprecv'}->{'size'} / serv: $im{'data'}->{'ymsg'}->{'flaprecv'}->{'service'} / ID: $im{'data'}->{'ymsg'}->{'sessid'}");

    for (my $idx = 0; $idx < scalar @packet; $idx++) {
      if ($im{'data'}->{'ymsg'}->{'flapbytes'} > 0) {
        push @{$im{'data'}->{'ymsg'}->{'flaprecv'}->{'data'}}, $packet[$idx];

        if (--$im{'data'}->{'ymsg'}->{'flapbytes'} <= 0) {
          &ymsg_incoming(\%{$im{'data'}->{'ymsg'}->{'flaprecv'}});
          $im{'data'}->{'ymsg'}->{'flaprecv'}  = undef;
          $im{'data'}->{'ymsg'}->{'flapbytes'} = 0;
        }
        next;
      }
    }
    
    return 0;

  } elsif (($sock{'list'}->{$socket}->{'type'} eq "MSN") || ($sock{'list'}->{$socket}->{'type'} eq "MSN-X")) {
    &msn_incoming($buffer);
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;
    &dcc_parse($socket, $buffer);
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-D") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

    if (++$sock{'list'}->{$socket}->{'warning'} > 3) {
      &del_socket($socket);
      return 0;
    } elsif ($buffer =~ /^\.(ID|IDENTIFY)/i) {
      my $passwd = (split(/ /, $buffer, 2))[1];
      &dcc_identify($socket, $passwd);
      return 0;
    } else {
      &putsock($socket, "You must .identify before you may use commands");
      return 0;
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-A") {
    my $remote = $socket->accept(); $remote->autoflush(1);

    $sock{'list'}->{$remote}->{'sock'}   = $remote;
    $sock{'list'}->{$remote}->{'type'}   = "DCC-B";
    $sock{'list'}->{$remote}->{'host'}   = $remote->peerhost();
    $sock{'list'}->{$remote}->{'port'}   = $remote->peerport();
    $sock{'list'}->{$remote}->{'fileno'} = $remote->fileno();
    $sock{'list'}->{$remote}->{'time'}   = $ts;
    $sock{'list'}->{$remote}->{'user'}  = "telnet";
    vec($bit, $sock{'list'}->{$remote}->{'fileno'}, 1) = 1;

    &putinfo("y", "Received connection from [%s:%d] on port %d", $sock{'list'}->{$remote}->{'host'}, $sock{'list'}->{$remote}->{'port'}, $cfg{'dcc-listen'});
    &putsock($remote, "Enter nickname:");
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-B") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s; $buffer =~ s/[\200-\377]//g;
    $sock{'list'}->{$socket}->{'nick'} = $buffer;
    $sock{'list'}->{$socket}->{'type'} = "DCC-C";
    $sock{'list'}->{$socket}->{'time'} = $ts;

    &putsock($socket, "Enter password:");
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "DCC-C") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;
    my $nick = $sock{'list'}->{$socket}->{'nick'};

    if (!exists $access{lc($nick)}) {
      &putsock($socket, "Invalid User\r");
      &putinfo("y", "Unauthorized connection attempt from [%s\@%s:%d] on port %d", $nick, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'}, $cfg{'dcc-listen'});
      $sock{'list'}->{$socket}->{'type'} = "DCC-E";
      &del_socket($socket);
      return 0;
    } elsif ($access{lc($nick)}->{'login'} eq $nick) {
      if ($buffer eq $access{lc($nick)}->{'passwd'}) {
        my $idx = 0;
        my $prefix;

        foreach (keys %{$sock{'list'}}) {
          if (($sock{'list'}->{$_}->{'type'} eq "DCC") && (++$idx > $cfg{'dcc-max'})) {
            &putsock($socket, $msg{'dcc-max'});
            $sock{'list'}->{$socket}->{'type'} = "DCC-E";
            &del_socket($socket);
            return 0;
          }
        }

        $sock{'list'}->{$socket}->{'type'}    = "DCC";
        $sock{'list'}->{$socket}->{'access'}  = $nick;
        $sock{'list'}->{$socket}->{'lastcmd'} = $ts;
        &send_motd($socket, $cfg{'dcc-motd'});

        if ($access{lc($nick)}->{'flags'} & $flags{'access'}->{'a'}) {
          $prefix = "Admin";
        } elsif ($access{lc($nick)}->{'flags'} & $flags{'access'}->{'o'}) {
          $prefix = "Oper";
        } else {
          $prefix = "User";
        }

        &putinfo(0, "%s %s (%s\@%s) has connected", $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
        &putbnet("(%s) %s %s (%s\@%s) has connected", $services{'DataServ'}->{'nick'}, $prefix, $sock{'list'}->{$socket}->{'nick'}, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'});
        return 0;
      } else {
        &putsock($socket, "Invalid password\r");
        &putinfo("y", "Unauthorized connection attempt from [%s\@%s:%d] on port %d", $nick, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'}, $cfg{'dcc-listen'});
        $sock{'list'}->{$socket}->{'type'} = "DCC-E";
        &del_socket($socket);
        return 0;
      }
    } else {
      &putsock($socket, "Invalid password\r");
      &putinfo("y", "Unauthorized connection attempt from [%s\@%s:%d] on port %d", $nick, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'port'}, $cfg{'dcc-listen'});
      $sock{'list'}->{$socket}->{'type'} = "DCC-E";
      &del_socket($socket);
      return 0;
    }

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT-A") {
    my $remote = $socket->accept(); $remote->autoflush(1);

    $sock{'list'}->{$remote}->{'sock'}   = $remote;
    $sock{'list'}->{$remote}->{'type'}   = "BOT-B";
    $sock{'list'}->{$remote}->{'host'}   = $remote->peerhost();
    $sock{'list'}->{$remote}->{'port'}   = $remote->peerport();
    $sock{'list'}->{$remote}->{'fileno'} = $remote->fileno();
    $sock{'list'}->{$remote}->{'time'}   = $ts;
    vec($bit, $sock{'list'}->{$remote}->{'fileno'}, 1) = 1;

    &putinfo("b", "Received bot connection from [%s:%d] on port %d", $sock{'list'}->{$remote}->{'host'}, $sock{'list'}->{$remote}->{'port'}, $cfg{'bot-listen'});
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT-B") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

    foreach (keys %bot) {
      if (!$bot{$_}->{'sock'} && ((split(/ /, $buffer, 0))[0] eq $services{'DataServ'}->{'nick'}) && ((split(/ /, $buffer, 0))[1] eq $bot{$_}->{'nick'}) && ((split(/ /, $buffer, 0))[2] eq $bot{$_}->{'passwd'})) {
        $bot{$_}->{'sock'}   = $socket;
        $bot{$_}->{'prefix'} = "<-";
        $sock{'list'}->{$socket}->{'type'} = "BOT";
        &putinfo("b", "Linked to %s [%s:%d]", $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $sock{'list'}->{$socket}->{'port'});
        &putbnet("(%s) Linked to %s [%s:%d]", $services{'DataServ'}->{'nick'}, $bot{$_}->{'nick'}, $bot{$_}->{'host'}, $sock{'list'}->{$socket}->{'port'});
        return 0;
      }
    }

    $sock{'list'}->{$socket}->{'type'} = "BOT-C";
    &del_socket($socket);
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV-A") {
    my $remote = $socket->accept(); $remote->autoflush(1);

    $sock{'list'}->{$remote}->{'sock'}     = $remote;
    $sock{'list'}->{$remote}->{'type'}     = "FSERV";
    $sock{'list'}->{$remote}->{'ip'}       = $remote->peerhost();
    $sock{'list'}->{$remote}->{'fileno'}   = $remote->fileno();
    $sock{'list'}->{$remote}->{'time'}     = $ts;
    $sock{'list'}->{$remote}->{'lastcmd'}  = $ts;
    $sock{'list'}->{$remote}->{'floodcnt'} = 0;
    $sock{'list'}->{$remote}->{'folder'}   = "/";
    $sock{'list'}->{$remote}->{'nick'}     = $sock{'list'}->{$socket}->{'nick'};
    $sock{'list'}->{$remote}->{'user'}     = $sock{'list'}->{$socket}->{'user'};
    $sock{'list'}->{$remote}->{'host'}     = $sock{'list'}->{$socket}->{'host'};
    vec($bit, $sock{'list'}->{$remote}->{'fileno'}, 1) = 1;

    my $ip = $sock{'list'}->{$remote}->{'ip'};

    if (!$fserv{$ip}) {
      $fserv{$ip}->{'ip'}      = $ip;
      $fserv{$ip}->{'nick'}    = $sock{'list'}->{$remote}->{'nick'};
      $fserv{$ip}->{'user'}    = $sock{'list'}->{$remote}->{'user'};
      $fserv{$ip}->{'host'}    = $sock{'list'}->{$remote}->{'host'};
      $fserv{$ip}->{'xfer'}    = 0;
      $fserv{$ip}->{'total'}   = 0;
      $fserv{$ip}->{'traffic'} = 0;
    }

    $fserv{$ip}->{'ondcc'} = $remote;

    &putinfo("s", "%s: %s (%s\@%s) has connected", $services{'FileServ'}->{'nick'}, $sock{'list'}->{$remote}->{'nick'}, $sock{'list'}->{$remote}->{'user'}, $sock{'list'}->{$remote}->{'host'});
    &send_motd($remote, $cfg{'fserv-motd'});
    &del_socket($socket);
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV-B") {
    my $remote = $socket->accept(); $remote->autoflush(1);

    $sock{'list'}->{$remote}->{'sock'}     = $remote;
    $sock{'list'}->{$remote}->{'type'}     = "FSERV-C";
    $sock{'list'}->{$remote}->{'ip'}       = $remote->peerhost();
    $sock{'list'}->{$remote}->{'fileno'}   = $remote->fileno();
    $sock{'list'}->{$remote}->{'start'}    = $ts;
    $sock{'list'}->{$remote}->{'time'}     = $ts;
    $sock{'list'}->{$remote}->{'absolute'} = $sock{'list'}->{$socket}->{'absolute'};
    $sock{'list'}->{$remote}->{'seek'}     = $sock{'list'}->{$socket}->{'seek'};
    vec($bit, $sock{'list'}->{$remote}->{'fileno'}, 1) = 1;

    &del_socket($socket);
    if ($cfg{'fserv-maxcps'} > 0) { $fserv{$sock{'list'}->{$remote}->{'ip'}}->{'limitcps'} = 0; }
    &fserv_send($remote, $sock{'list'}->{$remote}->{'seek'});
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV-C") {
    my $size = unpack("N", $buffer);

    if ($cfg{'fserv-maxcps'} > 0) {
      if ($fserv{$sock{'list'}->{$socket}->{'ip'}}->{'limitcps'} >= $cfg{'fserv-maxcps'}) {
        $fserv{$sock{'list'}->{$socket}->{'ip'}}->{'limitcps'} = 0;
        $sock{'list'}->{$socket}->{'type'} = "FSERV-X";
      } else {
        &fserv_send($socket, $sock{'list'}->{$socket}->{'seek'});
      }
    } else {
      &fserv_send($socket, $sock{'list'}->{$socket}->{'seek'});
    }

    return 0;
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV-G") {
    my $handle = $sock{'list'}->{$socket}->{'filehd'};
    print $handle $buffer;

    $sock{'list'}->{$socket}->{'time'}  = $ts;
    $sock{'list'}->{$socket}->{'read'} += length($buffer);

    print $socket pack("N", $sock{'list'}->{$socket}->{'read'});

    if ($sock{'list'}->{$socket}->{'read'} == $sock{'list'}->{$socket}->{'size'}) {
      close($handle);

      my $nick = $sock{'list'}->{$socket}->{'nick'};
      my $cps  = $sock{'list'}->{$socket}->{'size'};
      if ($ts - $sock{'list'}->{$socket}->{'start'} > 0) { $cps = $sock{'list'}->{$socket}->{'size'} / ($ts - $sock{'list'}->{$socket}->{'start'}); }

      if (exists $irc{'client'}->{lc($nick)}) { &putsock($sock{'irc'}, ":%s %s %s :DCC Get of %s (%s) complete (%s, %s/sec)", $services{'FileServ'}->{'nick'}, $cfg{'irc-comm'}, $nick, $sock{'list'}->{$socket}->{'file'}, &format_size($sock{'list'}->{$socket}->{'size'}), &format_time($sock{'list'}->{$socket}->{'start'}), &format_size($cps)); }
      &putinfo("s", "%s: %s!%s\@%s DCC Send %s (%s) [complete]", $services{'FileServ'}->{'nick'}, $nick, $sock{'list'}->{$socket}->{'user'}, $sock{'list'}->{$socket}->{'host'}, $sock{'list'}->{$socket}->{'file'}, &format_size($sock{'list'}->{$socket}->{'size'}));
      $sock{'list'}->{$socket}->{'type'} = "FSERV-Y";
      &del_socket($socket);
    }

    return 0;
  } elsif ($sock{'list'}->{$socket}->{'type'} eq "FSERV") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;
    &fserv_parse($socket, $buffer);
    return 0;

  } elsif ($sock{'list'}->{$socket}->{'type'} eq "BOT") {
    $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

    if (($buffer =~ /^<\S+> .*/) || ($buffer =~ /^\(\S+\) .*/)) {
      foreach (keys %{$sock{'list'}}) {
        if ($sock{'list'}->{$_}->{'type'} eq "DCC") { &putsock($sock{'list'}->{$_}->{'sock'}, $buffer); }
      }

      foreach (keys %bot) {
        if ($bot{$_}->{'sock'} ne $socket) { &putsock($bot{$_}->{'sock'}, $buffer); }
      }

      return 0;
    }
  }
}

sub http_get {
  my ($socket, $request, $host, $referer, $cookie) = (shift, shift, shift, shift, shift);

  &putsock($socket, "GET %s HTTP/1.0\r", $request);
  &putsock($socket, "Accept: text/html, */*\r");
  &putsock($socket, "Accept-Language: en-us\r");
  &putsock($socket, "User-Agent: Mozilla/5.0 (FXserv IRC services %s; http://fxserv.fxmakers.net/)\r", $VERSION);
  &putsock($socket, "Host: %s\r", $host);

  if ($referer) { &putsock($socket, "Referer: %s\r", $referer); }

  &putsock($socket, "Connection: Keep-Alive\r");

  if ($cookie) {
    &putsock($socket, "Cookie2: \$Version=\"1\"\r");
    &putsock($socket, "Cookie: %s\r", $cookie);
  }

  &putsock($socket, "\r");
  return 0;
}

sub http_post {
  my ($socket, $request, $host, $referer, $post, $boundary, $cookie) = (shift, shift, shift, shift, shift, shift, shift);

  &putsock($socket, "POST %s HTTP/1.0\r", $request);
  &putsock($socket, "Accept: text/html, */*\r");
  &putsock($socket, "Accept-Language: en-us\r");
  &putsock($socket, "User-Agent: Mozilla/5.0 (FXserv IRC services %s; http://fxserv.fxmakers.net/)\r", $VERSION);
  &putsock($socket, "Host: %s\r", $host);

  if ($referer) { &putsock($socket, "Referer: %s\r", $referer); }

  if ($boundary) {
    &putsock($socket, "Content-Type: multipart/form-data; boundary=%s\r", $boundary);
  } else {
    &putsock($socket, "Content-Type: application/x-www-form-urlencoded\r");
  }

  &putsock($socket, "Content-Length: %d\r", length($post));
  &putsock($socket, "Connection: Keep-Alive\r");

  if ($cookie) {
    &putsock($socket, "Cookie2: \$Version=\"1\"\r");
    &putsock($socket, "Cookie: %s\r", $cookie);
  }

  &putsock($socket, "\r");
  &putsock($socket, "%s\r", $post);
  &putsock($socket, "\r");
  return 0;
}

1;