# FXserv IRC services, Copyright (c) 2002 Christian Vandaele
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.
#
# $Id: ymsg.pl

use strict;
use Digest::MD5 qw(md5);

if ($] < 5.006) {
  use vars qw(%cfg %im %services %sock $ts);
} else {
  our (%cfg, %im, %services, %sock, $ts);
}

sub connect2ymsg {
  my $reconnect = shift;
  my $output;

  $cfg{'ymsg-attempt'} = $ts;

  if (!$reconnect) {
    $im{'data'}->{'ymsg'}->{'server'}    = "cs17.msg.yahoo.com";
    $im{'data'}->{'ymsg'}->{'port'}      = 5050;
    $im{'data'}->{'ymsg'}->{'flapbytes'} = 0;
    $im{'data'}->{'ymsg'}->{'flaprecv'}  = undef;
    $im{'data'}->{'ymsg'}->{'sessid'}    = 0x00000000;
    $im{'data'}->{'ymsg'}->{'cookieC'}   = "";
    $im{'data'}->{'ymsg'}->{'cookieT'}   = "";
    $im{'data'}->{'ymsg'}->{'cookieY'}   = "";
    @{$im{'data'}->{'ymsg'}->{'online'}} = ();
  }

  my $socket = IO::Socket::INET->new(
    Proto    => "tcp",
    PeerAddr => $im{'data'}->{'ymsg'}->{'server'},
    PeerPort => $im{'data'}->{'ymsg'}->{'port'},
    Timeout  => 3
  );

  if (!$socket) {
    &putinfo("y", "Unable to connect to YMSG server %s:%d", $im{'data'}->{'ymsg'}->{'server'}, $im{'data'}->{'ymsg'}->{'port'});
    return 0;
  }

  &putinfo("y", "Connected to YMSG server %s:%d", $im{'data'}->{'ymsg'}->{'server'}, $im{'data'}->{'ymsg'}->{'port'});
  $socket->autoflush(1);

  $sock{'ymsg'} = $socket;
  $sock{'list'}->{$socket}->{'sock'}   = $socket;
  $sock{'list'}->{$socket}->{'type'}   = "YMSG";
  $sock{'list'}->{$socket}->{'host'}   = $socket->peerhost();
  $sock{'list'}->{$socket}->{'port'}   = $socket->peerport();
  $sock{'list'}->{$socket}->{'fileno'} = $socket->fileno();
  vec($bit, $sock{'list'}->{$socket}->{'fileno'}, 1) = 1;

  $output->{'service'} = 0x004C;
  $output->{'status'}  = 0;

  &ymsg_output($sock{'ymsg'}, $output);

  return 0;
}

sub ymsg_output {
  my ($socket, $output) = (shift, shift);
  my @header;
  my $raw;
  my $len = 0; if (exists $output->{'data'}) { $len = scalar @{$output->{'data'}}; }

  push @header, &str2bytes("YMSG");
  push @header, &int2bytes(1, 0x09);
  push @header, &int2bytes(3, 0);
  push @header, &int2bytes(2, $len);
  push @header, &int2bytes(2, $output->{'service'});
  push @header, &int2bytes(4, $output->{'status'});
  push @header, &int2bytes(4, $im{'data'}->{'ymsg'}->{'sessid'});

my $test;
  foreach (@header) { $raw .= chr($_); $test .= sprintf("%02X ", $_); }
  foreach (@{$output->{'data'}}) { $raw .= chr($_); $test .= sprintf("%02X ", $_); }
putlog(0, "sent: $test");

  if ($raw) { print $socket $raw; }
  return 0;
}

sub ymsg_incoming {
  my $incoming = shift;
  my $recv;
  my $output;

  if ($incoming->{'service'} eq "L") {
    $output->{'service'} = 0x0057;
    $output->{'status'}  = 0;

    push @{$output->{'data'}}, &int2bytes(1, 0x31);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    &ymsg_output($sock{'ymsg'}, $output);
    return 0;

  } elsif ($incoming->{'service'} eq "W") {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    my @challenge   = split(//, $recv->{'94'});
    my $passwd_hash = &ymsg_y64(md5($cfg{'ymsg-passwd'}));
    my $crypt_hash  = &ymsg_y64(md5(&ymsg_md5_crypt($cfg{'ymsg-passwd'}, "_2S43d5f")));

    my $hash_passwd;
    my $hash_crypt;
    my $checksum;

    my $sv = ord($challenge[15]) % 8;

    if (($sv == 1) || ($sv == 6)) {
      $checksum = $challenge[ord($challenge[9]) % 16];
      $hash_passwd = sprintf("%s%s%s%s", $checksum, $cfg{'ymsg-uid'}, join('', @challenge), $passwd_hash);
      $hash_crypt  = sprintf("%s%s%s%s", $checksum, $cfg{'ymsg-uid'}, join('', @challenge), $crypt_hash);
    } elsif (($sv == 2) || ($sv == 7)) {
      $checksum = $challenge[ord($challenge[15]) % 16];
      $hash_passwd = sprintf("%s%s%s%s", $checksum, join('', @challenge), $passwd_hash, $cfg{'ymsg-uid'});
      $hash_crypt  = sprintf("%s%s%s%s", $checksum, join('', @challenge), $crypt_hash, $cfg{'ymsg-uid'});
    } elsif ($sv == 3) {
      $checksum = $challenge[ord($challenge[1]) % 16];
      $hash_passwd = sprintf("%s%s%s%s", $checksum, $cfg{'ymsg-uid'}, $passwd_hash, join('', @challenge));
      $hash_crypt  = sprintf("%s%s%s%s", $checksum, $cfg{'ymsg-uid'}, $crypt_hash, join('', @challenge));
    } elsif ($sv == 4) {
      $checksum = $challenge[ord($challenge[3]) % 16];
      $hash_passwd = sprintf("%s%s%s%s", $checksum, $passwd_hash, join('', @challenge), $cfg{'ymsg-uid'});
      $hash_crypt  = sprintf("%s%s%s%s", $checksum, $crypt_hash, join('', @challenge), $cfg{'ymsg-uid'});
    } elsif (($sv == 0) || ($sv == 5)) {
      $checksum = $challenge[ord($challenge[7]) % 16];
      $hash_passwd = sprintf("%s%s%s%s", $checksum, $passwd_hash, $cfg{'ymsg-uid'}, join('', @challenge));
      $hash_crypt  = sprintf("%s%s%s%s", $checksum, $crypt_hash, $cfg{'ymsg-uid'}, join('', @challenge));
    }

    $output->{'service'} = 0x0054;
    $output->{'status'}  = 0x5A55AA55;
    my $r06 = &ymsg_y64(md5($hash_passwd));
    my $r96 = &ymsg_y64(md5($hash_crypt));

    push @{$output->{'data'}}, &int2bytes(1, 0x30);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(1, 0x36);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($r06);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(2, 0x3936);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($r96);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(1, 0x32);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(1, 0x31);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(1, 0x31);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    &ymsg_output($sock{'ymsg'}, $output);

    return 0;
  } elsif ($incoming->{'service'} eq "T") {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

    if ($recv->{'60'} == 13) { &putinfo("y", "Could not log into Yahoo service. Please verify that your username and password are correctly typed."); }
    if ($recv->{'60'} == 14) { &putinfo("y", "Could not log into Yahoo service. Your account has been locked. Visit %s to reactivate it.", $recv->{'20'}); }
    &del_socket($sock{'ymsg'});

    return 0;

  } elsif ($incoming->{'service'} eq "U") {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    foreach (split(/\n/, $recv->{'59'})) {
      ($im{'data'}->{'ymsg'}->{'cookieC'}) = $_ =~ /^C\t(\S+)$/;
      ($im{'data'}->{'ymsg'}->{'cookieT'}) = $_ =~ /^T\t(\S+) expires/;
      ($im{'data'}->{'ymsg'}->{'cookieY'}) = $_ =~ /^Y\t(\S+) expires/;
    }

    return 0;
  } elsif ($incoming->{'service'} eq chr(1)) {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    if ($recv->{'1'} eq $cfg{'ymsg-uid'}) {
      $output->{'service'} = 0x0016;
      $output->{'status'}  = 0;

      push @{$output->{'data'}}, &int2bytes(1, 0x31);
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      push @{$output->{'data'}}, &int2bytes(2, 0x3235);
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      push @{$output->{'data'}}, &str2bytes("C=0" . chr(1) . "F=1,P=0,C=0,H=0,W=0,O=0,G=0" . chr(1) . "M=0,P=0,C=0,S=0,L=1,D=0,N=0,G=0,F=0,T=0");
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      &ymsg_output($sock{'ymsg'}, $output);
      $output = undef;

      $output->{'service'} = 0x0015;
      $output->{'status'}  = 0;

      push @{$output->{'data'}}, &int2bytes(2, 0x3236);
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      push @{$output->{'data'}}, &str2bytes("IMV,[none]=0,0,0;");
      push @{$output->{'data'}}, &int2bytes(2, 0xC080);
      &ymsg_output($sock{'ymsg'}, $output);

      if (scalar @{$im{'data'}->{'ymsg'}->{'list'}}) {
        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) { &ymsg_requser($im{'data'}->{'ymsg'}->{'list'}[$sub][0]); }
      }
    }

    if ($recv->{'7'}) {
      foreach (split(/\n/, $recv->{'7'})) { push @{$im{'data'}->{'ymsg'}->{'online'}}, $_; }

      if (!exists $recv->{'1'}) {
        my $nick = $recv->{'7'};
        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          if (($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $recv->{'7'}) && ($im{'data'}->{'ymsg'}->{'list'}[$sub][1] ne "")) { $nick = $im{'data'}->{'ymsg'}->{'list'}[$sub][1]; last; }
        }

        &ymsg_adduser($recv->{'7'}, $nick, $recv->{'10'});
      }
    }

    return 0;

  } elsif ($incoming->{'service'} eq chr(2)) {
    if ($incoming->{'packlen'} == 0) {
      if ($incoming->{'status'} == 0xFFFFFFFF) {
        &putinfo("y", "You have been logged out of the yahoo service, possibly due to a duplicate login.");
        &del_socket($sock{'ymsg'});
      }

      return 0;
    } else {
      my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

      if (scalar @{$im{'data'}->{'ymsg'}->{'list'}}) {
        for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
          if ($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $recv->{'7'}) { &ymsg_killuser($recv->{'7'}, "Signed off"); last; }
        }
      }
 
      return 0;
    }

  } elsif ($incoming->{'service'} eq chr(131)) {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});
    my $nick = $recv->{'7'};
    my $add  = 0;
foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    foreach (@{$im{'data'}->{'ymsg'}->{'online'}}) {
      if ($_ eq $recv->{'7'}) { $add = 1; last; }
    }

    if (!$add) { return 0; }

    for (my $sub = 0; $sub < scalar @{$im{'data'}->{'ymsg'}->{'list'}}; $sub++) {
      if (($im{'data'}->{'ymsg'}->{'list'}[$sub][0] eq $recv->{'7'}) && ($im{'data'}->{'ymsg'}->{'list'}[$sub][1] ne "")) { $nick = $im{'data'}->{'ymsg'}->{'list'}[$sub][1]; last; }
    }

    &ymsg_adduser($recv->{'7'}, $nick, $incoming->{'status'});
    return 0;

  } elsif ($incoming->{'service'} eq chr(46)) {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

# someone is adding me to his buddy list

    return 0;

  } elsif ($incoming->{'service'} eq chr(132)) {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    return 0;
  } elsif ($incoming->{'service'} eq "K") {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

# notify

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    return 0;
  } elsif ($incoming->{'service'} eq chr(6)) {
    my $recv = &ymsg_read_hash($incoming->{'data'}, 20, $incoming->{'packlen'});

# instant message

foreach (keys %{$recv}) {
putlog(0, "recv $_ : $recv->{$_}");
}

    return 0;
  }
}

sub ymsg_read_hash {
  my ($array, $start, $max) = (shift, shift, shift);
  my (@key, @value);
  my $pos  = $start;
  my $idx  = 0;
  my $get  = 0;
  my $name;
  my $recv;

  if (!$start) { $start = 0; }
  if (!$max) { $max = 65536; }

  while ($idx < $max) {
    if (($array->[$pos] == 0xC0) && ($array->[$pos + 1] == 0x80)) {
      $pos += 2;
      $idx += 2;

      if (!$get) {
        $get  = 1;
        $name = &bytes2str(\@key, 0, scalar @key);
        if (exists $recv->{$name}) { $recv->{$name} .= "\n"; }
        @key  = ();
      } else {
        $get = 0;
        $recv->{$name} .= &bytes2str(\@value, 0, scalar @value);
        @value = ();
        $name  = "";
      }
    }

    if (($array->[$pos] != 0xC0) && ($array->[$pos + 1] != 0x80)) {
      if (!$get) {
        push @key, $array->[$pos];
      } else {
        push @value, $array->[$pos];
      }

      ++$pos;
      ++$idx;
    }
  }

  return $recv;
}

sub ymsg_md5_crypt {
  my($pw, $salt) = (shift, shift);
  my $magic = "\$1\$";
  my $passwd;
  my $idx;

  $salt =~ s/^\Q$magic//;

  $salt =~ s/^(.*)\$.*$/$1/;
  $salt = substr($salt, 0, 8);

  my $ctx = new Digest::MD5;
  $ctx->add($pw);
  $ctx->add($magic);
  $ctx->add($salt);

  my ($final) = new Digest::MD5;
  $final->add($pw);
  $final->add($salt);
  $final->add($pw);
  $final = $final->digest;

  for ($idx = length($pw); $idx > 0; $idx -= 16) { $ctx->add(substr($final, 0, $idx > 16 ? 16 : $idx)); }

  for ($idx = length($pw); $idx; $idx >>= 1) {
    if ($idx & 1) {
      $ctx->add(pack("C", 0));
    } else {
      $ctx->add(substr($pw, 0, 1));
    }
  }

  $final = $ctx->digest;

  for ($idx = 0; $idx < 1000; $idx++) {
    my $ctx1 = new Digest::MD5;
    if ($idx & 1) {
      $ctx1->add($pw);
    } else {
      $ctx1->add(substr($final, 0, 16));
    }
    if ($idx % 3) { $ctx1->add($salt); }
    if ($idx % 7) { $ctx1->add($pw); }
    if ($idx & 1) {
      $ctx1->add(substr($final, 0, 16));
    } else {
      $ctx1->add($pw);
    }
    $final = $ctx1->digest;
  }

  $passwd = '';
  $passwd .= &base64(int(unpack("C", (substr($final, 0, 1))) << 16) | int(unpack("C", (substr($final, 6, 1))) << 8) | int(unpack("C", (substr($final, 12, 1)))), 4);
  $passwd .= &base64(int(unpack("C", (substr($final, 1, 1))) << 16) | int(unpack("C", (substr($final, 7, 1))) << 8) | int(unpack("C", (substr($final, 13, 1)))), 4);
  $passwd .= &base64(int(unpack("C", (substr($final, 2, 1))) << 16) | int(unpack("C", (substr($final, 8, 1))) << 8) | int(unpack("C", (substr($final, 14, 1)))), 4);
  $passwd .= &base64(int(unpack("C", (substr($final, 3, 1))) << 16) | int(unpack("C", (substr($final, 9, 1))) << 8) | int(unpack("C", (substr($final, 15, 1)))), 4);
  $passwd .= &base64(int(unpack("C", (substr($final, 4, 1))) << 16) | int(unpack("C", (substr($final, 10, 1))) << 8) | int(unpack("C", (substr($final, 5, 1)))), 4);
  $passwd .= &base64(int(unpack("C", substr($final, 11, 1))), 2);

  $final = '';
  return $magic . $salt . '$' . $passwd;
}

sub ymsg_y64 {
  pos($_[0]) = 0;

  my $result = join('', map(pack('u', $_) =~ /^.(\S*)/, ($_[0] =~ /(.{1,45})/gs)));
  $result =~ tr{` -_} {AA-Za-z0-9\._};

  my $padding = (3 - length($_[0]) % 3) % 3;
  if ($padding) { $result =~ s/.{$padding}$/'-' x $padding/e; }
  return $result;
}

sub ymsg_adduser {
  my ($uid, $nick, $status) = (shift, shift, shift);
  if (!$nick) { $nick = (split(/\@/, $uid, 0))[0]; }

  my $irc_nick = $cfg{'ymsg-nick'}; $irc_nick =~ s/\%/$nick/;

  $im{'ymsg'}->{$uid}->{'uid'}    = $uid;
  $im{'ymsg'}->{$uid}->{'nick'}   = $nick;
  $im{'ymsg'}->{$uid}->{'status'} = $status;
  $im{'ymsg'}->{$uid}->{'irc'}    = $irc_nick;

  if ($cfg{'server-type'} eq "hybrid") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s :%s", $irc_nick, $ts, "+i", $uid, $uid . ".messenger.yahoo.com", $cfg{'irc-fakeserv'}, $uid);
    &putsock($sock{'irc'}, ":%s MODE :%s", $irc_nick, "+i");
  } elsif ($cfg{'server-type'} eq "bahamut") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s %s 0 0 :%s", $irc_nick, $ts, "+i", $uid, $uid . ".messenger.yahoo.com", $cfg{'irc-fakeserv'}, $uid);
  } elsif ($cfg{'server-type'} eq "ultimate") {
    &putsock($sock{'irc'}, "NICK %s 1 %d %s %s %s 0 :%s", $irc_nick, $ts, $uid, $uid . ".messenger.yahoo.com", $cfg{'irc-fakeserv'}, $uid);
    &putsock($sock{'irc'}, "MODE %s %s", $irc_nick, "+i");
  }

  $irc{'client'}->{lc($irc_nick)}->{'nick'} = $irc_nick;
  $irc{'client'}->{lc($irc_nick)}->{'user'} = $uid;
  $irc{'client'}->{lc($irc_nick)}->{'host'} = $uid . ".messenger.yahoo.com";
  $irc{'client'}->{lc($irc_nick)}->{'real'} = $uid;
  $irc{'client'}->{lc($irc_nick)}->{'serv'} = $cfg{'irc-fakeserv'};
  $irc{'client'}->{lc($irc_nick)}->{'time'} = $ts;
  $irc{'client'}->{lc($irc_nick)}->{'mode'} |= $flags{'client'}->{'i'};

  &client_stats($irc_nick);

  if ($cfg{'ymsg-chan'}) {
    my $chan_ts = $ts;
    my $chan_mode;
    my $mode;

    if (exists $irc{'channel'}->{lc($cfg{'ymsg-chan'})}) {
      $chan_ts = $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'time'};
    } else {
      $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'channel'}  = $cfg{'ymsg-chan'};
      $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'time'}     = $ts;
      @{$irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'max'}}   = (1, $ts);
      @{$irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'topic'}} = (undef, undef, 0);
      $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'mode'} |= $flags{'channel'}->{'n'};
      $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'mode'} |= $flags{'channel'}->{'t'};
      $chan_mode = "nt";
    }

    $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'names'}->{lc($irc_nick)}->{'nick'} = $irc_nick;
    $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'names'}->{lc($irc_nick)}->{'mode'} = '';

    &putsock($sock{'irc'}, ":%s SJOIN %d %s +%s :%s%s", $cfg{'irc-fakeserv'}, $chan_ts, $irc{'channel'}->{lc($cfg{'ymsg-chan'})}->{'channel'}, $chan_mode, $mode, $irc_nick);
    &channel_stats($cfg{'ymsg-chan'});
  }

  return 0;
}

sub ymsg_killuser {
  my ($uid, $reason) = (shift, shift);
  my $nick = $im{'ymsg'}->{$uid}->{'irc'};

  delete $im{'ymsg'}->{$uid};
  if (exists $irc{'client'}->{lc($nick)}) {
    &putsock($sock{'irc'}, ":%s QUIT :%s", $nick, $reason);
    &del_client($nick);
  }

  return 0;
}

sub ymsg_requser {
  my $uid = shift;
  my $output;

  $output->{'service'} = 0x0083;
  $output->{'status'}  = 0;

  push @{$output->{'data'}}, &int2bytes(1, 0x31);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(1, 0x37);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes($uid);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3134);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3635);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes("contact");
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  &ymsg_output($sock{'ymsg'}, $output);

  return 0;
}

sub ymsg_eraseuser {
  my $uid = shift;

  &ymsg_killuser($uid, "Deleted");

  if ($sock{'ymsg'}) {
    my $output;

    $output->{'service'} = 0x0084;
    $output->{'status'}  = 0;

    push @{$output->{'data'}}, &int2bytes(1, 0x31);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(1, 0x37);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes($uid);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &int2bytes(2, 0x3635);
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    push @{$output->{'data'}}, &str2bytes("contact");
    push @{$output->{'data'}}, &int2bytes(2, 0xC080);
    &ymsg_output($sock{'ymsg'}, $output);
  }

  return 0;
}

sub ymsg_msguser {
  my ($nick, $uid, $query) = (shift, shift, shift);

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAVERSION %s\cA", $im{'ymsg'}->{$uid}->{'irc'}, $nick, "FXserv IRC services $VERSION");
    } elsif ($ctcp eq "PING") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cAPING %s\cA", $im{'ymsg'}->{$uid}->{'irc'}, $nick, (split(/ /, $1, 0))[1]);
    } elsif ($ctcp eq "TIME") {
      &putsock($sock{'irc'}, ":%s NOTICE %s :\cATIME %s\cA", $im{'ymsg'}->{$uid}->{'irc'}, $nick, scalar localtime());
    } elsif ($ctcp eq "DCC") {
    }

    return 0;
  }

  $query = "<" . $nick . "> " . $query;

  my $output;
  $output->{'service'} = 0x0006;
  $output->{'status'}  = 0x5A55AA56;

  push @{$output->{'data'}}, &int2bytes(1, 0x31);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes($cfg{'ymsg-uid'});
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(1, 0x35);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes($uid);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3134);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &str2bytes($query);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3937);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(1, 0x30);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3633);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3B30);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(2, 0x3634);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(1, 0x30);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(4, 0x31303032);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  push @{$output->{'data'}}, &int2bytes(1, 0x31);
  push @{$output->{'data'}}, &int2bytes(2, 0xC080);
  &ymsg_output($sock{'ymsg'}, $output);

  return 0;
}

sub ymsg_newnick {
  my ($uid, $nick) = (shift, shift);

  if (exists $im{'ymsg'}->{$uid}) {
    my $newnick = $cfg{'ymsg-nick'}; $newnick =~ s/\%/$nick/;
    my $oldnick = $im{'ymsg'}->{$uid}->{'irc'};

    &putsock($sock{'irc'}, ":%s NICK %s", $oldnick, $newnick);

    $irc{'client'}->{lc($newnick)}->{'nick'} = $newnick;
    $irc{'client'}->{lc($newnick)}->{'user'} = $irc{'client'}->{lc($oldnick)}->{'user'};
    $irc{'client'}->{lc($newnick)}->{'host'} = $irc{'client'}->{lc($oldnick)}->{'host'};
    $irc{'client'}->{lc($newnick)}->{'real'} = $irc{'client'}->{lc($oldnick)}->{'real'};
    $irc{'client'}->{lc($newnick)}->{'serv'} = $irc{'client'}->{lc($oldnick)}->{'serv'};
    $irc{'client'}->{lc($newnick)}->{'time'} = $irc{'client'}->{lc($oldnick)}->{'time'};
    $irc{'client'}->{lc($newnick)}->{'mode'} = $irc{'client'}->{lc($oldnick)}->{'mode'};;
    delete $irc{'client'}->{lc($oldnick)};

    $im{'ymsg'}->{$uid}->{'nick'} = $nick;
    $im{'ymsg'}->{$uid}->{'irc'}  = $newnick;
    return 0;
  }

  return 0;
}

1;